INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s

	STITLE Symbolic derivative

**********************************************************************
* EXPORT:
*
*	DERIVext	( obj, idnt -> obj' )
*			( obj, symb -> obj' )
*			( obj, {} -> {}' )
*
*		for idnt computes the derivative
*		for symb, computes the derivative, assuming the variable
*		is the first one in symb
*		for {}, computes the gradient of obj w.r.t. all idnt of {}
*		obj is allowed to be a list
*
*	DERIVIDNT1	( obj -> obj' )
*
*	DERIV		( symb -> symb' )
*			Derivative of symb wrt the idnt stored in LAM1
*
*	METADERIV	( meta -> meta )
*			Derivative of a meta representing a symbolic
*
**********************************************************************

* algebraic derivative
NULLNAME ssSYMDER
NULLNAME SYMDER		( 'idnt/lam' symb -> symb )
::
ASSEMBLE
	IF	f39
RPL
  NOTEQWcase 48ssSYMDER
ASSEMBLE
	ENDIF	f39
RPL
  SWAP DERIVext
;


**********************************************************************
* Desc:		Calculates the derivative of the object. For a list
*		argument calculates the gradient with respect to
*		the variables in the list. If the variable is
*		a symbolic, the first variable in it is used.
*		Note that the gradient is a vector quantity, thus
*		the result is returned as a list.
* Stack:	( ob id --> ob' )
*		( ob sym --> ob' )
*		( ob vector --> vector' )
**********************************************************************
NULLNAME DERIVext
::
  NOSTEPEVAL INsiSYMDER
;

* RPN derivative 
NULLNAME siSYMDER	( symb idnt/lam -> symb )
::
ASSEMBLE
	IF	f39
RPL
  NOTEQWcase 48siSYMDER
ASSEMBLE
	ENDIF	f39
RPL
  INsiSYMDER
;

LABEL INsiSYMDER
::
  CK&DISPATCH1
  idnt	DERIVIDNT
  lam	DERIVIDNT
  matrix DERIVMATRIX
  symb	
	::
          DUP' TYPEIDNTLAM? Find1stTrue	( obj, f[t], [ t, true | false ] )
          NOTcase INTVARERR
          DUPUNROT			( obj, t, f[t], t )
          DERIVIDNT			( obj, t, f'[t] )
          UNROT
          DERIVIDNT
          SWAP RDIVext
	;
;


LABEL DERIVMATRIX
::
  DUPTYPEMATRIX? NcaseTYPEERR
  INNERCOMP DUP#0=csedrp SETTYPEERR
  get1 SWAPDUP
  ZERO_DO (DO)			( v1, ..., vn, obj, #n )
    get1 3PICKSWAP DERIVext	( v2, ..., vn, obj, #n, obj' )
    UNROT
  LOOP
  SWAPDROP []N
;

* ====================================================================
* Desc:		Principal entry point for derivative with respect
*		to a identifier.
* Stack:	( ob id --> ob' )
* ====================================================================
NULLNAME DERIVIDNT
::
  1LAMBIND
    DERIVIDNT1
  ABND
;

* ====================================================================
* Desc:		Principal entry point for derivative with respect
*		to the identifier stored in LAM1.
* Stack:	( ob --> ob' )
* ====================================================================
NULLNAME DERIVIDNT1
::
  NOTMATRIXcase DERIV
  DUPNULL[]? ?SEMI
  INNERCOMP DUPDUP
  ZERO_DO (DO)
     get1 DERIVIDNT1 UNROT
  LOOP
  DROP []N
;

* ====================================================================
* Desc:
* Stack:
* ====================================================================
NULLNAME DERIV
::
  DUPTYPEARRY? case :: Z0 MAT*SCL ;
  QUOTEEVAL SYMBINCOMP INMETADERIV 
  DUP#0=csedrp Z0
  CKSYMBN
;

LABEL QUOTEEVAL
* symb -> symb, evaluate embedded quote
::
  DUPTYPESYMB? NOT?SEMI
				( symb )
  DUP' TYPESYMB? Find1stTrue	( symb obj true | symb false )
  casedrop SEVAL
;

* ====================================================================
* Desc:		Derivative of Meta object
* Stack:	( Meta --> Meta' )
* ====================================================================
NULLNAME METADERIV
::
  ?ATTNQUIT
  DUP#1= case INMETADERIV
  stepflag TestSysFlag NOTcase INMETADERIV
  SYMBN 1GETLAM ONE SYMBN SWAP TWO addt: xDER
;
	
LABEL INMETADERIV
::
* si meta ne contient pas lam1, renvoie un meta-nul

  1GETLAM EQUALPOSMETA		( meta, var, # )
  #0=
* ITE FALSE :: OBJ>R ' xFCNAPPLY EQUALPOSMETA R>OBJ ROTDROPSWAP #0= ;
  casedrop NDROPZERO

				( meta, var )
  DROP				( meta )
  DUP#1= case2drop Z1ONE
  M1st+case METADER+
  M1st-case METADER-
  M1st*case METADER*
  M1st/case METADER/
  M1st^case METADER^
  M1stNEGcase METADERNEG
  M1stFcncase METADERFCN
  AEQ1stcase xDER METADERDER
  AEQ1stcase x\.S_ALGEBRAIC METADERI4
  AEQ1stcase 
ASSEMBLE
	IF	f39
RPL
  xint
ASSEMBLE
	ELSE	f39
RPL
  xINT
ASSEMBLE
	ENDIF	f39
RPL
  METADERI3
  AEQ1stcase xIFTE METADERIFTE
  AEQ1stcase xXROOT METADERXROOT
  AEQ1stcase x= METADER=
  OVER METADEROP EQLookup		( meta [ op FALSE | prg TRUE ] )
  NOTcasedrop METADERIVd1
  ROTDROP SWAP#1-SWAP EVAL
;

LABEL METADERIVd1
::					( meta )
  CKSYMBN EXPR>				( o1 .. on % op )
  SWAP COERCE SWAP DO>STRID		( o1 .. on # $op )
  "d1" SWAP&$ $>ID			( o1 .. on # $d1op )
  SIMPUSERIDNT				( symb )
  SYMBINCOMP				( meta )
;

* --------------------------------------------------------------------
* Desc:		Like DO>STR but without quotes for id
* Stack:	id/obj -> $
* --------------------------------------------------------------------
NULLNAME DO>STRID
:: 
  DUPTYPEIDNT? ITE ID>$ DO>STR
;
	
* --------------------------------------------------------------------
* Desc:		Table of derivable functions and the respective
*		derivative calculation subroutines.
* --------------------------------------------------------------------
NULLNAME METADEROP
{	
  xEXP METADEREXP
  xLN METADERLN
  xINV METADERINV
  xSQ METADERSQ
  x\v/ METADERSQRT
  xSIN METADERSIN
  xCOS METADERCOS
  xTAN METADERTAN 
  xSINH METADERSINH
  xCOSH METADERCOSH
  xTANH METADERTANH
  xASIN METADERASIN
  xACOS METADERACOS
  xATAN METADERATAN 
  xASINH METADERASH
  xACOSH METADERACH
  xATANH METADERATH
  xEXPM METADEREXP
  xLNP1 METADERLNP1
  xALOG METADERALOG
  xLOG METADERLOG
  xSIGN NDROPZ0ONE
  xABS METADERABS
}

* --------------------------------------------------------------------
* Desc:		Meta derivative of addition
* Stack:	( Meta&+ --> Meta' )
* Note:		label since added after 1st public release
* --------------------------------------------------------------------
LABEL METADER=
::
  plDRPpZparg METADERIV		( right left' )
  DUP#0=ITDROPZ0ONE		( right left' )
  psh METADERIV			( left' right' )
  DUP#0=ITDROPZ0ONE		( left' right' )
  top& addt: x=
;

LABEL DUP#0=ITDROPZ0ONE
::
  DUP#0=csedrp Z0ONE
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of addition
* Stack:	( Meta&+ --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADER+
::
  plDRPpZparg METADERIV 
  DUP#0=csedrp METADERIV
  psh METADERIV ckaddt+
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of substraction
* Stack:	( Meta&- --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADER-
::
  plDRPpZparg METADERIV 
  DUP#0=csedrp METADER&NEG
  psh METADERIV ckaddt-
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of multiplication
* Stack:	( Meta&* --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADER*
::
  plDRPpZparg			( v, u )
  dup METADERIV			( v, u, u' )
  rotdup METADERIV		( u, u', v, v' )
  psh rot pshckaddt*		( u, v', u'*v )
  DUP#0=csedrp ckaddt*		( -> u*v' )
  unrot 			( u'*v, u, v' )
  ckaddt* ckaddt+
;

LABEL pshckaddt*
::
  DUP#0=csedrp NDROPZERO
  OVER j#1=case 2DROP
  psh ckaddt* 
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of division
* Stack:	( Meta&/ --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADER/
::
  plDRPpZparg			( v, u )
  dup METADERIV			( v, u, u' )
  rotdup METADERIV		( u, u', v, v' )
  psh dup 4roll			( u, v', v, v, u' )
  ckaddt*			( u, v', v, v*u' )
  ::
     DUP#0=csedrp
     ::				( u, v', v )
        unrot ckaddt* metaneg
     ;
				( u, v', v, v*u' )
     4roll 4roll ckaddt*	( v, v*u', u*v' )
     ckaddt-
  ;
				( v, v*u'-u*v' )
  psh addtSQ top& addt/
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of power
* Stack:	( Meta&^ --> Meta' )
* Note:		[u^v] ' = u^[v-1] u'+ u^v*v'*ln[u]
* --------------------------------------------------------------------
NULLNAME METADER^
::
  dup dup			( u^v, u^v, u^v )
  plDRPpZparg psh		( u^v, u^v, u, v )
  dup METADERIV			( u^v, u^v, u, v, v' )
  rotdup METADERIV		( u^v, u^v, v, v', u, u' ) 
  psh addtLN rot ckaddt*	( u^v, u^v, v, u', ln[u]*v' )
  4roll psh ckaddt*		( u^v, v, u', u^v*ln[u]*v' )
  4roll				( v, u', u^v*ln[u]*v', u^v )
  ROT OBJINT? SWAP 4UNROLL	( v, u', u^v*ln[u]*v', u^v, ? )
  ITE
     ::	ROT Z1 QSub 
	DUP Z1 EQUAL		( ..., o1, .., on-2, on, #n, k-1, ?k-1=1 )
        NOTcase UNROT
        DROPSWAPDROP #2-
     ; 
     ::	addt2: Z1 x- 4ROLLSWAP ;

				( v, u', u^v*ln[u]*v', u^[v-1] )
  4roll ckaddt*			( u', u^v*ln[u]*v', u^[v-1]*v )
  rot ckaddt* 
  DUP#0=csDROP
  psh ckaddt+
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of a function
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERFCN
::
  ( * first evaluate user fcn * )
					( meta )
  dup SYMBN EXPR>			( meta arg1..argk %k fnc_name )
  SWAP COERCE SWAP 
  METADEREVALFCN CKALG?	( meta symb flag )
  NOTcasedrop METADERUNDEFFCN		( invalid user fcn, -> meta )
					( meta symb )
  OVER#2+UNROLL NDROP SYMBINCOMP	( meta )
  ( * look if we have still a user fcn, if not -> METADERIV * )
  M1stFcncase METADERUNDEFFCN
  METADERIV
;

LABEL METADEREVALFCN
( arg1...argn #n fnc_name -> value )
* return #0 if error
::					
  DEPTH 3PICK #- 1LAMBIND
  ERRSET SIMPUSERFCN ERRTRAP TRAPMETADERFCN
  ABND
;

LABEL TRAPMETADERFCN
:: DEPTH #2+ 1GETLAM #- NDROP ZERO ;

LABEL METADERUNDEFFCN
::
  DERARG		( arg1...argk, der1...derk, #k , fnct utilisateur )
  DO>STRID SWAPDUP SWAP#1+SWAP
  OVER#2+UNROL psh CKSYMBN
			( der1, ..., derk, id, #k+1, symbol arg1 ... argk ; )
  Z0 ROT #1-
  ZERO_DO (DO)			( der1, .., derl, id, argl, res )
     OVER			( der1..derl, id, argl, res, argl )
     "d"
     ISTOP-INDEX #>Z
     DO>STR &$			( der1..derl, id, argl, res, argl, "der_l" )
     5PICK &$ $>ID ONE SYMBN	( der1..derl, id, argl, res, argl, der_lfnct )
     SWAP SYMBINCOMP get1
     ISTOP@
     ' xFCNAPPLY 4ROLL #3+ SYMBN ( der1..derl, id, argl, res, der_lfnct[argl] )
     5ROLL SWAP x*ext x+ext
  LOOP
  UNROT2DROP			( symb )
  SYMBINCOMP
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of a derivative of a function
* Stack:	( symb_id_; sym_fcn_; xDER #3  --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERDER
::
  DROP EvalNoCK SYMBINCOMP INMETADERIV
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of a defined integral
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERI4
::
  DERARG
  2DROP			( inf sup fnct' var_muette inf', sup' fnct' Z0 )
  DROP 5UNROLL		( inf sup fnct' fnct var_muette, inf' sup' )
  2OVER 8PICK MEVALext
  x*ext			( inf sup fnct' fnct var_muette inf' sup'*fnct[sup] )
  2SWAP DUP 5UNROLL	( inf sup fnct' var_muette inf' sup'*fnct fnct muette )
  8PICK MEVALext	( inf sup fnct' var_muette inf' sup'*fnct fnct )
  ROTSWAP x*ext x-ext	( inf sup fnct' var_muette sup'*fnct-inf'*fnct )
  5UNROLL		( symb inf sup fnct' var_muette )
  SWAP symcomp
  SWAP symcomp		( symb inf sup fnct' var_muette )
  MEVALMAKEs		( symb meta )
  get1 SYMBINCOMP psh	( symb-meta meta )
  MetaAdd		( meta )
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of an undefined integral
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERI3
::
  DERARG		( fnc intvar apply derfcn derintvar derapply #3 xINT )
  2DROP SWAP		( fnc intvar apply derfcn derapply derintvar )
  Z0 EQUAL 
  NOTcase :: BINT46 ERABLEERROR ; ( invalid derivative of integration var )
			( fnc intvar apply derfcn derapply )
  5ROLL 5PICK 5PICK 	( intvar apply derfcn derapply fcn intvar apply )
  MEVALext		( intvar apply derfcn derapply fcn|intvar=apply )
  x*ext	OBJ>R		( intvar apply derfcn )
  symcomp ROT symcomp	( apply derfcn intvar )
  xf[t]t2int		( meta-int[derfcn,intvar,apply] )
  R>OBJ SYMBINCOMP	( meta-int meta )
  MetaAdd		( meta )
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of IFTE
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERIFTE
::					( meta )
  pulldrop pshzerpsharg			( meta meta )
  pshzerpsharg				( else #1 then #1 meta-cond )
  get1 DROP get1 DERIV			( else #1 meta-cond then' )
  SWAP#1+ get1 DROP get1 DERIV		( meta-cond-then' else' )
  SWAP#1+ addt: xIFTE			( meta )
;

* --------------------------------------------------------------------
* Desc:		Find derivative of arguments
* Stack:	( meta-symb -> arg1 ... argk der1 ... derk #k op )
* --------------------------------------------------------------------
NULLNAME DERARG
::
  SYMBN EXPR> OBJ>R COERCE	( arg1, ..., argk, #k )
  dup DUPDUP
  ZERO_DO (DO)
     get1 DERIV UNROT
  LOOP
  DROP				( arg1, ..., argk, #k, der1, ..., derk, #k )
  get1 DROP
  R>OBJ				( arg1 ... argk der1 ... derk #k op )
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of XROOT
* Stack:	( Meta&XROOT --> Meta' )
* Note:		[XROOT[v,u]] ' = XROOT[v,u]/v *[u'/u -[v'/v]*ln[u] ]
* --------------------------------------------------------------------
LABEL METADERXROOT
::
  dup plDRPpZparg				( xroot u v )
  dup dup METADERIV				( xroot u v v v' )
  DUP#0=ITE swapdrop				
    :: pshtop& addt/				( xroot u v v'/v )
       rotdup 4unroll				( xroot u v v'/v u )
       addtLN top& addt*
       ;					( xroot u v v'/v*ln[u] )
  rotdup METADERIV				( xroot v v'/v. u u' )
  DUP#0=ITE :: DROPdrop metaneg ;
    :: pshtop& addt/ psh ckaddt- ;		( xroot v u'/u-v'/v. )
  unrot top& addt/ top& addt*
;
	
* --------------------------------------------------------------------
* Desc:		Meta derivative of EXP
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADEREXP
::
  dup addt: xEXP psh METADERIV ckaddt*
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of LN
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERLN
::
  AEQ1stcase xABS :: pulldrop METADERLN ;
  dup METADERIV psh top& addt/
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of LNP1
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERLNP1
::
  Z1ONE MetaAdd METADERLN
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of LOG
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERLOG
::
  METADERLN 
  DUP#0=csedrp ZERO
  metaln10 MetaDiv
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of ALOG
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERALOG
::
  metaln10 MetaMul METADEREXP
;


* --------------------------------------------------------------------
* Desc:		Meta derivative of ABS
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERABS
::
  dup METADERIV DUP#0=csedrp NDROPZERO
  psh addtSIGNonly top& addt*
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of INV
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERINV
::
  dup METADER&NEG psh addtSQ top& addt/
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of NEG
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERNEG
:: 
  pulldrop METADER&NEG
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of SQRT
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERSQRT
::
  dup METADERIV psh addtSQRTonly meta2* top& addt/
;

* --------------------------------------------------------------------
* Desc:		Meta derivative and negate
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADER&NEG
:: 
  METADERIV metaneg 
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of SQ
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERSQ
::
  dup METADERIV ckaddt* meta2*
;

LABEL Z180
:: ZINT 180 TOTEMPOB ;

LABEL Z200
:: ZINT 200 TOTEMPOB ;
	
LABEL DERSINCOSTAN
* multiply expression by pi/180 if degree mode or pi/200 if grade mode
* meta -> meta
::
  RAD? ?SEMI
  metapi DEG?Z180Z200 ONE MetaDiv		( meta meta-pi/180|200 )
  psh ckaddt* 
;

LABEL DERASINACOSATAN
* multiply expression by 180/pi if degree mode or 200/pi if grade mode
* meta -> meta
::
  RAD? ?SEMI
  DEG?Z180Z200 ONE metapi MetaDiv		( meta meta-pi/180|200 )
  psh ckaddt* 
;
	
* --------------------------------------------------------------------
* Desc:		Meta derivative of SIN
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERSIN
::
  dup addtCOS pshder*
  DERSINCOSTAN
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of COS
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERCOS
::
  dup addtSIN pshder* metaneg
  DERSINCOSTAN
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of TAN
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERTAN
::
  dup addtTAN metasq+1 pshder*
  DERSINCOSTAN
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of SINH
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERSINH
::
  dup addtCOSH pshder*
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of COSH
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERCOSH
::
  dup addtSINH pshder*
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of TANH
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERTANH
::
  dup addtTANH meta1-sq pshder*
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of ASIN
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERASIN
::
  dup meta1-sq SQRTINVpshd*
  DERASINACOSATAN
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of ACOS
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERACOS
::
  METADERASIN metaneg
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of ATAN
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERATAN
::
  dup metasq+1 addtINV pshder*
  DERASINACOSATAN
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of ASINH
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERASH
::
  dup metasq+1 SQRTINVpshd*
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of ACOSH
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERACH
::
  dup metasq-1 SQRTINVpshd*
;

* --------------------------------------------------------------------
* Desc:		Meta derivative of ATANH
* Stack:	( Meta --> Meta' )
* --------------------------------------------------------------------
NULLNAME METADERATH
::
  dup meta1-sq addtINV pshder*
;


* --------------------------------------------------------------------
* Desc:		Meta derivative utility.
* Stack:	( Meta1 Meta2 --> Meta2&Meta1'&* )
* --------------------------------------------------------------------
NULLNAME pshder*
::
  psh METADERIV ckaddt*
;

* --------------------------------------------------------------------
* Desc:		Meta derivative utility.
* Stack:	( Meta1 Meta2 --> Meta2&SQRT&INV&Meta1'&* )
* --------------------------------------------------------------------
NULLNAME SQRTINVpshd*
::
  addtSQRTonly addtINV pshder*
;

* --------------------------------------------------------------------
* Desc:		Multiply 2 meta objects with trivial simplifications.
* Stack:	( Meta1 Meta2 --> Meta1*Meta2
* --------------------------------------------------------------------
NULLNAME ckaddt*
::
  DUP#0=csedrp NDROPZERO
  OVER j#1=case 2DROP
  iflag SysITE DUP#1= FALSE 	( meta1 meta2 ? )
  ITE pshtop& top& 		( meta1&2 )
  addt*				( meta1&2&* )
;

* --------------------------------------------------------------------
* Desc:		Add 2 meta objects with trivial simplifications.
* Stack:	( Meta1 Meta2 --> Meta1+Meta2
* --------------------------------------------------------------------
NULLNAME ckaddt+
::
  DUP#0=csDROP
  top& OVER EQ: xNEG NOTcase addt+
  pulldrop addt-
;

* --------------------------------------------------------------------
* Desc:		Substract 2 meta objects with trivial simplifications.
* Stack:	( Meta1 Meta2 --> Meta1+Meta2
* --------------------------------------------------------------------
NULLNAME ckaddt-
::
  DUP#0=csDROP
  top& OVER EQ: xNEG NOTcase addt-
  pulldrop addt+
;


**********************************************************************

