INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s

	STITLE GCD (Euclide) and simplifications

**********************************************************************
* EXPORT:
* SLVARext	( {} -> {} )
*		Simplifies all elements of the list that are supposed to
*		be variables
*
* SIMP1ext	( obj -> obj )
*		simplifies symbolically one object
*
* SIMPVAR	( obj -> obj )
*		simplifies symbolically one variable
*
* ID>DERext	( idnt -> {}, idnt' )
*		Looks for all d1, d2, ... at the beginning of the name of
*		idnt to determine if idnt represent a derivative of a user
*		function. Stops if at a time the stripped idnt is in the
*		current directory.
*		Example 'd2d1Y' returns { #2 } 'd1Y' if 'd1Y' is defined
*		or { #2 #1 } 'Y', ... 
*
* SIMPIDNT	( idnt -> obj )
*		Evaluate idnt (looks recursively for it's content if defined)
*		Does not error for circular definition, but displays
*		a warning
*
* SIMPSQRT	( sqrt[obj] -> sqrt[obj] )
*		Presimplification of square roots for SIMP1ext
*
* SIMPNUMS	( obj, op -> obj ) if op != integral
*		( obj, op -> NUM[obj] [ RDROP] ) otherwise
*		Integral simplification in numeric mode
*
* SIMPSYMBS	( inf, sup, fcn, var -> integral[inf,sup,fnc,var] )
*
* SIMPUSERFCN	( o1, ..., on, #n, id -> id[] )
*		simplification of user fcns, tests for derivative of user fcn
*
* EVALUSERFCN	( v1, ..., vn, #n, fcn -> f[] )
*		evaluate a user fcn with stack checking
*
* SIMP|		( obj, liste -> obj )
*		execute the | operator
*
* DENOLCMext	( list -> o )
*		o is the lcm of the denominator of the elements of list
*		if stk1 is not a list returns the denominator of stk1
*
* METADENOLCM	( meta -> o )
*		o is the lcm of the denominators of the elements of meta
*
* SWPSIMPNDXF	( o2, o1 -> o1/o2 )
*		simplified fraction (internal)
*
* SIMPNDXFext	( o2, o1 -> o2/o1 )
*		simplified fraction (internal)
*
* SIMPext	( o1, o2 -> o1', o2' )
*		Simplify 2 objects (internal representation)
*		Checks that o2 is not a complex or an irrquad
*		because decomposition of the corresponding fraction
*		with larg would -> Try to recover memory
*
* MAKEPROFOND	( o, # -> { .. { o } .. } )
*		Embedding of o in # lists
*
* SIMPGCDext	( o1, o2, gcd -> o1 div gcd, o2 div gcd )
*
* SIMP3ext	( a, b -> g, a'', b'' )
*		g is the gcd of a and b, a''=a/g and b''=b/g
*
* LPGCDext	( list -> list, obj )
*		obj is the GCD of all elements of list
*		Algorithm far from optimal :-(
*
* SLOWGCDext	( c, 1, A, B -> c* pgcd[A,B] )
*		Euclidean algorithm for polynomial GCD
*		Used if A or B contains irrquad
*		c is the gcd of the contents of the original polynomials
*		returned after failure of GCDHEUext
*
* GCDext	( o2, o1 -> gcd )
*		Generic internal GCD
*
* CGCDext	( C2, C1 -> C )
*		GCD for Gauss integers
*
* CMODext	( C2, C1 -> C1, C2 mod C1 )
*
* ZGcd,ZGCDext	( Z2, Z1 -> Z )
*		GCD for integers
*
* SQFF2ext	( meta' -> meta'' )
*		meta' is a meta of lists l1, ..., ln, #n-1 [and not #n]
*		returns the same type of meta such that all elements
*		of a list of meta" are equal or prime together with all
*		elements of another list of meta" 
*
**********************************************************************


**********************************************************************
* Desc:		Simplifies all elements of the list that are
*		supposed to be variables.
* Stack:	( Lvar --> Lvar' )
**********************************************************************
NULLNAME SLVARext
::
  nonrecurflag TestSysFlag ?SEMI
  DUPTYPELIST? NOT?SEMI
  DUPNULL{}? ?SEMI
  INNERDUP #1-SWAP
  ZERO_DO			( var1, ..., varn, #n )
     get1			( var2, ..., varn, #n, var1 )
     ALGCASCOMPEVAL
  SWAPLOOP
  #1+ {}N
;


**********************************************************************
* Desc:		Simplify one object like EVAL
* Stack:	( symb --> symb' )
**********************************************************************
NULLNAME SIMPLIFY
::
  SYMBINCOMP metasimp CKSYMBN
;

**********************************************************************
* Desc:		Simplify one object like EXPAND
* Stack:	( symb --> symb' )
* Note:		object not necessarily symbolic
**********************************************************************
NULLNAME SIMP1ext
::					( obj )
  COMMONSIMP				( # lvar obj )
  ( back to symbolic )
  SWAPVAL2
;

**********************************************************************
* Desc:		Simplify one object like EXPAN
* Stack:	( symb --> symb' )
* Note:		object must be symb/real/cmplx
**********************************************************************
NULLNAME SYMEXPAN
::
ASSEMBLE
	IF	f39
RPL
  NOT40case 48SYMEXPAN
ASSEMBLE
	ENDIF	f39
RPL

  CKALG COMMONSIMP			( # lvar obj )
  DUP FXNDext SWAP CONSTANTPOLY?	( # lvar obj d n ? )
  case2drop SWAPVAL2
					( # lvar obj d n )
  DROP CONSTANTPOLY? SWAPDROP 		( # lvar obj ? )
  IT RISCH13				( # lvar obj )
  SWAPVAL2
;

LABEL COMMONSIMP
::
						( obj )
  QOBJext XROOTSIMP

  DUP LISTOPext LVARDext			( obj, lvar )

  ( * simplify all var * )
						( obj, lvar )
  OVER TWO DEPTHOBJext UNROT			( #, obj, lvar )
  DUP SLVARext					( #, obj, lvar, newlvar )

  ( * if vars were still simplified, just remove square roots * )
  ( * otherwise do the subsitution and recompute var without sqrt * )

  2DUP EQUAL 
  ITE
    :: DROP SQRT_IN? NOT?SEMI
       DROP LVARext SWAPDROP
    ;
    :: DUP4UNROLL MEVALext			( #, newlvar, obj )
       SWAP LVARext SWAPDROP
    ;

  ( * now make top level rational simplification * )
						( #, obj, newlvar )
  VXXLext					( # newlvar obj )
;

LABEL SWAPVAL2
::
  SWAP					( # obj_interne lvar )
  VAL2ext				( simplified_obj )
;

LABEL XROOTSIMP
* if obj contains some XROOT, calls ALGCASCOMPEVAL for simplifications
:: 				( obj )
  ' xXROOT Contains? 		( obj xXROOT flag )
  NOTcaseDROP 			( -> obj )
				( obj xXROOT )
  DROP ALGCASCOMPEVAL
;


**********************************************************************
* Desc:		Simplify variable
* Stack:	( ob --> ob' )
**********************************************************************
NULLNAME SIMPVAR
::	
  idntlamcase SIMPIDNT
  DUPTYPESYMB? NOT?SEMI
					( ..., var1 )
  ALGCASCOMPEVAL 
;

**********************************************************************
* Desc:		Look for all d1, d2, ... at the beginning of the name
*		of idnt to determine if idnt represent a derivative of
*		a user function. Stops if at a time the stripped idnt
*		is in the current directory.
*		Example 
*		'd2d1Y' returns { #2 } << >>
*		if 'd2d1Y' is not defined and 'd1Y' is defined as << >>
*		or { #2 #1 } 'Y' if d2d1Y d1Y and Y are not defined
* Stack:	( id --> der_list stripped_id )
**********************************************************************
NULLNAME ID>DERext
::
  ID>$ NULL{} SWAP
  BEGIN					( liste, $ )
     DUP palparse			( liste $ [ ob TRUE | $ # $ FALSE ] )
     ID>DERLP				( liste $ [ id FALSE | TRUE ] )
  WHILE					( liste $ )
     CDR$ ZEROSWAP			( liste, #, $ )
     BEGIN
     ::
       DUP CAR$ DUPNULL$? caseFALSE
					( liste, #, $, car$ )
       CHR># FORTYEIGHT OVER #>		( liste, #, $, car#, ? )
       caseFALSE
       FORTYEIGHT #- DUP TEN #<
     ;					( liste, #, $, [ id FALSE | # TRUE ] )
     WHILE
       ROT TEN #* #+ SWAP CDR$
     REPEAT
     DROP				( liste, #, $ )
     UNROT >TCOMP SWAP			( liste, $ )
  REPEAT
					( liste $ id )
  SWAPDROP				( liste id )
;

( [ ob TRUE | $ # $ FALSE ] -> [ id false  true ] )
LABEL ID>DERLP
::					( [ ob TRUE | $ # $ FALSE ] )
   NOTcase2drop :: $>ID FALSE ;		( id FALSE )

					( ob )
   DUPTYPEIDNT? NOTcaseFALSE 
   DUPSAFE@				( [ id ob TRUE | id FALSE ] )
   case DROPFALSE			( -> id FALSE )
					( id )
   OVER "d" ONE POS$ #1=		( id flag )
   NOTcaseFALSE				( -> id FALSE )
   DROPTRUE				( TRUE )
;

**********************************************************************
* Desc:		Evaluate idnt (looks recursively for its content if
*		defined). Does not error for circular definition,
*		but displays a warning.
* Stack:	( idnt --> ob )
**********************************************************************
NULLNAME SIMPIDNT
::	
  NULL{} DUPDUP 3LAMBIND
  RCL1IDNT
  ABND
  ( * check that ob is a valid symbolic element * )
  DUP CK[]NCK
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME RCLALLIDNT
::
  idntlamcase RCL1IDNT
  DUP LIDNText				( obj, lidnt )
  DUPNULL{}? caseDROP
  DUP INNERDUP #1-SWAP
  ZERO_DO
     get1 1GETLAMSWAP 
     RCL1IDNT 
     SWAP 1PUTLAM
  SWAPLOOP
  #1+ {}N 				( obj lidnt lnew_value )
  SWAP 2LAMBIND SYMBINCOMP replacelam1bylam2 METASYMBN ABND
;

**********************************************************************
* Desc:           Recursive content of an idnt
* Stack:          ( idnt/lam -> obj )/
* Note:           lam1 to 3 must be bound
**********************************************************************
NULLNAME RCL1IDNT
::
  DUP 3GETLAM ListPos                   ( idnt/lam # )
  DUP#0=csedrp RCL1IDNT1
                                        ( idnt/lam # )
  2GETLAM ROTDROPSWAP                   ( replace{} # )
  NTHCOMPDROP
;

LABEL RCL1IDNT1
::
  DUPSAFE@ NOT?SEMI
                                        ( idnt/lam, obj )
  1GETLAM ROT2DUP                       ( obj, liste, idntlam, liste, idnt )
  EQUALPOSCOMP                          ( obj, liste, idntlam, #pos )
  #0=case :: >HCOMP 1PUTLAM RCLALLIDNT ;
*  UNROT2DROP                            ( idnt/lam )
*  DUP DO>STR ": circular def." &$
*  EXITMSGSTO ERRJMP
  ( * warning: if circular def does not error, might cause infinite loop * )
  # 129 ERRORSTO ERRJMP
;


**********************************************************************
* Desc:
* Stack:	( inf sup fcn var --> integral(inf,sup,fcn,var) )
**********************************************************************
NULLNAME SIMPSYMBS
NULLNAME SYMINTEGRAL
::
ASSEMBLE
	IF	f39
RPL
  NOTEQWcase 48SYMINTEGRAL
ASSEMBLE
	ENDIF	f39
RPL

  EXACT? NOTcase INT4APPROX
  CLEANIDLAM 				( inf sup fcn id )
  ( * should look in fcn for ABS and split interval if needed * )
  4PICK APPROXCOMPEVAL 			( inf sup fcn id inf% )
  DUPTYPEREAL? NOTcasedrop INSYMINTEGRAL
  4PICK APPROXCOMPEVAL			( inf sup fcn id inf% sup% )
  DUPTYPEREAL? NOTcase2drop INSYMINTEGRAL
					( inf sup fcn id inf% sup% )
  2DUP %= case2drop :: 4DROP Z0 ;
  %> NOTcase SYMINTEGRAL<
  4ROLL UNROT SYMINTEGRAL< QNeg
;

LABEL SYMINTEGRAL<
::
  ZEROZERO RCLVX 3LAMBIND STOVX			( inf sup fcn )
  LVARXNXext UNROT2DROP			( inf sup fcn lvarx )
  INNERCOMP symintegral			( symb )			
  1GETABND STOVX
;

LABEL symintegral
* inf sup fcn varx1..varxn #n -> symb
::					( inf sup fcn meta-var )
  DUP#0=csedrp VXINSYMINTEGRAL
					( inf sup fcn meta-var )
  #1-SWAP				( inf sup fcn meta-var varx )
  DUPTYPESYMB? NOTcasedrop symintegral	( -> )
  					( inf sup fcn meta-var varx )
  DUP LASTCOMP EQ: xABS			( inf sup fcn meta-var varx ?abs )
  NOTcasedrop symintegral		( -> )
  OVER#2+UNROLL {}N 			( inf sup fcn varx lvar )
  OVER EXPR> 2DROP DUP SIGNE1ext	( inf sup fcn varx lvar arg signtab )
  7ROLL SIGNELEFT 6ROLL SIGNERIGHT	( fcn varx lvar arg signtab )
  ( * now for each sign in the table replace var by + or -arg of ABS * )
  Z0 OVER LENCOMP #2/ ZERO_DO (DO)	( fcn varx lvar arg signtab res )
    OVERINDEX@ #2* #1+ NTHCOMPDROP	( fcn varx lvar arg signtab res inf )
    3PICK INDEX@ #2* #3+ NTHCOMPDROP	( fcn varx lvar arg sgntb res inf sup )
    8PICK 8PICK 7PICK	( fcn varx lvar arg signtab res inf sup fcn varx arg )
    7PICK INDEX@ #2* #2+ NTHCOMPDROP
    ?ext OVER EQUALcase SIGNEERROR
    EQ: x- IT QNeg MEVALext	( fcn varx lvar arg signtab res inf sup fcn )
    7PICK INNERCOMP symintegral		( fcn varx lvar arg signtab res symb )
    x+ext				( fcn varx lvar arg signtab res )
  LOOP					( fcn varx lvar arg signtab res )
  6UNROLL 5DROP
;

LABEL VXINSYMINTEGRAL
::
  RCLVX INSYMINTEGRAL
;

LABEL INSYMINTEGRAL
* inf sup fcn vx -> symb
::					( inf sup fcn vx )
  DUPUNROT RISCHext			( inf sup vx primitive )
  4UNROLL PREVALext			( symb )
;

**********************************************************************
* Desc:		Simplification of user functions. Tests for derivative
*		of user functions.
* Stack:	( ob1..obn #n id/symb_id_;/seco/romptr --> id[] )
**********************************************************************
NULLNAME SIMPUSERFCN
::	

  ( fonction utilisateur )
  DUPTYPESYMB? IT CARCOMP
  IDNTLAM? NOTcase EVALUSERFCN
  DUP ID>DERext		( obj #nombres idnt list stripped_idnt )
  DUPSAFE@		( obj #n idnt list stripped [ ob true | false ] )
  NOTcase2drop SIMPUSERIDNT
			( ob1 .. obn #n idnt list stripped ob )
  DROP SIMPUSEREXIST
;

LABEL SIMPUSEREXIST
::
			( objets, #nombre, idnt, liste, function )
  ROTDROP		( objets, #nombre, liste, fonction )
  SWAP DUPNULL{}?
  casedrop EVALUSERFCN
			( objets, #nombre, fonction, liste )
  ZERO 4PICK 
  ZERO_DO 
     #1+DUP UNCOERCE
     DO>STR CHR_# >H$ $>ID
  SWAPLOOP
			( objets, #n, fonction, liste, v1, .., vn, #n )
  {}N DUP1LAMBIND	( objets, #n, fonction, liste, v )
  ROTSWAP		( objets, #n, liste, fonction, v )
  INNERCOMP get1	( objets, #n, liste, v1, .., vn, #n, fnct )
  EVALUSERFCN CKALG	( objets, #n, liste, fnct[v1, .., vn] )
  SWAPINCOMP get1 SWAP	( objets, #n, %var1, ..., %vark, fnct[v1, .., vn], #k )
  ZERO_DO (DO)
     1GETLAM
     ROT NTHELCOMP
     NOTcase INTERNALERR
			( ..., fnc, v% )
     DIRECTDERIV
  LOOP
			( objets, #n, fnc )
  OVER#2+UNROL {}N	( fnc, { o1 ... on } )
  1GETABND
  SWAP MEVALext
;


LABEL SIMPUSERIDNT
::				( obj1..objn #n idnt )
  SYMB? NOTcase SETNONEXTERR
  ONE SYMBN
  OBJ>R				( obj1..objn #n )
  reversym DUP OBJ>R ZEROSWAP 
  ZERO_DO (DO)			( objn..obj1 meta )
    get1 SYMBINCOMP top&	( objn..obj2 meta&obj1 )
  LOOP				( meta )
  R>OBJ	R>OBJ			( meta #n idnt )
  SWAPROT #2+ addt: xFCNAPPLY	( meta idnt #n xFCNAPPLY # )
  SYMBN
;

**********************************************************************
* Desc:		Evaluate a user function with stack checking
* Stack:	( V1..Vn #n fcn --> f[] )
**********************************************************************
NULLNAME EVALUSERFCN
::				( v1, ..., vn, #n, fcn )
*  DEPTH ROT#- 1LAMBIND		( v1, ..., vn, fcn )
*  EVAL				( fcn[v1,...,vn] )
*  1GETABND
*  DEPTH #= NOTcase SETSTACKERR
  fnResults TestSysFlag NOTcase EVALUSERFCN1
  fnResults ClrSysFlag EVALUSERFCN1
  fnResults SetSysFlag
;

LABEL EVALUSERFCN1
::
  CODEM
    GOSBVL =SAVPTR
    A=DAT1.A D1=A A=DAT1.A
    LC(5)DOROMP ?A#C.A -> .false %Not a romptr
    D1+5 A=DAT1.3
    LC(3)$DE ?A=C.X -> .true
    LC(3)$AB ?A=C.X -> .true
    *.false
    GOVLNG =GPPushFLoop
    *.true
    GOVLNG =GPPushTLoop
  ENDCODE
  NOTcase EVALUSERFCN2
  ROMPTR@ DROPSWAPDROP COLA EvalNoCK
 
*  DUPTYPEROMP? NOTcase EVALUSERFCN2
*  DUP ROMPTR>#				( ... romptr #libnum #libcmd )
*  DROP # DE #= NOTcase EVALUSERFCN2
*  ROMPTR@ DROPSWAPDROP EvalNoCK
;

LABEL EVALUSERFCN2
::
  DUPTYPECOL? case EVALUSERFCNCOL
  ONE SYMBN SWAP xFCNAPPLY
;

LABEL EVALUSERFCNCOL
::
  SWAPDROP EvalNoCK
;

**********************************************************************
* Desc:		Execute the WHERE operator
* Stack:	( ob list --> ob' )
**********************************************************************
NULLNAME SIMP|
::	
  INNERCOMP REPWHERE1
;

**********************************************************************
* Desc:		Calculate the LCM of the denominator of the
*		elements of the list. If input is not a list,
*		returns the denominator of the object.
* Stack:	( list --> ob )
**********************************************************************
NULLNAME DENOLCMext
::
  FXNDext SWAP
  CODE
		ST=1	sALG48
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A
		GOSBVL	=QSKIP
		GOSBVL	=GETPTR
		?ST=1	sALG48
		GOYES	deno=1
		LA(5)	denolcm!
		PC=(A)
deno=1		LA(5)	=Z1
		GOVLNG	=OverWrALoop
  ENDCODE
  ONE>POLY RMULText
;

* ( liste -> ppcm des denos ) 
LABEL denolcm!
::
  INNERCOMP METADENOLCM
;

**********************************************************************
* Desc:		Calculate LCm of the denominators of the elements
*		of Meta.
* Stack:	( Meta --> ob )
**********************************************************************
NULLNAME METADENOLCM
LABEL metadenolcm!
::
  DUP#0=csedrp Z1
  #1-SWAP FXNDext SWAPDROPSWAP	( o1, ..., on-1, res, #n-1 )
  DUP#0=csDROP
  ZERO_DO (DO)			( o1, ..., on-1, res )
      SWAPFXND SWAPDROP		( o1, ..., on-2, res, dn-1 )
      2DUP GCDext RDIVext RMULText
  LOOP
;

LABEL REWRITEASQ
::						( { q1...qn } )
  CKINNERCOMP EXTRACTCOMMONDENO {}POLY 		( d Q )
  SWAP ONE>POLY QDiv				( Q/d )
;

**********************************************************************
* Desc:		Simplified fraction (internal)
* Stack:	( ob2 ob1 --> ob1/ob2 )
**********************************************************************
NULLNAME SWPSIMPNDXF
CODE
=swpsimpndxf	GOSBVL	=swapml
=simpndxf	LA(5)	=simpndxfext
		PC=(A)
ENDCODE

**********************************************************************
* Desc:		Simplified fractin (internal)
* Stack:	( ob2 ob1 --> ob2/ob1 )
**********************************************************************
ASSEMBLE
=simpndxfext
RPL
NULLNAME SIMPNDXFext
::
  SIMPext NDXFext
;

**********************************************************************
* Desc:		Simplify two objects in internal representation.
*		Checks that o2 is not a complex or an irrquad
*		because decomposition of the corresponding fraction
*		with larg would -> Try to recover memory
*
* Stack:	( ob1 ob2 --> ob1' ob2' )
**********************************************************************
NULLNAME SIMPext
LABEL RCanno
CODE
simp		GOSBVL	=a=stkwc1
		?A#C	W
		GOYES	simpext
		GOVLNG	=Loop
simpext		LA(5)	simpext!
		PC=(A)
ENDCODE


LABEL simpext!
::
*  OVER DEPTHext DO>STR OVER DEPTHext DO>STR &$ DISPROW1 ( test )
  mult2conj!

* 2eme etape de la routine de simplification: o1 n'est pas un cmplx ni un irrq

  nogcdflag TestSysFlag case FLOATSIMP
  SIMPEXTOK

* o2 est-il negatif?

  POSITIFext				( o1, o2, o2<0? )
  ?SEMI
  SWAPRNEG SWAPRNEG
;

LABEL FLOATSIMP
::
  CK&DISPATCH1
  ereal 'QDivLOPAZ1
  ecmp 'QDivLOPAZ1
  EXT 'QDivLOPAZ1
  list FLOATSIMPL
  zint :: EXACT? ?SEMI Z2%% 'QDivLOPAZ1 ;
  any NOP
;

LABEL 'QDivLOPAZ1
:: 'QDivLOPA Z1 ;

LABEL FLOATSIMPL
::
  ( * divide num & deno by the deno's leading coeff * )
  DUP				( n d d )
  [BEGIN]
    DUPTYPELIST?
  [WHILE]
    CARCOMP
  [REPEAT]
				( n d leading_coeff )
  CK&DISPATCH1
  ereal 2QDivLOPA
  ecmp 2QDivLOPA
  EXT 2QDivLOPA
  any DROP
;

LABEL 2QDivLOPA
::
  ROTOVER 'QDivLOPA		( d leading_coeff n' )
  UNROT 'QDivLOPA		( n' d' )
;

LABEL mult2conj!
CODE

* 1ere etape: tant que o1 est un complexe/irrquad -> * conjugue 

mult2conj	GOSBVL	=SAVPTR
		B=0	A	B=compteur de profondeur
		C=DAT1	A
		CD0EX		D0=@stk1
		LC(5)	=DOLIST
mult2lp		A=DAT0	A
		?A#C	A
		GOYES	mult2col?
		B=B+1	A
		D0=D0+	5
		GOTO	mult2lp
mult2col?	LC(5)	=DOCOL
		?A#C	A
		GOYES	mult2ptrlp

* { .. { :: on teste si c'est le seul objet de la liste: SKIPOB puis des SEMI

		AD0EX
		R1=A.F	A	sauvegarde D0
		AD0EX
		A=B	A	
		R0=A.F	A	sauvegarde profondeur
		GOSBVL	=SKIPOB
		LC(5)	=SEMI
mult2lp2	B=B-1	A
		GOC	mult2yes
		A=DAT0	A
		D0=D0+	5
		?A=C	A
		GOYES	mult2lp2
mult2ptrlp	GOVLNG	=GETPTRLOOP
mult2yes	GOSBVL	=PUSH#
		C=R1.F	A
		D1=D1+	5
		DAT1=C	A	-> overwrite stk2 avec le complexe/irrquad
		D1=D1-	5
		CD1EX		-> teste si c'est un complexe/irrquad
		D1=D1+	5
		A=DAT1	A
		CD1EX
		LC(5)	=x<<
		?A=C	A
		GOYES	qmult2conj
cmult2conj	LA(5)	cmult2conj!
		PC=(A)
qmult2conj	LA(5)	qmult2conj!
		PC=(A)
ENDCODE


**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME SIMPEXTOK
::

* 3eme etape: simplification

  CODE
		GOSBVL	=rsubrcl		A, R1=stk1; R2=stk2
		A=R2
		C=A	W
		LCIRRQ
		?A#C	W
		GOYES	simpextnotirrq
simpextirrq	LA(5)	simpextirrq!
		PC=(A)
simpextnotirrq	A=R1
		LC(5)	=DOINT
		?A#C	A
		GOYES	simpext2l?
		A=R2
		LC(5)	=DOINT
		?A#C	A
		GOYES	simpextch?
		LA(5)	simpext2zint!
		PC=(A)
simpextch?	LC(5)	=DOCOL
		?A#C	A
		GOYES	simpextobjh
simpextch	LA(5)	simpextch!
		PC=(A)
simpextobjh	LA(5)	simpextobjh!
		PC=(A)
simpext2l?	A=R2.F	A
		LC(5)	=DOLIST
		?A#C	A
		GOYES	simpextany
		A=R1.F	A
		?A#C	A
		GOYES	simpextany
simpext2l	LA(5)	simpext2l!
		PC=(A)
simpextany	LA(5)	simpextany!
		PC=(A)
  ENDCODE

* test again for a complex/irrquad 

  mult2conj!
;

LABEL cmult2conj!
::				( o1, z2, # )
  DROPDUP RCONJext		( o1, z2, conj[z2] )
  DUPUNROT RMULText		( z1, conj[z2], |z2|^2 )
  UNROT RMULText SWAP
;

LABEL qmult2conj!
::				( o1 irrquad # )
  UNROTDUP OBJ>R QCONJext 	( # o1 irrq_barre )
  DUP QXNDext DROP GCDext	( # o1 irrq_barre gcd[a,b] )
  QDiv				( # o1 irrq_barre/gcd )
  DUP R>OBJ QMul		( # o1 irrq_barre/gcd norme/gcd )
  mult2conj!
  4PICK MAKEPROFOND		( #, o1, irrquad', {{ norme }} )
  4UNROLLROT			( {{ norme }}, o1, irrquad', # )
  MAKEPROFOND RMULText SWAP
  mult2conj!
;

**********************************************************************
* Desc:		Embedd ob in the given number of lists.
* Stack:	( ob # --> {{{...{o}...}}} )
**********************************************************************
NULLNAME MAKEPROFOND
::
  DUP#0=csDROP
  ZERO_DO (DO)
     ONE>POLY
  LOOP
;

LABEL simpext2zint!
:: 
  2DUP ZGCDext
  ROTOVER ZQUOText
  UNROT ZQUOText
;

LABEL simpextirrq!
::				( q, o )
  OVER INNERCOMP		( q, o, <<, a, b, c, >>, #5 )
  get1 6ROLL 6ROLL		( q, <<, c, >>, #5, o, a, b )
  ROTDUP 2OVER GCDext		( q, <<, c, >>, #5, a, b, o, o, gcd[a,b] )
  SWAP SIMPext			( q, <<, c, >>, #5, a, b, o, gcd', o' )
  SWAPDROPDUP NINE UNROLL	( q, o', <<, c, >>, #5, a, b, o, o' )
  OVER EQUALcasedro 7DROP	( q, o', <<, c, >>, #5, a, b, o, o' )
  RDIVext DUPUNROT		( q, o', <<, c, >>, #5, a, o/o', b, o/o' )
  RDIVext 6UNROLL		( q, o', <<, b', c, >>, #5, a, o/o' )
  RDIVext 5UNROLL		( q, o', <<, a', b', c, >>, #5 )
  ::POLY ROTDROPSWAP		( q', o' )
;

LABEL simpextch!
::
  2DUPSWAP			( c1, z2, z2, c1 )
  EVAL ZGCDext ZGCDext		( c1, z2, z )
  ROT CXRIext			( z2, z, r1, i1 )
  3PICK ZQUOText		( z2, z, r1, i1/z )
  SWAP3PICK ZQUOText IRXCext	( z2, z, r1/z+i*i1/z )
  UNROT ZQUOText
;

LABEL simpextobjh!
::				( obj, zint )
  SWAP PPZZ			( zint, obj', z' )
  ROT SIMPext			( obj', z'', zint' )
  UNROT RMULText SWAP
;

LABEL simpext2l!
::
  GCDHEUext			( a b c G' A'/G' B'/G' flag )
  ?SKIP SLOWSIMP2L		( a b c G' A'/G' B'/G' )
  2SWAP 2DROP			( a b A'/G' B'/G' )
  2SWAP
  mult2conj!
				( A'/G' B'/G' a' b' )
  ONE>POLY ROT RMULText		( A'/G' a' B/pgcd )
  UNROT ONE>POLY RMULText	( B/pgcd A/pgcd )
  SWAP
  NEWTRIMext
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME SLOWSIMP2L
::	
  Z1 DUP 2OVER			( a b c G' A'/G' B'/G' 1 1 A'/G' B'/G' )
  SLOWGCDext			( a b c G' A'/G' B'/G' newG' )
  SIMPGCDext			( a b c G' newA' newB' )
;

* simplifie o1 et o2 si GCDHEU ne s'applique pas
LABEL simpextany!
::	
  2DUP GCDext			( o1, o2, gcd )
  SIMPGCDext
;

**********************************************************************
* Desc:
* Stack:	( o1, o2, gcd -> o1 div gcd, o2 div gcd )
**********************************************************************
NULLNAME SIMPGCDext
::
  Z1 EQUALcasedro NOP
  ROTOVER NEWDIVext DROP FXNDext	( o2, gcd, o1.n, o1.d )
  2SWAP NEWDIVext DROP FXNDext		( o1.n, o1.d, o2.n, o2.d )
  4ROLL RMULText			( o2.d*o1.n, o1.d, o2.n )
  UNROT RMULText
;

**********************************************************************
* Desc:		Calculate g = gcd(a,b) and a''=a/g and b''=b/g
* Stack:	( a b --> g a'' b'' )
**********************************************************************
NULLNAME SIMP3ext
::
  CK&DISPATCH1
  2LIST SIMP3LISText
  any   :: 2DUP GCDext			( a, b, g )
           ROTOVER RDIVext		( b, g, a/g )
           ROT 3PICK RDIVext		( g, a/g, b/g )
        ;
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME SIMP3LISText
::					( A B )

* g=pgcd[A,B], A''=A/g, B''=B/g

  2DUP GCDHEUext			( A B a b c pr[pgcd] A'/G' B'/G' fl )
  NOTcase SIMP3LSTSLOW
					( A B a b c pr[pgcd] A'/G' B'/G' )
  4ROLL ONE>POLY 4ROLL RMULText		( A B a b A'/G' B'/G g )
  4ROLL ONE>POLY ROT RMULText		( A B a A'/G' g B'' )
  4ROLL ONE>POLY 4ROLL RMULText		( A B g B'' A'' )
  4ROLLDROP 4ROLLDROP			( g B'' A'' )
  SWAP					( g A'' B'' )
;

**********************************************************************
* Desc:
* Stack:
* Note:		QUOText is protected because the gcd might not
*		divide A and B since irrquad gcd is ill-defined
**********************************************************************
NULLNAME SIMP3LSTSLOW
::				( A, B, a, b, c, pr[pgcd], A'/G', B'/G' )
  SLOWGCDext			( A, B, a, b, pgcd )
  UNROT2DROP			( A, B, gcd )
  DUPUNROT DIVEVAL QUOText	( A, gcd, B'' )
  ROT 3PICK			( gcd, B'', A, gcd )
  DIVEVAL QUOText		( gcd, B'', A'' )
  SWAP				( gcd, A'', B'' )
;

* evaluate next runstream object with a protection for divisions (e.g. QUOText)
* return Undefined result if an error occurs
LABEL DIVEVAL
::
  'R
  divflag TestSysFlag case EVAL
  divflag SetSysFlag EVAL
  divflag ClrSysFlag
;

**********************************************************************
* Desc:		Calculate the GCD of all the elements in the list.
* Stack:	( {} --> {} ob )
* Note:		The algorithm is far from optimal.
**********************************************************************
NULLNAME LPGCDext
CODE
lpgcd		C=DAT1	A
		CD1EX
		A=DAT1	A
		CD1EX
		LC(5)	=DOLIST
		?A=C	A
		GOYES	lpgcdtest1
		LC(5)	=DOMATRIX
		?A=C	A
		GOYES	lpgcdtest1
		LA(5)	=DUP
		PC=(A)
lpgcdtest1	GOSBVL	=SAVPTR
		C=DAT1	A
		D0=C
		D0=D0+	5
lpgcdloop	A=DAT0	W
		LC(5)	=SEMI
		?A=C	A
		GOYES	lpgcdext
		C=A	W
		LCZ1
		?A=C	A
		GOYES	lpgcdzint
lpgcdskip	GOSBVL	=SKIPOB
		GOTO	lpgcdloop
lpgcdzint	?A=C	W
		GOYES	lpgcdpush1
		P=	11
		SETDEC
		C=-C-1	P
		SETHEX
		P=	0
		?A=C	W
		GOYES	lpgcdpush1
		D0=D0+	5
		C=DAT0	A
		AD0EX
		A=A+C	A
		AD0EX
		GOTO	lpgcdloop
lpgcdpush1	GOSBVL	=GETPTR
=zint1		LA(5)	=Z1
		PC=(A)
lpgcdext	GOSBVL	=GETPTR
		LA(5)	=lpgcdext!
		PC=(A)
ENDCODE

*  ( LAM1: liste -> liste, lpgcd )
LABEL lpgcdext!
::
  nogcdflag TestSysFlag case Z1
  DUPINCOMP
  DUP#0=csedrp Z1
  lpgcdexec
  Z0 EQUALcasedro Z1
;

LABEL lpgcdexec
::
  DUP#1= caseDROP
  ONE_DO (DO)
    CODE
		C=DAT1	A
		CD1EX
		A=DAT1	12
		CD1EX
		C=A	W
		LCZ1
		?A=C	W
		GOYES	lpgcdis1
lpgcd2		GOLONG	gcd
lpgcdis1	LA(5)	=lpgcdexit!
		PC=(A)
    ENDCODE
  LOOP
;

* ( o1, ..., ok -> 1 )
LABEL lpgcdexit!
::
  ISTOP-INDEX #1+NDROP
  ExitAtLOOP
  Z1
;

**********************************************************************
* Desc:		Euclidean algorithm for polynomial GCD.
*		Used if A or B contains irrquads. c is the GCD
*		of the contents of the original polynomials
*		returned after failure of GCDHEUext.
* Stack:	( c 1 A B --> c* gcd(A,B) )
**********************************************************************
NULLNAME SLOWGCDext
::
  ROTDROP				( c A B )
  nogcdflag TestSysFlag case2DROP	( return c if no simpl. allowed -> c )
  { "GCDHeu failure" DISPROW2 } Sys1IT
  DUPLENCOMP				( c l1 l2 #l2 )
  3PICK LENCOMP				( c l1 l2 #l2 #l1 )
  2DUP #MIN 5UNROLL			( c #min l1 l2 #l2 #l1 )
  #>ITE SWAPROT ROT		( c l1, l2 avec l1 plus long que l2, #min )
  #1=case
  :: &COMP LPGCDext SWAPDROP ONE>POLY RMULText ;
					( c l1 l2 )
  LPGCDext DUP4UNROLL DIVOBJext		( c pgcd[l2]/autres_var l1 l2' )
  SWAP LPGCDext				( c pgcd[l2]/autr_var l2' l1 pgcd[l1] )
  DUP4UNROLL DIVOBJext			( c pgcd[l2] pgcd[l1] l2' l1' )
  2SWAP GCDext 4ROLL RMULText
  ONE>POLY UNROTSWAP

  ( pgcd[l1,l2] par rapport aux autres var, l1', l2' )

  Z1 DUP 2SWAP
  [BEGIN]				( lpgcd g h l1 l2 )
     DUPCKLEN{} #1<>
  [WHILE]
     SLOWGCDLP
  [REPEAT]
					( lpgcd  g' h' l1 l2_avec_degre=0 )
  2SWAP 2DROP				( lpgcd  l1 l2_avec_degre=0 )
  CKCARCOMP Z0 EQUAL			( lpgcd l1 l2=0? )
  NOTcaseDROP				( -> lpgcd )
					( lpgcd l1 )
  LPGCDext DIVOBJext RMULText
;

LABEL SLOWGCDLP
* sub-resultant algorithm innerloop
* g h l1 l2 -> g' h' l1' l2'
::				( g h l1 l2 )
  ( * multiplies numerator by an invertible element * )

  DUPDUP CARCOMP SWAP LENCOMP	( g h l1 l2 a2.1 #l2 )
  4PICK LENCOMP SWAP#-		( g h l1 l2 a2.1 #l1-#l2 )
  DUPUNROT #1+ RP# ONE>POLY	( g h l1 l2 #l1-l2 {a2.1^} )
  4ROLL QMul			( g h l2 #l1-l2 l1' )

  ( division de l1' par l2 )

  Z0 SWAP4PICK DIV2LISText	( g h l2 #l1-l2 l1divl2 l1modl2 )
  SWAPDROP			( g h l2 #l1-l2 l1modl2 )
	
* commented next two lines for the resultant algorithm
*  DUPTYPELIST? ?SKIP ONE{}POLY	
*  DUPLENCOMP #1=case SWAPDROP
* end change
	
				( g h l2 #l1-l2 l1modl2 )
  5ROLL 5PICK			( h l2 #l1-l2 l1modl2 g h )
  4PICK RP#			( h l2 #l1-l2 l1modl2 g h^# )
  DUP4UNROLL RMULText DIVOBJext	( h l2 #l1-l2 h^# l1modl2/[g*h^#] )
  UNROT 4PICK CARCOMP		( h l2 l1modl2/[g*h^#] #l1-l2 h^# g' )
  DUP 7UNROLL ROT RP#		( g' h l2 l1modl2/[g*h^#] h^# g'^# )
  5ROLL RMULText SWAP RDIVext	( g' l2 l1modl2/[g*h^#] h^1-#*g'^# )
  UNROT				( g' h' l1'=l2 l2'=l1modl2 )
;
	
**********************************************************************
* Desc:		Generic internal GCD
* Stack:	( ob2 ob1 --> gcd )
* Note:		( LAM2: GCDext obj1, obj2 -> pgcd )
**********************************************************************
NULLNAME QGcd
NULLNAME GCDext
::
  nogcdflag TestSysFlag case2drop Z1
  CODE
=gcd		GOSBVL	=rsubrcl
* si stk1 ou stk2=1 ou 0, le pgcd est 1 ou l'autre niveau
		C=A	W
		LCZ0
		?A=C	W
		GOYES	todroploop
gcdext1		LCZ1
		?A#C	W
		GOYES	gcdext2
swapdroploop	GOSBVL	=swapml
todroploop	GOVLNG	=DropLoop
* test avec -1, desactive a cause du signe du pgcd
*gcdext2	P=	11
*		SETDEC
*		C=-C-1	P
*		SETHEX
*		P=	0
* test avec signe de C change, car pgcd[obj,-1]=obj
*		?A=C	A
*		GOYES	swapdroploop
gcdext2		A=R2
		C=A	W
		LCZ0
		?A=C	W
		GOYES	swapdroploop
		LCZ1
		?A=C	W
		GOYES	todroploop
* test avec -1
*		P=	11
*		SETDEC
*		C=-C-1	P
*		SETHEX
*		P=	0
*		?A=C	W
*		GOYES	todroploop
* on teste si l'un des niveaux est SYMB -> 1
		LC(5)	=DOSYMB		A=stk2
		?A#C	A
		GOYES	gcdext3
todrop2zint1	GOTO	drop2zint1
gcdext3		A=R1.F	A
		?A=C	A
		GOYES	todrop2zint1
* on teste entier/cmplx, cmplx/entier, entier/entier
		LC(5)	=DOINT		A=stk1
		?A#C	A		
		GOYES	gcdstk1notzint
		A=R2.F	A
		?A#C	A
		GOYES	gcdstk1zint
zgcd		A=PC
		LC(5)	(zgcdext!)-(*)
		GOTO	gcdextbranch
gcdstk1notzint	A=R2.F	A
		?A#C	A
		GOYES	gcdnozint	
		GOTO	gcdstk2zint	stk2 entier
* 1 niveau est entier, on le met au niveau 2, on teste niv1: complex/irrq
gcdstk1zint	GOSBVL	=swapml
		A=R2
		AR1EX
		AR2EX
gcdstk2zint	A=R1
		C=A	W
		LCIRRQ
		?A#C	W
		GOYES	gcdzintnotirrq
gcdobjirrq	A=PC
		LC(5)	(gcdobjirrq!)-(*)
gcdextbranch	A=A+C	A
		PC=(A)
gcdzintnotirrq	?A=C	A
		GOYES	cgcdext
		GOTO	gcdnoirrq
cgcdext		A=PC
		LC(5)	(cgcdext!)-(*)
		GOTO	gcdextbranch
* niveau 1 et 2 ne sont pas des zint, chercher des irrq ou des listes
* test obj/irrq, irrq/obj, irrq/irrq, cmplx/cmplx
gcdnozint	A=R1
		C=A	W
		LCIRRQ
		?A#C	W
		GOYES	gcdstk1notirrq
		A=R2
		C=A	W
		LCIRRQ
		?A#C	W
		GOYES	gcdobjirrq
gcd2irrq	A=PC
		LC(5)	(gcd2irrq!)-(*)
		GOTO	gcdextbranch
gcdstk1notirrq	A=R2
		C=A	W
		LCIRRQ
		?A#C	W
		GOYES	gcd2cmp?
		GOSBVL	=swapml
togcdobjirrq	GOTO	gcdobjirrq
gcd2cmp?	?A#C	A
		GOYES	gcdnoirrq
		A=R1.F	A
		?A=C	A
		GOYES	cgcdext
* pas d'irrquad et pas cmplx/cmplx, on cherche 1 liste
gcdnoirrq	A=R2.F	A
		LC(5)	=DOLIST
		?A#C	A
		GOYES	gcdobjlist?
		A=R1.F	A
		?A#C	A
		GOYES	gcdlistobj
gcd2list	A=PC
		LC(5)	(gcd2list!)-(*)
		GOTO	gcdextbranch
gcdobjlist	GOSBVL	=swapml
gcdlistobj	GOVLNG	=ppz
gcdobjlist?	A=R1.F	A
		?A=C	A
		GOYES	gcdobjlist
* type non reconnu pour PGCD -> on renvoie 1
drop2zint1	D1=D1+	10
		D=D+1	A
		D=D+1	A
		GOLONG	zint1
  ENDCODE
  POSITIFext ?SEMI RNEGext
;

* ( LAM3: list1, list2 -> pgcd )
LABEL gcd2list!
::				( l1, l2 )
  2DUP EQUAL caseDROP
  GCDHEUext			( a, b, c, G', A'/G', B'/G', [ FALSE | TRUE ] )
  6ROLL DROP 6ROLL DROP		( c, G', A'/G', B'/G', [ FALSE | TRUE ] )
  NOTcase SLOWGCDext
  2DROP				( c, G' )
  SWAP ONE>POLY
  RMULText NEWTRIMext
  { "GCDHeu success" DISPROW2 }
  Sys1IT
;

LABEL gcd2irrq!
::
  2DUP EQUAL caseDROP
  PREPARext			( a, b, c, a', b', c' )
  num#1=case 
  ::
    DROP RADDext		( a, b, c, a'' )
* unsafe NDXQext?
    4UNROLL NDXQext GCDext
  ; 
  4PICK EQUALcasedro GCDSAMEIRRQ
			( a, b, c, a', b', c' avec c et c' premiers entre eux )
  DROP				( a b c a' b' )
  GCDext 4UNROLL		( g a b c )
  NDXQext GCDext		( g' )
;

LABEL GCDSAMEIRRQ
::				( a b c a' b' )
  ( * if ab'-a'b=0 return gcd[a,b]+gcd[a',b']*sqrt[c] * )
  OVER5PICK QMul		( a b c a' b' a'*b )
  6PICK 3PICK QMul QSub		( a b c a' b' a'*b-a*b' )
  j#0=case GCDSAMEIRRQ0
				( a b c a' b' )
  4ROLL GCDext			( a c a' b'' )
  ROT				( a a' b'' c )
  CODE			-> jump to label gcdobjirrq1!
	A=D0
	LC(5)	(gcdobjirrq1!)-(gcdobjirrq2!)
	GOVLNG	=D0=A+C_lp
  ENDCODE
  LABEL gcdobjirrq2!
;

LABEL GCDSAMEIRRQ0
::				( a b c a' b' )
  SWAPDROP			( a b c b' )
  3PICK GCDext			( a b c b" )
  ROTOVER QDiv			( a c b" b/b" )
  4ROLLSWAP QDiv		( c b" a/[b/b"] )
  SWAPROT NDXQext		( a"+b"*sqrt[c] )
;

* c'est le pgcd de pgcd[a,a'] et de b*sqrt[c]
LABEL gcdobjirrq!
::
  QXNDext			( a', a, b, c )
  LABEL gcdobjirrq1!
  2SWAP GCDext			( b, c, d=pgcd[a,a'] )
  ROTOVER GCDext		( c, d, pgcd[b,d] )
  DUPUNROT RDIVext		( c, pgcd[b,d], d/pgcd[b,d] )
  3PICK GCDext			( c, pgcd[b,d], pgcd[c,d/pgcd[b,d]] )
  ( * must insure that this pgcd is real * )
  DUP RREext SWAP RIMext GCDext	( c, pgcd[b,d], pgcd[c,d/pgcd[b,d]] )

  Z1 EQUALcasedro SWAPDROP	( -> c )
				( c, pgcd[b,d], pgcd[c,d/pgcd[b,d]] )
  ( * test positivity of c*pgcd to modify sign of pgcd if necessary * )
  ROTOVER QMul 			( pgcd[b,d], pgcd[c,d/pgcd[b,d]], c*pgcd )
  POSITIFext SWAPDROP		( pgcd[b,d], pgcd[c,d/pgcd[b,d]], flag )
  ?SKIP QNeg			( pgcd[b,d], pgcd[c,d/pgcd[b,d]] )
  Z0 UNROT NDXQext
;

**********************************************************************
* Desc:		GCD for Gauss integers
* Stack:	( C2 C1 --> C )
**********************************************************************
NULLNAME CGCDext
LABEL cgcdext!
::					( a, b )
  [BEGIN]
      DUP Z0 EQUAL
  NOT_WHILE
      CMODext
  [REPEAT]
					( b, r=0 )
  DROP
  DUPTYPEZ? case ZAbs
  INCOMPDROP				( im, re )
  Z0 EQUALcasedro ZAbs
					( im, re )
  DupZIsNeg? SWAP ZAbs SWAPROT		( re, re<0, im )
  DupZIsNeg? SWAP ZAbs UNROT		( re, im, re<0?, im<0? )
  XOR ?SKIP SWAP IRXCext
;

**********************************************************************
* Desc:
* Stack:	( C2 C1 --> C1 C2_mod_C1 )
**********************************************************************
NULLNAME CMODext
::					( a, b )
  DUP RCONJext				( a, b, conj )
  2DUP					( a, b, conj, b, conj )
  RMULText				( a, b, conj, |b|^2 )
  4ROLLROT				( b, |b|^2, a, conj )
  RMULText OVER				( b, |b|^2, a*conj, |b|^2 )
  QMODSYMext				( a'=b, |b|^2, a*conj mod |b|^2 )
  3PICK RMULText SWAP ZQUOText		( a'=b, b*[a*conj mod |b|^2]/|b|^2 )
;

**********************************************************************
* Desc:		Integer GCD
* Stack:	( Z2 Z1 --> Z )
**********************************************************************
NULLNAME ZGCDext
NULLNAME ZGcd
LABEL zgcdext!
:: 
  CODE
* Desc:		FastGcd? tries to calculate GCD[Z1,Z2] in registers
*		If success then pushes GCD & TRUE, else FALSE

=FastGcd?	GOSBVL	=SAVPTR
		GOSBVL	=SHORTPOPZ?	A[W]C[S]=Z2
		GOC	fastgcdfail	Not short integer - fail
		?A=0	W
		GOYES	fastgcddrop1	GCD[any,0] = any
		B=A	W		B[W] = N2
		GOSBVL	=SHORTPOPZ?	A[W]C[S]=Z1
		GOC	fastgcdfail
		?A=0	W
		GOYES	fastgcddrop2	GCD[0,any]=any
		C=B	W

* Suspect following might be better:
*		GOSUBL	SHORTGCD

		SETDEC
-		GOSBVL	=BDIV		B[W]=remainder	C[W]=divisor
		A=C	W		A[W]=divisor
		C=B	W		C[W]=remainder
		?C#0	W
		GOYES	-
		GOSBVL	=PushN		Push GCD	(HEX mode)
		A=DAT1	A		Overwrite 2nd stack level with Z
		D1=D1+	10
		D=D+1	A
		GONC	fastgcdtrue

fastgcdfail	GOVLNG	=GPPushFLoop	( Z1 Z2 --> Z1 Z2 FALSE )

fastgcddrop1	GOVLNG	=GPOverWrTLp	( Z1 0 --> Z1 TRUE )

fastgcddrop2	GOSBVL	=GETPTR		( 0 Z2 --> Z2 TRUE )
		A=DAT1	A
		D1=D1+	5
fastgcdtrue	DAT1=A	A
		D1=D1-	5
		GOVLNG	=OverWrTLoop

  ENDCODE
  ?SEMI					( Done already )

* Atleast one integer too big for a register. Use slow division
* until both integers fit in

  ZAbs SWAP ZAbs
  BEGIN
*    DupQIsZero? NOT
     DUP Z0 EQUALNOT
  WHILE				( A B )
     SWAPOVER ZDIVext		( B A/B AmodB )
     SWAPDROP			( B AmodB )
     CODE
	GOLONG	FastGcd?
     ENDCODE
     case :: 2RDROP ;		( FastGcd finished )
  REPEAT
  DROP
;


**********************************************************************
* Desc:		Transcendental simplifications
* Stack:	( symb -> symb )
* Note:		convert only sqrt ^ and XROOT to EXP/LN
*		LN are returned as -1/INV[-LN[]] for use by SERIES
**********************************************************************
NULLNAME TSIMP2ext
::
  invlnflag SetSysFlag
  SERIESEXPLN TSIMP3ext
  invlnflag ClrSysFlag
;

**********************************************************************
* Desc:		Transcendental simplifications
* Stack:	( symb -> symb )
* Note:		Convert transc. functions to EXP and LN
**********************************************************************
NULLNAME TSIMPext
::
  SETEXACT EXPLNext
  TSIMP3ext
;

**********************************************************************
* Desc:
* Stack:	( symb -> symb )
* Note:		on decompose la liste des variables en ln, exp et autres
*		les arguments des ln sont simplifies recursivement, puis on
*		fait une decomposition sans carre avant de rassembler s'il y a
*		des pgcd non triviaux
**********************************************************************
NULLNAME TSIMP3ext
::				( symb )
* 2eme etape: la simplification proprement dite

  stepflag TestSysFlag
  stepflag ClrSysFlag
  TWENTYSEVEN TestSysFlag
  TWENTYSEVEN ClrSysFlag	( symb, ?display, ?FS )

  ROT				( ?display, ?FS, symb )


  TSIMP1			( ?display, ?FS, symb )
  TWENTYSEVEN ROT ITE SetSysFlag ClrSysFlag 
  stepflag ROT ITE SetSysFlag ClrSysFlag
;

	
LABEL TSIMP1
* main transcendental simplification code
( obj -> obj )
::	
  LVARext		( obj lvar )
  DUPNULL{}? caseDROP
  INNERCOMP
  NULL{} DUPDUP 4ROLL	( obj v1..vn largexp largln lautre #n )
  ZERO_DO (DO)		( obj v1..vk largexp largln lautre )
    4ROLL TSIMPVAR
    LOOP
  UNROT			( obj, lautre, largexp, largln )
  2DUP
  nonrecurflag TestSysFlag ?SKIP TSIMP1REC

			( obj lautre largexp largln largexp largln )
  TSIMPLN		( obj, lautre,  largexp, largln, largexp, lln simpl )
  TSIMPEXP		( obj lautre largexp largln llns lexps )


  ( * creation de la liste pour modifier dans l'objet de depart )
  &COMP UNROT		( obj lautre llns+lexps largexp largln  )
  ' TSIMPxLN
  1LAMBIND LISTSECOext ABND
  SWAP' TSIMPxEXP
			
  1LAMBIND LISTSECOext ABND	( obj lautre llns+lexps lln largexp )
  &COMP				( obj lautre llns+lexps lln+lexp )
  ROT &COMP SWAP		( obj lln+lexp+lautres llns+lexps )

  MEVALext		( obj )
;

LABEL TSIMPxLN
:: SYMBINCOMP addt: xLN SYMBN ;

LABEL TSIMPxEXP
:: SYMBINCOMP addt: xEXP SYMBN ;

LABEL TSIMP1REC
:: TSIMP1 SWAP TSIMP1 SWAP ;


LABEL TSIMPVAR
::	
  DUPTYPESYMB? NOTcase >HCOMP
				( obj v1..largexp largln lautre symb )
  DUPINCOMP #1-SWAP
  OVER#2+UNROL CKSYMBN		( obj v1..lexp lln lautre symb op arg )
  TSIMPVARLST
  ROTSWAP EQLookup
  NOTcase2drop >HCOMP
  ROTDROP			( obj v1..lexp lln lautre arg prg )
  EVAL
;

LABEL TSIMPVARLST
:: invlnflag TestSysFlag case { xEXP TSIMPVAREXP }
   { xEXP TSIMPVAREXP xLN >HCOMPstk3 }
;

LABEL TSIMPVAREXP
:: 4ROLLSWAP >HCOMP UNROT ;

LABEL >HCOMPstk3
:: ROTSWAP >HCOMP SWAP ;

LABEL TSIMPLN
* liste of ln -> liste of simplified ln
:: 				( l-arg_ln )
  DUPNULL{}? ?SEMI		
  LVARext VXXLext		( lvar l-arg_ln-l )

* square-free factorization of polynomials and full factorization of integers

  INNERDUP #1-SWAP 
  ZERO_DO			( lvar arg_ln1, ..., arg_lnn,  #n-1 )
    get1 FXNDext 
    MSQFF %-1 METAMM2
    get1 MSQFF top& 
    METACOMPRIM {}N
    SWAPLOOP			( lvar arg_ln1, ..., llnn, #n-1 )

* compute gcd of obj list by list

  SQFF2ext			( lvar meta-arg_ln )

* now compute the ln

  #1+DUP #1-SWAP
  ZERO_DO			( lln1, ..., llnn, #n-1 )
    get1 LISTLN			( lln2, ..., llln, #n-1, ln1 )
  SWAPLOOP
  #1+ {}N			( ob lautre lrgexp largln largexp lln-v lln-l )
  ONE UNROTSWAP VAL2ext		( ob lautre lrgexp largln largexp lln )
;

**********************************************************************
* Desc:		Logarithm of an object whose internal factorization is given.
*		See sqff.s for the format of the list:
*		{ fact1 m1 .. factn mn }
* Stack:	( {} --> ob )
**********************************************************************
LABEL LISTLN
::
  DUPLENCOMP DUP#0= case2drop Z0
					( {} # )
  SWAP >R Z0ONE ROT #2/ 		( meta # )
  ZERO_DO (DO)
    RSWAP 'R'R RSWAP			( meta f % )
    R>Z ONE ROT LNATANMETA		( meta % #1 meta )
    MetaMul MetaAdd			( meta )
    LOOP
  CKSYMBN
;


LABEL TSIMPEXP{}
::		( obj lautre largexp largln {largexp lln-s} ?pi lvar lvar )
  INNERDUP #1-SWAP
  ZERO_DO (DO)
    get1 DUPTYPESYMB? TSIMPEXP{}LP
  LOOP
  #1+ {}N
;

LABEL TSIMPEXP{}LP
::
  NOTcasedrop #1- 
  DUPDUP LENCOMP NTHCOMPDROP EQ: xLN 
  NOTcasedrop #1-
  SWAP 
;

LABEL TSIMPEXP
:: 
  SWAP DUPNULL{}? casedrop 4PICK
		      ( obj lautre largexp_orig largln_orig lln-s largexp )
  5PICK
  ' TSIMPxLN
  1LAMBIND LISTSECOext ABND
  &COMP COMPRIMext
	      ( obj lautre largexp_orig largln_orig lln-s largexp+ln[lautres] )

  SWAP TWO{}N		( obj lautre largexp largln {largexp lln-s} )
  LVARext		( obj lautre  largexp largln {largexp lln-s} lvar )

      ( on met pi en tete de liste si necessaire )

  pi
  2DUP EQUALPOSCOMP #0<> ( COMPLEX? AND )

      ( obj lautre largexp largln {largexp lln-s} lvar pi ?a_rajouter )

  DUP4UNROLL
  ITE
        :: >HCOMP NEXTPext CARCOMP >TCOMP ;
        DROP

  ( obj lautre largexp largln {largexp lln-s} ?pi_a_rajouter lvar )
  ( * compute lvar that contain LN * )
  DUP DUPNULL{}? ?SKIP TSIMPEXP{}


  ( obj lautre largexp largln {largexp l-lns} ?pi lvar lvarLN )

  ( on rajoute i* pi a lvarLN )

  ROT IT :: ipi >TCOMP ;

  ( obj lautre largexp largln {largexp l-lns} lvar lvarLN )

  DUP' xEXPext 1LAMBIND LISTSECOext ABND 

* commented out next line since VXXLext might return TRANSCERROR
* NULL{} VXXLext SWAPDROP

  ( obj lautre largexp largln {largexp l-lns} lvar lvarLN lnargs )
  4ROLL INCOMPDROP 5UNROLL
  ( obj lautre largexp largln l-lns lvar lvarLN lnargs largexp )

  ( teste si les arguments des exp sont CL \`a coef. rationnels )
  ( des autres arguments des exp ou des arguments des LN )

  ROT TWO{}N ROT 
 
  ( obj lautre largexp largln l-lns =... lnargs {largexp lvarLN} lvar )

  VXXLext
  INCOMPDROP		( ..., arg des LN, lvar, largexp, lvarLN )

  ( on rajoute Z1 a lvarLN pour les constantes )

  Z1 >TCOMP 4ROLL	( ..., lvar, largexp, lvarLN, arg des LN )

  ( repercussion sur arg des LN )

  base_ln >TCOMP ROT	( ..., lvar, lvarLN, arg des LN, largexp )
  NULL{} SWAP
  [BEGIN]		( ..., lvar, lvarLN, argLN, lexps, largexp )
     DUPNULL{}?
  NOT_WHILE
  ( * determines if last element of largexp is a rational C.L. of lvarLN * )
    TSIMPEXPLP
  [REPEAT]

  ( obj lautre largexp largln l-lns lvar lvarLN argLN lexps largexp )

  DROPOVER

  ( obj lautre largexp largln l-lns lvar lvarLN argLN lexps argLN )

  ( on reajuste toutes les longueurs des listes de lexps )

  LENCOMP #1- ' LRDMext 2Ob>Seco

  ( obj lautre largexp largln l-lns lvar lvarLN argLN lexps prg )

  1LAMBIND LISTSECOext ABND

  ( obj lautre largexp largln l-lns lvar lvarLN argLN lexps )

  ( on calcule les PGCD des d\'enominateurs verticalement )
  ( lexps -> lexps*pgcd, lpgcd colonne par colonne )

  METATRAN []N DUP' denolcm!

  1LAMBIND LISTSECOext ABND		( lexps, lpgcd )
  DUPUNROT ' VPMULT VBINARYOP
  METATRAN {}N				( lpgcd, lexps*lpgcd )
  4UNROLL

  ( obj, lautre, largexp, largln, l-lns, lvar, lexps, lvarLN, argLN, lpgcd ) 

  ' TSIMPXROOT VBINARYOP
    ( obj, lautre, largexp, largln, l-lns, lvar, lexps, lvarLN, argLN )
  SWAPDROP ONESWAP 4ROLL VAL2ext
  ( obj, lautre, largexp, largln, l-lns, lexps, argLN )
  ' TSIMPPI 2Ob>Seco 1LAMBIND 
  ( obj, lautre, largexp, largln, l-lns, lexps )
  LISTSECOext ABND
  ( obj, lautre, largexp, largln, l-lns, lexps )
;

LABEL TSIMPXROOT
::	( x, y -> x^[1/y] )
  num#1=case DROP
  invlnflag TestSysFlag case TSIMPXROOTN
  Z2 EQUALcasedro XSQRext
  OVER DUPTYPESYMB?
  ITE IS_EXP? DROPFALSE					( x, y, flag )
  case TSIMPXREXP
							( x y )
  SYMBINCOMP get1 SYMBINCOMP 				( y-m x-m )
*  top& addt: xXROOT SYMBN
  addtLN psh MetaDiv addtEXP SYMBN
;

LABEL TSIMPXROOTN
::
  SWAP SYMBINCOMP addtLN2 CKSYMBN 
  SWAP RDIVext xEXPext 
;

LABEL IS_EXP?
:: DUPLENCOMP NTHCOMPDROP EQ: xEXP ;

LABEL TSIMPXREXP
::						( x, y )
  FXNDext SWAPROT				( d, n, x )
  INNERCOMP #1-SWAP DROP SYMBN SWAP		( d, ln[x], n )
  RDIVext SIMP1ext xEXPext			( d, x' )
  SWAP x^ext
;


LABEL TSIMPPI
::	( x, y -> pi y_i^x_i )
  INNERCOMP #1+ROLL INNERCOMP
  DUP#1= casedrop TSIMP^
  get1				( y1, ..., yn-1, x1, ..., xn, #n, yn )
  ROT TSIMP^			( y1, ..., yn-1, x1, ..., xn-1, #n, yn^xn )
  OVER ONE_DO (DO)
    OVER #2+PICK
    4ROLL TSIMP^ x*ext
  LOOP
				( y1, ..., yn-1, #n, X.Y )
  OVER #1+UNROLL #1- NDROP
;

LABEL TSIMP^
( x, y -> x^y )
::
  Z0 EQUALcasedro DROPZ1
  num#1=case DROP
  OBJ>R SYMBINCOMP R>OBJ SWAP#1+ addt: x^ SYMBN
;


LABEL TSIMPEXPLP
::
  INNERCOMP #1-SWAP
  OVER#2+UNROL {}N
  5UNROLL		( ...lvar largexp lvarLN lnargs lexps argexp )
  4ROLLSWAP		( ...lvar largexp argLN lexps lvarLN argexp )
  ( * look if argexp is a linear combination of elements of lvarLN * )
  TSIMPEXPLP1

  ( ..., lvar, largexp, argLN, lexps, lvarLN, argexp, matr/lcoef, impossible? )
	
  casedrop TSIMPEXPLP2
  SWAPDROPSWAP		( ... lvar largexp argLN lexps lcoef lvarLN )
  4UNROLL		( ... lvar largexp lvarLN argLN lexps lcoef )
  >HCOMP 4ROLL		( ... lvar lvarLN argLN lexps largexp )
;

LABEL TSIMPEXPLP1
* list of objs, obj -> list of objs, obj, matr/lcoef, impossible?
* determines if obj is a lin. comb. of list of objs
::	
  ( * teste si lvarLN est vide  * )
  OVER DUPNULL{}? SWAPDROP
  case TrueTrue
			( lvarLN argexp )
  2DUP FXNDext		( lvarLN argexp lvarLN argexpn argexpd )
  3PICK INNERCOMP
  METADENOLCM		( lvarLN argexp lvarLN argexpn argexpd ppcmlvarLN )
  DUPUNROT SIMPext	( lvarLN argexp lvarLN argexpn ppcm argexpd_s ppcm_s )
  4ROLL RMULText	( lvarLN argexp lvarLN ppcm argexpd_s argexpn*ppcm_s )
  UNROT RMULText	( lvarLN argexp lvarLN argexpn*ppcm_s argexpd_s*ppcm )
  ROT LOPMext SWAP	( lvarLN argexp lvarLN*argexpd_s*ppcm argexpn*ppcm_s )
  NULL[] ZERO
  TSIMP8 	( lvarLN argexp lvarLN*argexpd_s*ppcm argexpn*ppcm_s lcoef # )
  2SWAP 2DROP		( lvarLN argexp { l-poly } poly lcoef # )

  ( si #=2 impossible, si #=1 on teste )

* si #=0, on devrait tester si la matrice reduite (systeme sous-determine)
* admet des racines rationnelles

  #1= NOTcase TRUE	( -> lvarLN argexp {l-poly} poly lcoef true )
			( lvarLN argexp {l-poly} poly lcoef )
  3PICKOVER XYext	( lvarLN argexp {l-poly} poly lcoef {l-poly}.lcoef )
  3PICK RSUBext 
* NEWTRIMext
  Z0 EQUALNOTcase TRUE
			( lvarLN argexp { l-poly } poly lcoef )
  TRIMext
  ( parait suspect ici mais peut-etre contrebalance par LRDMext dans TSIMPEXP )
  DUPINCOMP TRUE SWAP
  ZERO_DO (DO)
    TSIMPEXPLPR
  LOOP
  NOT			( lvarLN argexp { l-poly } poly lcoef flag )
;

LABEL TSIMPEXPLPR
:: 
  NOTcasedrop FALSE
  FXNDext TYPEZ? SWAP TYPEZ? AND
;


LABEL TSIMPEXPLP2
::
  SWAPOVER >HCOMP
  5UNROLL	( lvar lvarLN largexp argLN lexps argexp )
  ROTSWAP	( lvar lvarLN largexp lexps argLN argexp )

* weak normalization: retirer toutes les parties dependant lineairement
* des variables logarithmiques

  6PICK		( lvar lvarLN largexp lexps argLN argexp-l lvar )
  DUP1LAMBIND
  NULL{} 3PICK
  ROT LENCOMP	
   ( lvar lvarLN largexp lexps argLN argexp-l lin-argexp-l argexp-l #profond. )
  DUP#0= ITE_DROP TSIMPWEAK1

	( lvar, lvarLN, largexp, lexps, argLN, argexp, lin-argexp-l, a_0 )

  extract! >TCOMP

          ( lvar, lvarLN, largexp, lexps, argLN, argexp, lin-argexp-l )

  INNERCOMP reversym Z0 Z1 ROT
  DUP#1= ITE_DROP TSIMPWEAK3

  ABND

  ( lvar lvarLN largexp lexps argLN argexp cn+1 list symb )

  base_ln 4PICK x^ext x*ext

  ( lvar lvarLN largexp lexps argLN argexp cn+1 list symb )

  4UNROLL RADDext RSUBext

  ( lvar lvarLN largexp lexps argLN lin-argexp reste-argexp )

  COMPLEX? ITE xEXPext TSIMPREEXP
  x*ext			( lvar lvarLN largexp lexps argLN exp )
  >HCOMP UNROT 		( lvar lvarLN argLN largexp lexps )
  Z1 Z0
  5PICK LENCOMP #1- NDUPN
  #1+ {}N		( lvar lvarLN argLN largexp lexps expl )
  >HCOMP SWAP		( ... lvar lvarLN argLN lexps largexp )
;

LABEL TSIMPREEXP
:: 
  DUP RREext NEWTRIMext xEXPext
  SWAP RIMext NEWTRIMext DUP xCOSext
  SWAP xSINext 'xi
  SWAP x*ext x+ext x*ext
;

LABEL TSIMPWEAK1
::
  ZERO_DO			( lin-argexp-l, argexp-l )
    INTERNALX2
    NEWDIVext SWAPDROP		( lin-argexp-l, a_1x+a_0 )
    FXNDext SWAP 
    CKINNERCOMP TSIMPWEAK2
    3PICK RDIVext 		( lin-argexp-l, d, a_1, a_0 )
    SWAPROT RDIVext		( lin-argexp-l, a_0, a_1 )
    1GETLAM INDEX@ #1+ NTHCOMPDROP 
    pi EQUAL
    IT :: Z-i RMULText ;
    extract!			( lin-argexp-l, a_0, a_1 )
    ROTSWAP >TCOMP 
  SWAPLOOP
;

LABEL TSIMPWEAK2
::
  DUPTWO #>case NDROP0DUP
  #1= NOT?SEMI
  Z0 SWAP 
;

LABEL TSIMPWEAK3
::
  ONE_DO (DO)				( cn+1...c1 list symb )
  ROT
  1GETLAM INDEX@ NTHCOMPDROP	( cn+1...c2 list symb c1 var )
  TSIMPWEAK3LP
  LOOP
;

LABEL TSIMPWEAK3LP
::
  DUPTYPESYMB? NOTcase2DROP
  DUPLENCOMP #1= ITE 
  :: DUP pi EQUAL ;
  :: DUPDUP LENCOMP NTHCOMPDROP 
     EQ: xLN caseFALSE
     2DROP RDROP
     ;
					( ...list symb c1 var ?pi )
  DUPUNROT ITE :: DROP Z-1 ; xEXPext 	( ...list symb c1 ?pi exp[var] )
  5ROLL 4PICK				( ...symb c1 ?pi exp[var] list c1 )
  4ROLL ITE INTERNALiX INTERNALX
  RMULText INDEX@ n-1{}Next
  RADDext 4UNROLL
  SWAP x^ext x*ext
;

LABEL NDROP0DUP
:: NDROP Z0 DUP ;

LABEL extract!
::
  FXNDext extract1!
  DUPTYPEZ?
  NOTcase2drop Z0
  num#0=case SWAPDROP
  SWAP extract1!
  DUPTYPEZ?
  NOTcase2drop Z0
  SWAP RDIVext
;
                 
LABEL extract1!
::
  [BEGIN]
    DUPTYPELIST?
  [WHILE]
    LASTCOMP
  [REPEAT]
;

NULLNAME LASTCOMP
:: DUPLENCOMP NTHCOMPDROP ;

LABEL n-1{}Next
::
  DUP#1= caseDROP
  ONE_DO (DO) ONE>POLY LOOP
;


	

LABEL TSIMP8
::	

  ( #0 signifie continuer a chercher, #1 a tester, #2 impossible ) 
  ( { l-poly }, poly, matr, # -> { l-poly }, poly, matr, # )

  DUP#0= NOT?SEMI
  ( * find the #rows of the matrix to build [assuming univariate poly ] * )
  3PICK TSIMPLEN
				( { l-poly }, poly, matr, #0, #len )
  5PICK INNERCOMP get1 SWAP
		( { l-poly }, poly, matr, #0, poly1, ..., polyn, #len, #n )
  ZERO_DO (DO)
    SWAP TSIMPLEN #MAX
  LOOP
				( {l-poly} poly matr #0 #len )
  #1+_ONE_DO (DO)		( {l-poly} poly matr # )
    TSIMP8LP
  LOOP				( {l-poly} poly matr # )
;

LABEL TSIMPLEN
::
  DUPTYPELIST? ( was DTYPEMATRIX? )
  case LENCOMP DROPONE
;

LABEL TSIMP8LP
( * build one line of the matrix, reduce it, stop if no solution * )
( * INDEX@ is the line number * )
::
  DUP#0= NOTcase ExitAtLOOP
  3PICK INDEX@ TSIMP7		( {l-poly} poly matr #0 coef )
  5PICK INNERDUP #1-SWAP
  ZERO_DO			( {lp} poly matr #0 coef p1..pn #n-1 )
    get1 JINDEX@ TSIMP7
    SWAPLOOP
				( {lp} poly matr #0 coef c1..cn #n-1 )
  #1+ get1 SWAP#1+ []N		( {lp} poly matr #0 [c1..cn coef] )
  ( * look if there are polynomials inside the ck * )
  DUPINCOMP FALSE SWAP
  ZERO_DO (DO)
    SWAP TYPELIST? OR
  LOOP
				( {lp} p m #0 [c1..cn coef] list? )
  ( * if there are poly we must write the system for each power of the poly * )
  ( * since the coeff/ck were not univariate * )
  case TSIMP8LPL

  ( * add line to the matrix * )
  SWAPDROP			( {lp} p m line )
  SWAPINCOMP get1 SWAP#1+	( {lp} p meta-m )
  ( * now reduce it without step by step * )
  NULL{} ONE NOSTEPEVAL METAMATRED 2DROP	
				( {l-poly} poly flag1 meta_matr_reduite )

  ( * meta-matr reduite -> matr reduite, #0 ou #2 ou lcoef #1 * )

  OVER trimlist! 		( {l-poly} poly meta-redmat lastline{} )

  ( si non liste, c'etait une ligne du type { 0 ... 0 x } )

  NOTLISTcase TSIMP8LPNL
  DROPOVER			( {l-poly} poly meta-redmat lastline[] )
  LENCOMP #1- OVER#=		( {l-poly} poly meta-matr_reduite Cramer? )
  NOTcase :: []N ZERO ;
  NULL{} SWAP			( {l-poly} poly l1..ln {} #n )
  ZERO_DO (DO)			( {l-poly} poly l1...lk lcoef )
    SWAP trimlist!
    DUPDUP LENCOMP NTHCOMPDROP
    SWAP CARCOMP RDIVext
    >HCOMP
  LOOP				( {l-poly} poly lcoef )
  ONE				( {l-poly} poly lcoef #1 )
;

LABEL TSIMP8LPNL
::				( {l-poly} poly meta-redmat last_obj )
  j#0=case TSIMP8LP0
  ( x<>0: incompatible ) 
  NDROP NULL[] TWO		( -> {l-poly} poly null[] #2 )
;

LABEL TSIMP8LP0
::				( {l-poly} poly meta-redmat )
  pulldrop []N ZERO		( -> {l-poly} poly matrix #0 )
;

LABEL TSIMP8LPL
( * at least one of the poly was not univariate * )
( * call recursively TSIMP8 one var deeper * )
::
				( {lp} p m #0 [c1..cn coef] )
  INNERCOMP #1-SWAP
  OVER#2+UNROL []N		( {lp} poly matr #0 coef [c1..cn] )
  SWAP 2SWAP 			( {lp} poly [c1..cn] coef matr #0 )
  TSIMP8
  2SWAP 2DROP			( {l-poly} poly matr # )
;


LABEL TSIMP7
::	
  ( TSIMP7 renvoie le terme de degre n-1 d'un objet vu comme un poly )
  ( obj, #n -> terme )
  DUP#1= casedrop TSIMP7_1
  OVER TYPELIST?
  NOTcase2drop Z0
					( obj, #n )
  OVER LENCOMP				( obj, #n, #l )
  2DUP#> case2drop DROPZ0
  SWAP#- #1+ NTHCOMPDROP
;

LABEL TSIMP7_1
:: 
  DUPTYPELIST? NOT?SEMI 
  DUPLENCOMP NTHCOMPDROP 
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME SQFF2ext ( l1, ..., ln, #n-1 -> l1', ..., ln', #n-1)
* teste s'il y a des pgcd non triviaux entre les facteurs des lj
* pour renvoyer une liste de facteurs premiers entre eux
::
  ( on recherche les pgcd non triviaux liste a liste )
  DUP#0= ?SEMI
  get1 SWAPDUP		( l2, ..., ln, l1, #n-1, #n-1 )
  ZERO_DO
      get1 INDEX@ #2+ DUP
      ONE_DO		( li+1, ...,ln, l1, ..., li-1, #n-1, li, #i )
         get1 ROT SQFF2LP
        UNROT 4UNROLL
    [LOOP]
    DROP 
  SWAPLOOP
;

LABEL SQFF2LP
::  ( simplification de 2 listes )
						( l1, l2 )
  DUPNULL{}? ?SEMI
  SWAP DUPNULL{}? case SWAP
  SWAP
  DUPLENCOMP #2/
  #1+_ONE_DO					( l1, l2 )
    SQFF2LPLP
  [LOOP]
;

LABEL SQFF2LPLP
::					( l1, l2 )
  DUPINDEX@ #2* #1- NTHCOMPDROP		( l1, l2, f2i )
  NOTLISTcase DROP
  ROT DUPLENCOMP #2/
  #1+_ONE_DO				( l2, f2i, l1 )
    SQFF2LPLPLP
  [LOOP]

    ( on remplace f2i par f2i' dans l2 )
					( l2 f2i' l1 )
  UNROTSWAP
  INDEX@ #2* #1-SWAP PUTLIST
;					( l1 l2 )

LABEL SQFF2LPLPLP
::
  DUPINDEX@ #2* #1- NTHCOMPDROP	( l2, f2i, l1, f1j )
  NOTLISTcase DROP
  3PICK TWO{}N LPGCDext		( l2 f2i l1 {f1j f2i} pgcd )
  SQFF2LP3_1
;

LABEL SQFF2LP3_1
::
  NEWTRIMext num#1=case 2DROP
  DUPUNROT LOPDext		( l2 f2i l1 pgcd {f1j' f2i'} )
  INCOMPDROP SQFF2LP3_2		( l2 f2i' l1, pgcd f1j' )
  NEWTRIMext
  num#1=case 2DROP
  ROT INDEX@ #2* #1-SWAP
  PUTLIST			( l2 f2i' pgcd l1 mf par f1j' )
  DUPINDEX@ #2* NTHCOMPDROP
  3PICK				( l2 f2i' pgcd l1 mf par f1j' multipl pgcd )
  ADDLISText			( l2 f2i' pgcd l1 )
	
           ( si pgcd est un reel ou un compl., on refait le pgcd )
	
  SWAP num#1=case DROP 
				( l2 f2i' l1 pgcd )
  XEQTYPE SWAPDROP %2 %>= ?SEMI
  INDEX@ #1- INDEXSTO
;

LABEL SQFF2LP3_2
::
  NEWTRIMext num#1=case DROP
  ( mise a jour de l2 )
  5ROLLDROP 4UNROLL		( l2 f2i' l1 pgcd f1j' )
  5ROLL DUP
  JINDEX@ #2* NTHCOMPDROP
  4PICK				( f2i' l1 pgcd f1j' l2 m pgcd )
  ADDLISText
  DUPLENCOMP #2/ #1+ JSTOPSTO
  5UNROLL			( l2+pgcd+m f2i' l1 pgcd f1j' )
;


