INCLUDE erromid.h
INCLUDE erext.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s

	STITLE Integration

**********************************************************************
* EXPORT:
*
* IBP		( symb, u -> u*v, -u*v' )
*		( symb, u -> symb' )
*		Integration By Parts. 
*		If symb does not contain an integral sign, returns u*v, -u*v'
*		where v=symb/u', 
*		Else symb'=symb mathematically with the same integration
*		by part applied on the integral
*		Note that if symb contains an = signe, this tranformation
*		applies only to the right hand side
*
* SQRT_IN?	( {} -> {}, flag )
*		Returns TRUE if one element of {} is a symb containing a sqrt
*
* STOPRIMIT	( symb -> )
*		Stores antiderivative in PRIMIT variable
*
* RISCHext	( symb, idnt -> symb )
*		Undefined integral
*
* LVARXNXext	( symb -> symb, x, {}, {}' )
*		x is the current var,
*		{} is the list of constant var w.r.t. x
*		{}' is the list of var that depends on x
*
* PREVALext	( symb, inf, sup, x -> symb|x=sup - symb|x=inf )
*		Evaluate an antiderivative between 2 bounds
*		Does not check for discontinuities of symb in this interval
*
* INText	( symb, x -> int[$,x, symb, xt] )
*		Return unevaluated integral
**********************************************************************


NULLNAME RISCHext
::
( algorithme de Risch: symb, rvx -> primitive )
  OVER DUPTYPESYMB? NOTcasedrop RISCH
					( symb rvx symb )
  LASTCOMP EQ: xAND NOTcase RISCH
					( symb rvx )
  SWAP EXPR> 2DROP			( rvx symb_[] symb_non_int )
  ROT RISCH x+ext			( symb )
;

LABEL RISCH
::				( symb rvx )
  RADON SWAP			( rvx symb )
* simplifies var inside expr, necessary e.g. for embedded integrals
*  SEVAL					( symb )
* ! Replace all division by EXP[] by a multiplication by EXP[-]
  { 
    x/ xABS xSIGN xEXP
    xSINH xCOSH xTANH 
    xASINH xACOSH xATANH 
    xLNP1 xEXPM xLOG xALOG x\.S_ALGEBRAIC
  } 
  { 
    risch/ rischABS addtSIGNonly addtEXPonly
    sinh2exp cosh2exp tanh2exp 
    asinh2ln acosh2ln atanh2ln 
    addtLNP1 addtEXPM addtLN/LN10 addt*LN10EXP REP.s
  } 
  REPLACE2BY1			( rvx symb )

  SWAP 				( symb rvx )

* extract idnt/lam part of the integration variable
  SYMBINCOMP #1= NOTcase INTVARERR
  IDNTLAM? NOTcase INTVARERR
				( symb rvx )
* Purge integration var and add it to the real var
  DUP SAFEPURGE			( symb rvx )
  DUP ADDTOREAL

* Lam binding

  RCLVX				( symb, rvx, savevx )
  2DUP				( symb, rvx, savevx, rvx, savevx )
  OVER EQUAL ITE_DROP STOVX
  SWAP				( symb, savevx, rvx )
  Z0 Z0				( rck and discontinu )
  DUPDUP			( rr# )
  				( profondeur -> LAM rprof )
  ZEROZEROZERO
  ZEROZEROZERO
  ZEROZEROZERO
  ZEROZEROZERO

*  {{ savevx rvx rck discontinu rr# dlvar rprof rcos rarg rischflag coeflog autolog type laurent ratpart temp_stor lvar_val depth_val }}

  DEFINE savevx 18GETLAM
  DEFINE rvx 17GETLAM
  DEFINE rck 16GETLAM
  DEFINE discontinu 15GETLAM
  DEFINE rr# 14GETLAM
  DEFINE dlvar 13GETLAM
  DEFINE rprof 12GETLAM
  DEFINE rcos 11GETLAM
  DEFINE rarg 10GETLAM
  DEFINE rischflag 9GETLAM
  DEFINE coeflog 8GETLAM
  DEFINE autolog 7GETLAM
  DEFINE type 6GETLAM
  DEFINE laurent 5GETLAM
  DEFINE ratpart 4GETLAM
  DEFINE temp_stor 3GETLAM
  DEFINE lvar_val 2GETLAM
  DEFINE depth_val 1GETLAM
  DEFINE savevx! 18PUTLAM
  DEFINE rvx! 17PUTLAM
  DEFINE rck! 16PUTLAM
  DEFINE discontinu! 15PUTLAM
  DEFINE rr#! 14PUTLAM
  DEFINE dlvar! 13PUTLAM
  DEFINE rprof! 12PUTLAM
  DEFINE rcos! 11PUTLAM
  DEFINE rarg! 10PUTLAM
  DEFINE rischflag! 9PUTLAM
  DEFINE coeflog! 8PUTLAM
  DEFINE autolog! 7PUTLAM
  DEFINE type! 6PUTLAM
  DEFINE laurent! 5PUTLAM
  DEFINE ratpart! 4PUTLAM
  DEFINE temp_stor! 3PUTLAM
  DEFINE lvar_val! 2PUTLAM
  DEFINE depth_val! 1PUTLAM
  ' NULLLAM EIGHTEEN NDUPN DOBIND
				( symb )
  CKSYMREALCMP
* initialise la variable CASINFO

  DUP XGROBext STOCASINFO		( symb )
  risch0!				( primitive )

  DUP rvx savevx ABND 2DUP		( prim prim rvx savevx rvx savevx )
  EQUAL case2drop STOPRIMIT
					( prim prim rvx savevx )
  DUP STOVX				( prim prim rvx savevx )
  NR_REPLACE STOPRIMIT
;

LABEL STOCASINFO
* obj ->
::
  CONTEXT@ CASDIR		( obj rep )
  SWAP' ID CASINFO STO		( rep )
  CONTEXT!
;

NULLNAME risch/
::
  ( meta1 meta2 )
  AEQ1stcase xEXP risch-EXP*
  MetaDiv
;

NULLNAME rischABS
::
  dup addtSIGNonly metamult
;


LABEL risch-EXP*
::
  pulldrop metaneg addtEXPonly MetaMul
;


**********************************************************************
* Desc:		Internal IBP
* Stack:	( u'*v u -> u*v -u*v' )
* Note:		if u is a constant return INTVX(u'*v)+u
*		if stack 2 is a list it must be of the form { olduv u'*v }
*		then olduv will be added to u*v at stack level 2
*		This permits multiple IBP in algebraic mode
*		IBP(ASIN(X)^2,X)
*		IBP(ANS(1),sqrt(1-X^2))
*		IBP(ANS(1),C) the last step with an integral containing a cst C
**********************************************************************
NULLNAME IBP
::					( expr u )
  SWAP dLISTcase IBPLIST
					( u expr )
  DUPTYPESYMB? NOTcase IBPSYMB		( uv -u*v' )
  DUP LASTCOMP EQ: xAND NOTcase IBPSYMB
					( u expr )
  EXPR> 2DROP INIBPLIST
;

LABEL IBPLIST
::					( u { olduv expr } )
  INNERCOMP #2= NcaseSIZEERR		
  INIBPLIST
;

LABEL INIBPLIST
::					( u olduv expr )
  SWAP CKALG				( u expr olduv )
  UNROT INIBPSYMB				( olduv uv -u*v' )
  UNROT x+ext SWAP			( olduv+uv -u*v' )
ASSEMBLE
	IF	f39
RPL
  AND39
ASSEMBLE
	ENDIF	f39
RPL
;

LABEL INIBPSYMB
::					( u expr )
  CKALG					( u expr )
  OVER RCLVX DIRECTDERIV		( u expr u' )
  Z0 EQUALcasedro IBPSYMB0
					( u expr=u'*v u' )
  NDXFext SIMP1ext			( u v )
  2DUP x*ext				( u v uv )
  UNROT RCLVX DIRECTDERIV		( uv u v' )
  2SYMBINCOMP MetaMul MetaNeg 
  metasimp CKSYMBN			( uv -u*v' )
;

LABEL IBPSYMB
::
  INIBPSYMB
ASSEMBLE
	IF	f39
RPL
  AND39
ASSEMBLE
	ENDIF	f39
RPL
;

LABEL IBPSYMB0
* if u is a constant then simply integrate expr and add u
::					( u expr )
  RCLVX RISCHext x+ext Z0		( integral 0 )
ASSEMBLE
	IF	f39
RPL
  AND39
ASSEMBLE
	ENDIF	f39
RPL
;


**********************************************************************
NULLNAME SQRT_IN?
( lvar -> lvar, [ varsqrt, true | false ] )
::
  DUP' IS_SQRT?
  Find1stTrue
;
**********************************************************************
NULLNAME IS_SQRT?
::	
  DUPTYPESYMB? NOTcase DROPFALSE
  DUPLENCOMP NTHCOMPDROP EQ: x\v/ 
;
**********************************************************************
NULLNAME XROOT_IN?
( lvar -> lvar, [ varxroot, true | false ] )
::
  DUP ' IS_XROOT?
  Find1stTrue
;
**********************************************************************
NULLNAME IS_XROOT?
::	
  DUPTYPESYMB? NOTcase DROPFALSE
  DUPLENCOMP NTHCOMPDROP EQ: xXROOT 
;
**********************************************************************
NULLNAME STOPRIMIT
::				( symb )
  CONTEXT@ CASDIR		( symb rep )
  SWAP' ID PRIMIT STO		( rep )
  CONTEXT!
;
**********************************************************************

NULLNAME CONTAINS_LN?
::	
  DUPTYPESYMB? NOTcase DROPFALSE
  ' xLN ' EQ POSCOMP #0<>
;

NULLNAME ISNT_IDNT?
::	
  TYPEIDNTLAM? NOT 
;


**************************************
* Integration of rational fractions
**************************************
LABEL rischpfstep
::
  stepflag TestSysFlag NOT?SEMI
					( newlvar fr[x] lvar )
  3PICK3PICK R2SYM 			( newlvar fr[x] lvar fr[x]-sym )
  BINT87 GETERABLEMSG rischstepout	( newlvar fr[x] lvar sym )
  DROP					( newlvar fr[x] lvar sym )
;

NULLNAME RISCHPF
( newlvar, fr[x], lvar -> primitive )
::					( newlvar fr[x] lvar )
  rischpfstep				( newlvar fr[x] lvar )
  NULL{} autolog!			( newlvar, symb, lvar )
  ROT lvar_val!				( symb, lvar )
  SEVEN PFext

  ( lvar, quotient, num1, deno1, f1, m1, ..., numk, denok, fk, mk, #4k )

* rational and logarithmic part
* -> lvar, quotient, result

  RISCHPF2				( lvar, p=quotient, result )

* polynomial part

  ROTDROPSWAP RISCH13			( result, p )
  INTEGRext
  VAL1M get1 SYMBINCOMP metaadd SYMBN	( symb )
;

LABEL RISCHPF2
::	
  Z0 SWAP DUP#0=csDROP
  FOUR #/ SWAPDROP
  ZERO_DO (DO)		( lvar quotient ... numk denok fk mk result )
    5UNROLL		( lvar quotient ... result numk denok fk mk )
* rational part
    %ABSCOERCE
    2DUP RP# rr#!	( lvar quotient ... result numk denok fk #mk )
    SWAPDUP DERext	( lvar quotient ... result numk denok #mk fk fk' )
    3PICK #1=
    ITE ROTDROP RISCHRAT
			( lvar quotient ... new-res new q.n new_q.d r2 r2' )
* logarithmic part
    SWAPDUP temp_stor! ONE{}N 4UNROLL
			( lvar quotient ... new-res new_q.n new_q.d r2' )
    rischlogpart	( lvar quotient ... newnewres )
  LOOP
;


NULLNAME RISCHRAT
::
  2DUP EGCDext

  ( lvar, quotient, ..., result, numk, denok, #mk, fk, fk', g, a, b )
  ( avec g=b*fk'+a*fk )

  Z0 ratpart! Z1 temp_stor!
  6PICK
  ONE_DO (DO)

* q.n=numk, q.d=denok, fk=r2
* nouveau q2= [q2/g]*a+ [ [q2/g]*b/[i-1] ]' 
* = q2.n*a*[i-1] +q2.n'*b / g*q2.d*[i-1] 
* a rajouter a result: -[q2/g]*b/[i-1]/fk^i-1 )

				( lvar quo...result numk denk # fk fk' g a b )
	8ROLL 8ROLL		( ... # r2 r2' g a b q.n q.d )
	OVERDUP 5PICK RMULText
	DERext			( # r2 r2' g a b q.n q.d q.n q.n*b' )
	SWAP 6PICK RMULText	( # r2 r2' g a b q.n q.d q.n'*b q.n*a )
	ISTOP-INDEX #>Z
	DUP4UNROLL		( # r2 r2' g a b q.n q.d i-1 q.n'*b q.n*a i-1 )
	RMULText RADDext	( # r2 r2' g a b q.n q.d i-1 q.n'*b+q.n*a*[i-1] )
	( on divise par r2#i-1 )
	rr# TEN PICK QUOText
	DUP rr#! PSEUDODIV
	SWAPDROP		( # r2 r2' g a b q.n q.d i-1 deno q.n'*b+q.n*a*[i-1] )
	ELEVEN UNROLL		( new_q-n # r2 r2' g a b q.n q.d i-1 den )
	UNROT RMULText		( new_q-n # r2 r2' g a b q.n den q.d*i-1 )
	DUPUNROT RMULText	( new_q-n # r2 r2' g a b q.n q.d*i-1 q.d*i-1*den )
	6PICK RMULText
	NINE UNROLL		( new_q-n new_q-d # r2 r2' g a b q.n q.d*i-1 )
	SWAP3PICK RMULText	( new_q-n new_q-d # r2 r2' g a b q.d*i-1 q.n*b )
	rr# PSEUDODIV SWAPDROP	( new_q-n new_q-d # r2 r2' g a b q.d*i-1 den q.n*b )
	temp_stor RMULText
	UNROT RMULText		( new_q.n new_q.d # r2 r2' g a b q.n*b* q.d*i-1*den )
	RDIVext	ratpart
	RADDext ratpart!	( new_q.n new_q.d # r2 r2' g b a )
	5PICK temp_stor
	RMULText temp_stor!	( new_q.n new_q.d # r2 r2' g b a )
  LOOP
  2DROP				( lvar quo...result new_q.n new_q.d # r2 r2' g )
  7ROLL ratpart ROT RDIVext	( lvar quo...new_q.n new_q.d # r2 r2' res coef )
  4PICK 6ROLL #1- UNCOERCE	( ...res coef den power )
  SWAP VAL1M get1 meta^		( ...res coef den^power-meta )
  get1 VAL1M pshtop& addt/	( lvar quo...new_q.n new_q.d r2 r2' res coef/r2^-meta )
  obj2-meta1
  SYMBN				( lvar quo...new_q.n new_q.d r2 r2' new-res )
  5UNROLL			( lvar quo...new-res new_q.n new_q.d r2 r2' )
;


***********************************
* All extension types: logarihtmic part
*************************************

( result, lfac, q2-n, q2-d, d2' -> result )

NULLNAME rischlogpart

* d2 is in temp_stor, degree of q2 must be <= degree d2' )
* updates coeflog
::

* first tests if there is something to do, is the list of factors empty?
  4PICK NULLCOMP? case 4DROP
					( result, lfac, q2-n, q2-d, d2'  )

* autolog is the authorized log part, NULL{} means all log are allowed
* autolog is used for logarithmic extensions

  autolog NULLCOMP? NOTcasedrop RISCH1LOG

					( result, lfac, q2-n, q2-d, d2'  )

* if current depth is 0, extract fast the log part q2 div d2'

  rprof #0= IT RISCHLOG			( result, lfac, r, q2-d, d2' )

* look if it remains something to do

  ROT num#0=case 4DROP
					( result, lfac, q2-d, d2', r )

* for rational fraction (depth=0), we can generate the ATAN part much faster
* if there is only 1 factor of degree 2

  rprof #0=ITE
    :: 4PICK LENCOMP #1= NOTcaseFALSE
       temp_stor LENCOMP #3= 
    ; 
    FALSE
  case RISCHATAN
					( result, lfac, q2-d, d2', r )
  UNROT RMULText RDIVext FXNDext	( result, lfac, q2, d2' )
  ROT					( result, q2, d2', lfac )

  ( factorization of lfac to the list of roots: lrac )

  INNERCOMP ZEROSWAP
  ZERO_DO (DO)				( fac1, ..., facn, meta-res )
     get1 EVIDENText DROP top&
  LOOP
					( result q2 d2' meta-rac )
  THREE OVER#2+UNROL			( result q2 d2' #3 r1...rk #k )
  psh DROP				( r1...rk #k result q2 d2' )
  4ROLL
  ZERO_DO (DO)				( r1...rk result q2 d2' )
     2DUP 6ROLL
     DUPTYPETAG? ITE
        generate_atan
        gen_ln_only
					( r1...rk-1 result q2 d2' ln[x-rk]*d2'[rk]/q2[rk] )
     4ROLLSWAP x+ext UNROT
  LOOP
					( result, q2, d2' )
  2DROP					( result )
;

*************************************************
NULLNAME PREVALext
( primitive, inf, sup, vx -> prim[sup]-prim[inf] )
* must check for INT inside and replace by int(inf,sup,fnc,.)
::
  3PICK3PICK EQUALcase :: 4DROP Z0 ;
					( primitive inf sup vx )
  ( * 03/2000 added check for inf and sup in bad order * )
  3PICK APPROXCOMPEVAL			( primitive inf sup vx inf% )
  DUPTYPEREAL? NOTcasedrop PREVAL
  3PICK APPROXCOMPEVAL			( primitive inf sup vx inf% sup% )
  DUPTYPEREAL? NOTcase2drop PREVAL
					( primitive inf sup vx inf% sup% )
  %< case PREVAL
					( primitive inf sup vx )
  ROTSWAP PREVAL QNeg			( -> -res )
;

LABEL PREVAL
::
  DUPTYPESYMB? IT CARCOMP		( primitive inf sup vx )
  CHECKSING				( primitive inf sup vx flag )
  IT WARNSING				( primitive inf sup vx )
  DUP4UNROLL				( primitive vx inf sup vx )
  5PICK UNROTSWAP			( primitive vx inf primitive vx sup )
  %-2 LIMIText STRIPTAGS		( primitive vx inf val_sup )
  4UNROLL %2 LIMIText STRIPTAGS		( val_sup val_inf )
  num#0=case DROP			( -> val_sup )
  OVER NEWTRIMext j#0=case :: SWAPDROP QNeg ;
					( val_sup val_inf )
  2SYMBINCOMP prevalmetasub CKSYMBN	( res )
  stepflag TestSysFlag ?SEMI		( res )
  SIMPLIFY				( simplify again )
;

LABEL prevalmetasub
::
  DUP #3+PICK 				( meta1 meta2 op1 )
  EQ: 
ASSEMBLE
	IF	f39
RPL
  xint
ASSEMBLE
	ELSE	f39
RPL
  xINT
ASSEMBLE
	ENDIF	f39
RPL
  case metasubint
  ( * changed dometasub to MetaSub because something-unsignef inf -> -inf * )
  MetaSub
;

**********************************************************************
* Desc:		Warn user for singularity
* Stack:	( symb inf sup vx -> symb inf sup vx )
**********************************************************************
NULLNAME WARNSING
::					( symb inf sup vx )
  BINT68 GETERABLEMSG
ASSEMBLE
	IF	f39
RPL
  DoQueryBox
ASSEMBLE
	ELSE	f39
RPL
  % 5000 % 0.3 DOBEEP
  AskQuestion
ASSEMBLE
	ENDIF	f39
RPL
  ?SEMI
  BINT69 ERABLEERROR
;


****************************************
NULLNAME INText
( f[x], x -> int[f[xt],xt,x] )
::
  SWAPOVER DUP ID>$ CHR_t >T$ $>ID 	( x f[x] x xt )
  DUP4UNROLL NR_REPLACE			( x t f[t] )
  symcomp SWAP symcomp 			( x f[t] t )
  xf[t]t2int
  SYMBN
;

LABEL xf[t]t2int
::					( x f[t] t )
  DUP ADDTOREAL				( x f[t] t )
  ROT SYMBINCOMP			( f[t] t meta-x )
  #2+ 					( meta )
  addt:
ASSEMBLE
	IF	f39
RPL
  xint
ASSEMBLE
	ELSE	f39
RPL
  xINT
ASSEMBLE
	ENDIF	f39
RPL
					( meta )
;

*******************************************
**********************************************************************
* Desc:		Undefined integration
* Stack:	( f(x), x, y -> F(y) where F'=f )
* Note:		No limit for underdetermined form
**********************************************************************
NULLNAME INT3
::					( f x y )
  CKSYMREALCMP SWAPROT 			( y x f )
  CKSYMREALCMP EXACT? 			( y x f ? )
  NOTcase INT3APPROX
					( y x f )
  OVER RISCHext				( y x F )
  SWAPROT				( F x y )
  2DUP EQUALcase 2DROP
					( F x y )
  NR_REPLACE
;

LABEL INT3APPROX
::				( y x f )
  %0 4UNROLL			( low high x f )
  SWAP INT4APPROX		( res )
;

LABEL INT4APPROX
::
  ' x\.S_ALGEBRAIC 		( low high f x integral )
  FIVE SYMBN ALGCASCOMPEVAL	( res )
;

**********************************************************************
* Desc:		compute n-th Fourier coefficient of a 2 pi periodic function
* Stack:	( symb, n  -> cn )
* Note		cn=1/[period]*int[0,period,symb*exp[-i*n*x*2*pi/period] dx] )
**********************************************************************
NULLNAME FOURIERext
::
  RCLVX DUP4UNROLL            ( x symb n x )
  SYMBINCOMP get1 SYMBINCOMP MetaMul    ( x symb n*x-meta )
  metai metaneg MetaMul         ( x symb -i*n*x-meta )
  RCLPERIOD DUP1LAMBIND            ( x symb -i*n*x period )
  metapi*2 get1 SYMBINCOMP        ( x symb -i*n*x pi/2-meta period-meta )
  MetaDiv MetaMul            ( x symb -i*n*x*2pi/periode-meta )
  addtEXPNO-CK                 ( x symb exp-meta )
  get1 SYMBINCOMP MetaMul CKSYMBN    ( x symb*exp )
  OVER RISCHext                ( x primitive )
  Z0 1GETLAM 4ROLL
  switchflag SysITE PREVALext
                :: switchflag SetSysFlag PREVALext switchflag ClrSysFlag ;
                    ( F[P]-F[0] )
  1GETABND x/ext            ( coef )
;
