INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s

	STITLE Linearisation and transc. expand

**********************************************************************
* EXPORT:
*
* TCOLLECT	( symb -> symb )
*		Does first trigonometric linearization then collects
*		sin and cos of the same angle
*
* SIGMAEXPext	( symb -> symb )
*		Conversion to exp and ln and exponential linearization
*
* LINEXPext	( symb -> meta )
*		meta is made of arg_exp1, coef1, ..., arg_expn, coefn, #2n
*
* SIGMAEXP2ext	( meta -> symb )
*		Back conversion from a arg_exp/coef-meta to a symbolic
*
* TCHEBext	( Z -> P )
*		If Z>0 1st type Tchebycheff polynomials,
*		If Z<0 2nd type
*
* SINEXPA	( symb -> symb )
*
* METASINEXPA	( meta -> meta )
*		expand sin()
*
* COSEXPA	( symb -> symb )
*
* METACOSEXPA	( meta -> meta )
*		expand cos()
*
* EXPEXPA	( symb -> symb )
*
* METAEXPEXPA	( meta -> meta )
*		expand exp()
*
* METAINT?	( meta -> meta, flag )
*		flag indicates if meta is an integer
*
* METAPOSINT?	( meta -> meta, flag )
*		flag indicates if meta is a positive integer
*
* OBJINT?	( obj -> obj, flag )
*		flag indicates if obj is an integer
*
* OBJPOSINT?	( obj -> obj, flag )
*		flag indicates if obj is a positive integer
*
* LNEXPA	( symb -> symb )
*
* METALNEXPA	( meta -> meta )
*		expand ln()
*
* LINEXPA	( symb -> meta )
*		meta alternates trig operator and coeff
*
* MTRIG2SYMB	( meta -> symb )
*		Back conversion of a trig-meta to a symbolic
*
* LNCOLCext	( symb -> symb )
*		Collect logarithms
*
* TEXPAext	( symb -> symb )
*		Expands all transcendental functions of symb
**********************************************************************

**********************************************************************
* Desc:		Perform trigonometric linearization and then
*		collect sin and cos of the same angle.
* Stack:	( symb --> symb' )
**********************************************************************
NULLNAME TCOLLECT
:: 
  LINEXPA			( trig1, coef1, ..., trign, coefn, #2n )
  COMPLEX? case MTRIG2SYMB
  DUP #2= case MTRIG2SYMB
  DUP #2/
  ZERO_DO (DO)			( trig1, coef1, ..., trign, coefn, #2n )
     UNROT TWO{}N		( trig1, coef1, ..., #2n, { trign coefn } )
     OVER UNROLL		( { trign coefn }, trig1, coef1, ..., #2n )
     #1-
  LOOP
				( { trig1 coef1 }, .., { trign coefn }, #n )
  {}N

  ' :: CARCOMP SWAP CARCOMP IsV>V? ;
  SortList

				( { { trig1 coef1 }, .., { trign coefn } } )
  INNERCOMP Z0 SWAP
  ZERO_DO (DO)			( { trig1 coef1 }, .., { trign coefn }, res )
     tcollect1!
     ITE tcollect2! tcollect3!
     x+ext			( ..., {trign-1 coefn-1}, res )
  LOOP
;

* --------------------------------------------------------------------
* Desc:
* Stack:	( -> ..., res, FALSE | ..., res, arg, TRUE )
* --------------------------------------------------------------------
LABEL tcollect1!
::
  ISTOP-INDEX #1=case FALSE

* teste si trign et trign-1 peuvent etre combines

  3PICK CARCOMP DUPTYPESYMB? NOTcasedrop FALSE
  INNERCOMP pulldrop SYMBN

  3PICK CARCOMP DUPTYPESYMB? NOTcase2drop FALSE
  INNERCOMP pulldrop SYMBN			( ..., arg_n-1, arg_n )

  OVER EQUAL NOTcasedrop FALSE TRUE
;

* --------------------------------------------------------------------
* Desc:
* Stack:
* --------------------------------------------------------------------
LABEL tcollect2!
::
				( ...{trign-1 coefn-1} {trign coefn} res arg )
  ISTOP@ #1- ISTOPSTO
  4ROLL TWO NTHCOMPDROP		( ...{trign coefn} res arg coefn-1 )
  4ROLL INCOMPDROP		( ..., res, arg, coefn-1, trign, coefn )
  4ROLLROT
  DUPLENCOMP NTHCOMPDROP	( ..., res, coef1, coef2, arg, secon )

* avec SortList, on doit toujours avoir secon=xSIN ??

  EQ: xSIN IT ROTSWAP		( res coefsin coefcos, arg )
  UNROTSWAP RNEGext SWAP2DUP	( res arg coefsin, coefcos, coefsin, coefcos )
  DUP RMULText
  SWAPDUP RMULText RADDext
  xvext SIMP1ext		( res, arg, coefsin, coefcos, sqrt[] )
  4UNROLL ARG2 CKSYMBN
  x+ext xCOSext x*ext		( res, a_ajouter )
;

* --------------------------------------------------------------------
* Desc:
* Stack:
* --------------------------------------------------------------------
LABEL tcollect3!
::					( ..., { trign coefn }, res )
  SWAP INCOMPDROP			( ... res trign coefn )
  SIMP1ext SWAP 			( ... res coefn trign )
  num#1=case DROP
  x*ext					( res, a_ajouter )
;

**********************************************************************
* Desc:		Conversion to exp and ln with exponential linearization
* Stack:	( symb --> symb' )
**********************************************************************
NULLNAME SIGMAEXPext
::
  EXPLNext				( symb )
  LINEXPext
  SIGMAEXP2ext
;

**********************************************************************
* Desc:		
* Stack:	( symb --> Meta )
* Note:		Meta = arg_exp1 coef1 ... arg_expn coefn #2n
**********************************************************************
NULLNAME LINEXPext
::
  DUPTYPESYMB? NOTcase Z0SWAP2
  INNERCOMP
  linexpmeta!					( meta )
  simpmetacp!					( meta )
;

LABEL Z0SWAP2
:: Z0 SWAP TWO ;

LABEL CKSYMBNZ0SWAP2
:: CKSYMBN Z0SWAP2 ;

* --------------------------------------------------------------------
* Desc:		
* Stack:	( meta -> meta' )
* Note:		meta represents a symbolic
* --------------------------------------------------------------------
LABEL linexpmeta!
::
  ' xEXP EQUALPOSMETA SWAPDROP			( meta # )
  #0=case CKSYMBNZ0SWAP2
  OVER						( meta op )
  linexpop ListPos				( meta # )
  DUP#0=csedrp linexpnot!
  linexpaction SWAP NTHCOMPDROP EVAL
;

LABEL linexpop
{ 
  x+ x- x* x/ x^ 
  xEXP xINV xSQ xNEG 
}

LABEL linexpaction
{ 
  linexp+! linexp-! linexp*! linexp/! linexp^! 
  linexpexp! linexpinv! linexpsq! linexpneg!
}
	
* --------------------------------------------------------------------
* Desc:		Simplifies the arg part of a exp-linearization meta
* Stack:	meta -> meta 
* --------------------------------------------------------------------
LABEL simpmetacp!
::
  TMETACOMPRIM
  DUPDUP #2/
  ZERO_DO (DO)
     get1 SIMPLIFY UNROT
     get1 UNROT
  LOOP
  DROP TMETACOMPRIM
;

* --------------------------------------------------------------------
* Desc:			unrecognized op
* Stack:		meta -> meta
* --------------------------------------------------------------------
LABEL linexpnot!
::
  #1-SWAP					( meta op )
  nonrecurflag TestSysFlag case linexpnotend	( -> meta )
						( meta op )
  enumargs					( meta op # )
  DUP#0=csedrp linexpnotend
  DUP #5= casedrop linexpnotend
						( meta op #args )
  SWAP OBJ>R ZEROSWAP				( meta meta-res #args )
  ZERO_DO (DO)					( meta meta-res )
    psh larg					( meta-res meta1 meta2 )
    OVER TYPESYMB? DUP OBJ>R			( meta-res meta1 meta2 ? )
    IT :: DROP SYMBINCOMP ;
    linSYMBN					( meta-res meta1 symb2 )
    R>OBJ ITE :: symcomp ONE ; SYMBINCOMP	( meta-res meta1 meta2 )
    rot top&					( meta1 meta2&meta-res )
  LOOP
  psh NDROP					( meta-res )
  R>OBJ linexpnotend
;

LABEL linexpnotend
:: SWAP#1+ CKSYMBN Z0SWAP2 ;

* --------------------------------------------------------------------
* Desc:			linearization of arg of a unary op
* Stack:		meta -> meta
* --------------------------------------------------------------------
LABEL linexpmono!
::
  pulldrop linexpmeta!
;

* --------------------------------------------------------------------
* Desc:			linearization of NEG
* Stack:		meta -> meta
* --------------------------------------------------------------------
LABEL linexpneg!
::
  linexpmono!
  Z-1 METAMM2
;

* --------------------------------------------------------------------
* Desc:			linearization of SQ
* Stack:		meta -> meta
* --------------------------------------------------------------------
LABEL linexpsq!
::
  linexpmono!
  dup
  linexpconv!
;	

* --------------------------------------------------------------------
* Desc:			linearization of INV
* Stack:		meta -> meta
* --------------------------------------------------------------------
LABEL linexpexp!
::
  pulldrop				( exparg )
  nonrecurflag SysITE CKSYMBN linSYMBN	( exparg-linearized )
  Z1 TWO
;


LABEL linSYMBN
* meta -> symb
:: 
   ' xEXP EQUALPOSMETA SWAPDROP		( meta # )
   #0=case CKSYMBN
   linexpmeta! SIGMAEXP2ext 
;

* --------------------------------------------------------------------
* Desc:			linearization of INV
* Stack:		meta -> meya
* --------------------------------------------------------------------
LABEL linexpinv!
::				( meta )
  linexpmono!			( meta )
  linexp1/			( meta )
;

LABEL linexp1/
::				( meta )
  DUP #2= NOTcase linexpinvfail
				( meta2 )
  ROT RNEGext ROT FXNDext	( #2 -argexp n d )
  SWAPNDXF ROT			( -argexp n/d #2 )
;

LABEL linexpinvfail
:: SIGMAEXP2ext xINVext Z0SWAP2 ;


* --------------------------------------------------------------------
* Desc:			linearization of arguments of binary op
* Stack:		meta -> meta1 meta2
* --------------------------------------------------------------------
LABEL linexpbin!
::
  plDRPpZparg linexpmeta!
  psh linexpmeta!
;

* --------------------------------------------------------------------
* Desc:			linearization of +
* Stack:		meta -> meta
* --------------------------------------------------------------------
LABEL linexp+!
::
  linexpbin! top&
;

* --------------------------------------------------------------------
* Desc:			linearization of -
* Stack:		meta -> meta
* --------------------------------------------------------------------
LABEL linexp-!
::
  linexpbin!
  Z-1 METAMM2
  top&
;


* --------------------------------------------------------------------
* Desc:		Convolution product of 2 exp-linearization meta
* Stack:	(  meta1, meta2 -> meta )
* --------------------------------------------------------------------
LABEL linexpconv!
::	
  get1

* {{ len2 len1 }}
DEFINE len2 2GETLAM
DEFINE len1 1GETLAM
DEFINE len2! 2PUTLAM
DEFINE len1! 1PUTLAM
  2LAMBIND
  ZERO				( meta1' meta2' meta-res )
  len1 #2/
  ZERO_DO (DO)			( meta1' meta2' meta-res )
     DUP len2 #+ #4+
     DUP ROLL
     SWAP#1- ROLL		( m1'' m2' meta-res arg1 coef1 )
     len2 #2/
     ZERO_DO (DO)		( m1'' m2' meta-res arg1 coef1 )
        ROTDUP INDEX@ #2* #+
        #6+ DUP PICK
        SWAP#1- PICK		( m1'' m2' meta-res' arg1 cf1 # arg2 coef2 )
        4PICK RMULText		( m1'' m2' meta-res' arg1 cf1 # arg2 cf2*cf1 )
        5UNROLL 4PICK x+ext	( m1'' m2' meta-res' cf^2 arg1 cof1 # arg+arg )
        5UNROLL #2+ UNROT	( m1'' m2' meta-res arg1 coef1 )
     LOOP
				( m1'' m2' meta-res arg1 coef1 )
     2DROP
  LOOP
				( meta2' meta-res )
  len2 OVER#2+UNROL
  psh NDROP
  ABND 
;

* --------------------------------------------------------------------
* Desc:			linearization of *
* Stack:		meta -> meta
* --------------------------------------------------------------------
LABEL linexp*!
::
  linexpbin!
  linexpconv!
;

* --------------------------------------------------------------------
* Desc:			linearization of /
* Stack:		meta -> meta
* --------------------------------------------------------------------
LABEL linexp/!
::
  linexpbin!				( meta1 meta2 )
  linexp1/				( meta1 meta2-inv )
  linexpconv!				( meta )
;

* --------------------------------------------------------------------
* Desc:			linearization of ^
* Stack:		meta -> meta
* --------------------------------------------------------------------
LABEL linexp^!
::
  pulldrop larg				( meta-arg meta-exp )
  CKSYMBN OVER#2+UNROLL			( exp meta-arg )
  linexpmeta!				( exposant meta-arg )
  DUP #2= casedrop linexp^k
					( exposant meta-arg )
  get1 OBJPOSINT?			( meta-arg exposant ? )
  NOTcase linexp^fail
					( meta-arg exposant )
  Z># DUP#1= caseDROP
					( meta-arg exposant )
  OBJ>R {}N R>OBJ			( list1 #exposant )
  SWAPDUP 1LAMBIND			( #exposant list1 )
  INNERCOMP get1			( meta #exposant )
  ONE_DO (DO)				( meta-res )
     1GETLAM INNERCOMP
     linexpconv!
  LOOP
  ABND
;

LABEL linexp^k
::					( k argexp coefexp )
  3PICK x^ext				( k argexp coefexp^k )
  UNROT RMULText			( coefexp^k k*argexp )
  SWAP TWO				( k*argexp coefexp^k #2 )
;

LABEL linexp^fail			
::					( meta-arg exposant )
  OBJ>R SIGMAEXP2ext			( arg )
  R>OBJ x^ext Z0SWAP2		( arg^exp 1 #2 )
;

**********************************************************************
* Desc:		Back conversion from arg_exp/coef_meta to symbolic
* Stack:	( Meta --> symb )
**********************************************************************
NULLNAME SIGMAEXP2ext
::
  #2/ Z0ONE ROT
  ZERO_DO (DO)			( pole1, coef1, ..., polen, coefn, meta )
    get1 SYMBINCOMP metasimp	( pole1 coef1 ... polen meta meta-coefn )
    UobROT SYMBINCOMP 
    addtEXPNO-CK 		( pole1 coef1 ... meta meta-coefn meta-exp )
    MetaMul MetaAdd 
  LOOP
  CKSYMBN
; 

**********************************************************************
* Desc:		Tchebycheff polynomial.
* Stack:	( zint --> P )
* Note:		If zint>0 then 1st kind, if <0 then second kind.
**********************************************************************
NULLNAME TCHEBext
::
  DUP ZIsNeg? SWAP ZABS Z>#		( flag # )
  DUP#0=csedrp    :: ITE Z0 Z1 ;
  DUP#1= casedrop :: ITE Z1 INTERNALX ;
  SWAP ITE
    :: Z0 Z1 ;
    :: Z1 INTERNALX ;
  ROT 
  ONE_DO (DO)				( T_k-1 T_k )
     DUP Z2 RMULText
     CKINNERCOMP
     Z0 SWAP#1+ {}POLY
     ROT RSUBext
  LOOP
  SWAPDROP
;

**********************************************************************
* Desc:		Expand SIN(x+/-y)
* Stack:	( Meta --> Meta' )
* Note:		Takes binary operator x+/x- from runstream
**********************************************************************
LABEL SINEXPA1
::
  plDRPpZparg			( meta2, meta1 )
  dup METASINEXPA
  rotdup METACOSEXPA
  rottop& addt*
  unrot METASINEXPA
  swap METACOSEXPA
  top&addt* top&
  'R SWAP#1+
;

**********************************************************************
* Desc:		Expand SIN
* Stack:	( symb --> symb' )
**********************************************************************
NULLNAME SINEXPA
::
  SYMBINCOMP
  METASINEXPA
  CKSYMBN
;

**********************************************************************
* Desc:		Expand SIN
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME METASINEXPA
::
  M1st+case SINEXPA+
  M1st-case SINEXPA-
  M1st*case SINEXPA*
  M1st/case SINEXPA/
  addtSIN
;

**********************************************************************
* Desc:		Expand SIN(x+y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME SINEXPA+
::
  SINEXPA1 x+
;

**********************************************************************
* Desc:		Expand SIN(x-y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME SINEXPA-
::
  SINEXPA1 x-
;

**********************************************************************
* Desc:		Expand SIN(x*y)
* Stack:	( Meta --> Meta' )
* Note:		expands if x or y be an integer
**********************************************************************
NULLNAME SINEXPA*
::
  plDRPpZparg
  METAINT?     case SINEXPA*1
  psh METAINT? case SINEXPA*1
  top&addt* DISTRIB* 
  NOTcase addtSIN
  METASINEXPA
;

LABEL SINEXPA/
::
  DISTRIB/ NOTcase addtSIN
  METASINEXPA
;

**********************************************************************
* Desc:		Expand SIN(x*y)
* Stack:	( Meta2, Meta1 --> Meta' )
* Note:		Meta1 is assumed to be an integer
**********************************************************************
NULLNAME SINEXPA*1
::				( meta-x, n, #1 )
  DROPDUP ZABS QNeg TCHEBext	( meta-x n Un )
  SWAP OBJ>R OBJ>R
  dup METASINEXPA
  psh METACOSEXPA
  CKSYMBN R>OBJ			( sin-meta-x cos Un )
  ZEROSWAP ROT ONE{}N VAL2ext	( sin-meta-x, res )
  symcomp INNERCOMP top&addt*
  R>OBJ ZIsNeg? NOT?SEMI metaneg
;

**********************************************************************
* Desc:		Expand COS(x+/-y)
* Stack:	( Meta --> Meta' )
* Note:		Takes x+ or x- from runstream
**********************************************************************
LABEL COSEXPA1
::
  plDRPpZparg			( meta2, meta1 )
  dup METACOSEXPA
  rotdup METACOSEXPA
  rottop& addt*
  unrot METASINEXPA
  swap METASINEXPA
  top&addt* top&
  'R SWAP#1+
;

**********************************************************************
* Desc:		Expand COS
* Stack:	( symb --> symb' )
**********************************************************************
NULLNAME COSEXPA
::
  SYMBINCOMP
  METACOSEXPA
  CKSYMBN
;

**********************************************************************
* Desc:		Expand COS
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME METACOSEXPA
::
  M1st+case COSEXPA+
  M1st-case COSEXPA-
  M1st*case COSEXPA*
  M1st/case COSEXPA/
  addtCOS
;

**********************************************************************
* Desc:		Expand COS(x+y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME COSEXPA+
::
  COSEXPA1 x-
;

**********************************************************************
* Desc:		Expand COS(x-y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME COSEXPA-
::
  COSEXPA1 x+
;

**********************************************************************
* Desc:		Expand COS(x*y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME COSEXPA*
::
  plDRPpZparg
  METAINT?     case COSEXPA*1
  psh METAINT? case COSEXPA*1
  top&addt* DISTRIB* 
  NOTcase addtCOS
  METACOSEXPA
;

LABEL COSEXPA/
::
  DISTRIB/
  NOTcase addtCOS
  METACOSEXPA
;

**********************************************************************
* Desc:		Expand COS(x*y)
* Stack:	( meta2, meta1 --> Meta' )
* Note:		meta1 represents an integer
**********************************************************************
NULLNAME COSEXPA*1
::					( meta-x, n, #1 )
  DROP ZABS TCHEBext			( meta-x Un  )
  OBJ>R METACOSEXPA
  CKSYMBN R>OBJ
  ZERO SWAPROT ONE{}N VAL2ext
  symcomp INNERCOMP
;

**********************************************************************
* Desc:		Expand EXP
* Stack:	( meta -> meta )
* Note:		Takes a binary operator from the runstream
**********************************************************************
LABEL EXPEXPABIN
::				( meta [runstream op] -> meta )
  plDRPpZparg			( meta2, meta1 )
  METAEXPEXPA psh METAEXPEXPA	( meta1', meta2' )
  top& 'R SWAP#1+
;

**********************************************************************
* Desc:		Expand EXP
* Stack:	( symb --> symb' )
**********************************************************************
NULLNAME EXPEXPA
::
  SYMBINCOMP				( meta -> meta )
  METAEXPEXPA
  CKSYMBN
;

**********************************************************************
* Desc:		Expand EXP
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME METAEXPEXPA
::
  M1st+case EXPEXPA+
  M1st-case EXPEXPA-
  M1st*case EXPEXPA*
  M1st/case EXPEXPA/
  M1stNEGcase EXPEXPANEG
  addtEXP
;

**********************************************************************
* Desc:		Expand EXP(x+y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME EXPEXPA+
::
  EXPEXPABIN x*
;

**********************************************************************
* Desc:		Expand EXP(x-y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME EXPEXPA-
::
  EXPEXPABIN x/
;

**********************************************************************
* Desc:		Expand EXP(x*y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME EXPEXPA*
::
  plDRPpZparg			( meta1, meta2 )
  METAINT?     case EXPEXPA*1
  psh METAINT? case EXPEXPA*1
  top&addt* DISTRIB*			( meta flag )
  NOTcase addtEXP
  METAEXPEXPA
;


**********************************************************************
* Desc:		Expand EXP(x/y)
* Stack:	( Meta --> Meta' )
* Note:		distribute division by y if possible
**********************************************************************
LABEL EXPEXPA/
::
  DISTRIB/			( meta flag )
  NOTcase addtEXP
  METAEXPEXPA
;

**********************************************************************
* Desc:		Expand EXP(-x)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME EXPEXPANEG
::
  pulldrop METAEXPEXPA
  addtINV
;

**********************************************************************
* Desc:		Expand EXP(x*y)
* Stack:	( Meta2, meta1 --> Meta' )
* Note:		meta1 represents an integer
**********************************************************************
NULLNAME EXPEXPA*1
::
  psh addtEXP
  psh top& addt^
;


**********************************************************************
* Desc:		Expand LN
* Stack:	( meta [runstream op] -> meta )
* Note:		takes a binary op from runstream
**********************************************************************
LABEL LNEXPABIN
::
  plDRPpZparg			( meta2, meta1 )
  METALNEXPA psh METALNEXPA	( meta1', meta2' )
  top& 'R SWAP#1+
;

**********************************************************************
* Desc:		Expand LN
* Stack:	( symb --> symb' )
**********************************************************************
NULLNAME LNEXPA
::
  SYMBINCOMP			( meta -> meta )
  METALNEXPA
  CKSYMBN
;

**********************************************************************
* Desc:		Expand LN
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME METALNEXPA
::
  M1st*case LNEXPA*
  M1st/case LNEXPA/
  M1st^case LNEXPA^
  M1stINVcase addtLNINV
  AEQ1stcase x\v/ addtLNv
  AEQ1stcase xSQ addtLNSQ
  AEQ1stcase xABS METALNEXPAABS
  addtLN
;

LABEL METALNEXPAABS
::
  pulldrop 
  lnabsflag TestSysFlag case METALNEXPA
  lnabsflag SetSysFlag METALNEXPA lnabsflag ClrSysFlag
;

LABEL addtLNINV
::
  pulldrop METALNEXPA addtNEG
;

LABEL addtLNv
::
  pulldrop METALNEXPA meta/2
;

LABEL addtLNSQ
::
  pulldrop METALNEXPA meta2*
;

**********************************************************************
* Desc:		Expand LN(x*y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME LNEXPA*
::
  LNEXPABIN x+
;

**********************************************************************
* Desc:		Expand LN(x/y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME LNEXPA/
::
  LNEXPABIN x-
;

**********************************************************************
* Desc:		Expand LN(x^y)
* Stack:	( Meta --> Meta' )
**********************************************************************
NULLNAME LNEXPA^
::
  plDRPpZparg
  METALNEXPA
  top& 'x* SWAP#1+
;

**********************************************************************
* Desc:		Alternate trig operator and coefficient
* Stack:	( symb --> Meta )
**********************************************************************
NULLNAME LINEXPA
::
  DUPTYPESYMB? NOTcase %1SWAP2
  INNERCOMP lintrigmeta!
;

* --------------------------------------------------------------------
* Desc:		Alternate trig operator and coefficient
* Stack:	( meta -> meta' )
* Note:		meta represents a symbolic object
* --------------------------------------------------------------------
LABEL lintrigmeta!
CODE
lintrigmeta	GOSBVL	=SAVPTR
		C=DAT1	A
		CD1EX
		D1=D1+	5
		A=DAT1	A
		B=A	A
		CD1EX
* B= number of elements of meta
* tests if one of them is SIN or COS
		LC(5)	=xCOS
		D=C	A
		LC(5)	=xSIN
lintriglp	B=B-1	A
		GOC	lintrignf
		D1=D1+	5
		A=DAT1	A
		CDEX	A
		?A=C	A
		GOYES	lintrigop
		CDEX	A
		?A#C	A
		GOYES	lintriglp
lintrigop	GOSBVL	=GETPTR
		D1=D1+	5
		A=DAT1	A
		D1=D1-	5
		LC(5)	=x+
		?A#C	A
		GOYES	lintrig1
		A=PC
		LC(5)	(lintrig+!)-(*)
lintrigbr	A=A+C	A
		PC=(A)
lintrignf	GOSBVL	=GETPTR
		GOTO	lintrignot
lintrig1	LC(5)	=x-
		?A#C	A
		GOYES	lintrig2
		A=PC
		LC(5)	(lintrig-!)-(*)
		GOTO	lintrigbr
lintrig2	LC(5)	=x*
		?A#C	A
		GOYES	lintrig25
		A=PC
		LC(5)	(lintrig*!)-(*)
		GOTO	lintrigbr
lintrig25	LC(5)	=x^
		?A#C	A
		GOYES	lintrig3
		A=PC
		LC(5)	(lintrig^!)-(*)
		GOTO	lintrigbr
lintrig3	LC(5)	=x/
		?A#C	A
		GOYES	lintrig4
		A=PC
		LC(5)	(lintrig/!)-(*)
		GOTO	lintrigbr
lintrig4	LC(5)	=xCOS
		?A#C	A
		GOYES	lintrig5
		A=PC
		LC(5)	(lintrigcos!)-(*)
		GOTO	lintrigbr
lintrig5	LC(5)	=xNEG
		?A#C	A
		GOYES	lintrig6
		A=PC
		LC(5)	(lintrigneg!)-(*)
		GOTO	lintrigbr
lintrig6	LC(5)	=xSIN
		?A#C	A
		GOYES	lintrig7
		A=PC
		LC(5)	(lintrigcos!)-(*)
		GOTO	lintrigbr
lintrig7	LC(5)	=xSQ
		?A#C	A
		GOYES	lintrignot
		A=PC
		LC(5)	(lintrigsq!)-(*)
		GOTO	lintrigbr
lintrignot	A=PC
		LC(5)	(lintrignot!)-(*)
		GOTO	lintrigbr
ENDCODE

* --------------------------------------------------------------------
* Desc:		COS(arg) or SIN(arg) -> arg COS #2 or arg SIN #2
* Stack:	( meta -> meta )
* --------------------------------------------------------------------
LABEL lintrigcos!
::
  SYMBN Z1 TWO
;

* --------------------------------------------------------------------
* Desc:
* Stack:		( meta -> meta-trig )
* --------------------------------------------------------------------
LABEL lintrigneg!
::
  pulldrop
  lintrigmeta!
  Z-1 METAMM2
;

* --------------------------------------------------------------------
* Desc:
* Stack:		meta -> meta-trig
* --------------------------------------------------------------------
LABEL lintrigsq!
::
  pulldrop
  lintrigmeta!
  dup
  lintrigconv!
;

* --------------------------------------------------------------------
* Desc:
* Stack:		meta -> meta-trig
* --------------------------------------------------------------------
LABEL lintrignot!
::
  CKSYMBN %1SWAP2
;

LABEL %1SWAP2
:: %1 SWAP TWO ;
	
* --------------------------------------------------------------------
* Desc:			Split meta with top binary op
* Stack:		meta -> meta-trig meta-trig
* --------------------------------------------------------------------
LABEL lintrigbop!
::
  plDRPpZparg
  lintrigmeta!
  psh
  lintrigmeta!
;

* --------------------------------------------------------------------
* Desc:
* Stack:		meta -> meta-trig
* --------------------------------------------------------------------
LABEL lintrig+!
::
  lintrigbop!
  top&
  TMETACOMPRIM
;

* --------------------------------------------------------------------
* Desc:
* Stack:		meta -> meta-trig
* --------------------------------------------------------------------
LABEL lintrig-!
::
  lintrigbop!
  Z-1 METAMM2
  top&
  TMETACOMPRIM
;

* --------------------------------------------------------------------
* Desc:
* Stack:		meta -> meta-trig
* --------------------------------------------------------------------
LABEL lintrig/!
::
  lintrigbop!				( meta1 meta2 )
  MTRIG2SYMB xINVext 			( meta1 1/coef2 )
  METAMM2				( meta1' )
;

* --------------------------------------------------------------------
* Desc:
* Stack:		meta -> meta-trig
* --------------------------------------------------------------------
LABEL lintrig^!
::
  pulldrop larg				( meta-arg meta-exp )
  CKSYMBN OVER#2+UNROLL			( exp meta-arg )
  lintrigmeta!				( exposant meta-arg )
  get1 OBJPOSINT?			( meta-arg exposant ? )
  NOTcase lintrig^fail
					( meta-arg z-exposant )
  num#1=case DROP			( -> meta-arg )
					( meta-arg z-exposant )
  OBJ>R {}N R>OBJ			( arg z )
  OVER LENCOMP #2= NOTcase lintrig^not2

* if length=2 linearization using Tchebycheff polynomials

  SWAPINCOMP DROP			( z sin/cos coef )
  3PICK x^ext UNROT			( coef^z z sin/cos )
  num#1=case :: SWAPDROPSWAP TWO ;
  EXPR> SWAPDROP			( coef^z z arg [xSIN | xCOS] )
  Z1 OVER EQ: xCOS
  ITE Z1 Z-1 TWO{}POLY	 		( coef^z, z, arg, xSIN | xCOS, {} )
  4PICK RPext
  4PICK Z4 ZMod Z># TWO #/		( coef^z z arg oper {} #rem #quot )
  #1= IT
    ::					( .., xSIN | xCOS, {}, # rem )
      3PICK EQ: xCOS ?SEMI
      SWAP RNEGext SWAP
    ;
				( coef^z z arg oper {} #rem )
  Z2 6ROLL Z1 QSub RPext	( coef^z arg oper {} #rem 2^z-1 )
  4ROLL EQ: xCOS		( coef^z arg {} #rem 2^z-1 flag )
  3PICK #0= IT DROPTRUE		( coef^z arg {} #rem 2^z-1 flag' )
  5ROLL 			( coef^z {} #rem 2^z-1 flag arg )

  ( lam4=#rem, lam3= 2^z-1, lam2=flag, lam1=arg )

* {{ coefficient liste #rem puissance2 sincos argument }}
DEFINE coefficient 6GETLAM
DEFINE liste 5GETLAM
DEFINE #rem 4GETLAM
DEFINE puissance2 3GETLAM
DEFINE sincos 2GETLAM
DEFINE argument 1GETLAM
DEFINE coefficient! 6PUTLAM
DEFINE liste! 5PUTLAM
DEFINE #rem! 4PUTLAM
DEFINE puissance2! 3PUTLAM
DEFINE sincos! 2PUTLAM
DEFINE argument! 1PUTLAM
  ' NULLLAM SIX NDUPN DOBIND

  liste INNERCOMP		(    cn, ..., c0, c-1, ..., c-n, #2n+1 )
				( OR cn, ..., c1, c-1, ..., c-n, #2n )
  #2/ DUPDUP #2+UNROLL NDROP
  #1+ #rem #- 			(    cn, ..., c0, #n+1 )
				( OR cn, ..., c1, #n )
  ZEROSWAP
  ZERO_DO (DO)			( cn, .., ck, meta )
     INDEX@ #2* #rem #+ 
     lintrig^1
     SWAP#1+			( cn, .., ck, meta&arg )
     get1
     coefficient RMULText
     puissance2			( cn..ck+1 meta&arg ck*coefficient puissance2 )
     4GETLAM INDEX@ #+
     #0=ITE Z2 Z1
     RMULText 			( cn..ck+1 meta&arg ck*coeff puiss2' )
     RDIVext SWAP#1+		( cn..ck+1 meta&arg&ck )
  LOOP
  ABND
;

LABEL lintrig^1
( # -> fnc[%*arg] )
::
  DUP#0=csedrp %1
  DUP#1= casedrop lintrig^11
  #>Z argument SYMBINCOMP #1+ addt* ifnegmoveneg
  lintrig^12
;

LABEL lintrig^11
::
  argument SYMBINCOMP lintrig^12
;

LABEL lintrig^12
::
  2GETLAM ITE addtCOS addtSIN SYMBN
;

LABEL lintrig^fail
::					( meta-arg exposant )
  OBJ>R MTRIG2SYMB R>OBJ x^ext		( arg^exposant )
  %1SWAP2				( meta )
;
		
LABEL lintrig^not2
::					( list z-exposant )
  Z>#					( list1 #exposant )
  SWAPDUP 1LAMBIND INNERCOMP get1
  ONE_DO (DO)				( meta-res )
    1GETLAM INNERCOMP
    lintrigconv!
  LOOP
  ABND
;



* --------------------------------------------------------------------
* Desc:
* Stack:		meta -> meta-trig
* --------------------------------------------------------------------
LABEL lintrig*!
::
  lintrigbop!			( liste1, liste2 )
  lintrigconv!
;

* --------------------------------------------------------------------
* Desc:			Convolution of 2 meta-trig
* Stack:		( meta1-trig, meta2-trig -> meta-trig )
* --------------------------------------------------------------------
LABEL lintrigconv!
::
  {}N
  1LAMBIND
  ZEROSWAP #2/
  ZERO_DO (DO)			( p1, c1.., pn, cn, meta-resultat )
     get1 rotomo		( p1, c1.., meta-resultat, cn, pn )
     convl!			( p1, c1.., meta-resultat, meta )
     top&			( p1, .., pn-1, cn-1, meta-resultat ) 
  LOOP
  ABND
  TMETACOMPRIM
; 

LABEL TMETACOMPRIM
* like METACOMPRIM but returns 1 0 #2 for a 0 result
::
  FALSE 1LAMBIND METACOMPRIM0 ABND
  DUP#0= NOT?SEMI
  DROP %1 Z0 TWO
;

* --------------------------------------------------------------------
* Desc:
* Stack:	( cn, pn -> meta=cn*liste2 convolee avec pn )
* --------------------------------------------------------------------
LABEL convl!
::
  num#1=case
    :: DROP 1GETLAM INNERCOMP METAMULMULT ;
						( cn, pn )
  1GETLAM INNERCOMP
  get1 SWAPDUP #2/
  ZERO_DO (DO)				( q1, d1, ..., qn, dn, pn, #2n )
     get1				( d1, ..., qn, dn, pn, #2n, q1 )
     convl1!
  LOOP
					( cn, q1, d1, ..., qn, dn, pn, #2n )
  SWAPDROP METAMULMULT
;

* --------------------------------------------------------------------
* Desc:
* Stack:
* --------------------------------------------------------------------
LABEL convl1!
::
  num#1=case
    :: DROP OVERSWAP			( d1 ... qn dn pn pn #2n )
       get1 UNROT
    ;

  SWAP get1			( q2 d2 ... qn dn pn q1 #2n d1 )
  Z2 RDIVext DUP		( q2 d2 ... qn dn pn q1 #2n d1/2 d1/2 )
  4ROLL 5PICK			( q2 d2 ... qn dn pn #2n d1/2 d1/2 q1 pn )
				( d1 d1 q1 pn -> +-d1 p1.1 +-d1 p1.2 )
  EXPR> SWAPDROP		( d1 d1 q1 arg[pn] op[pn] )
  ROT EXPR> SWAPDROP ROTSWAP	( d1 d1 arg1 arg2 op1 op2 )
  TWO{}N			( d1 d1 arg1 arg2 {} )
  convl2!			( q2 d2 ... qn dn cn #2n d1' q1.1 d1'' q1.2 )

* si last de q1.2 <> SIN/COS -> d1''=d1''*q1.2 et q1.2=1

  trigarrange!
  2SWAP trigarrange!		( q2 d2 ... qn dn cn #2n q1.2 d1'' q1.1 d1' )
  convl3!
;

* --------------------------------------------------------------------
* Desc:
* Stack:
* --------------------------------------------------------------------
LABEL convl2!
::
  { xCOS xCOS } EQUALcasedro convl2_1!
  { xSIN xSIN } EQUALcasedro convl2_2!

  CARCOMP EQ: xCOS IT SWAP		( d1, d1, arg1, arg2 )
  2DUP x-SIMPSIN			( d1, d1, arg1, arg2, arg-arg )
  4UNROLL x+SIMPSIN			( d1, arg-arg, d1, arg+arg )
;

* --------------------------------------------------------------------
* Desc:
* Stack:
* --------------------------------------------------------------------
LABEL convl2_1!
::					( d1 d1 arg1 arg2 )
  2DUP x-SIMPCOS			( d1 d1 arg1 arg2 c_arg-arg )
  4UNROLL x+SIMPCOS			( d1 c_arg-arg d1 c_arg+arg )
;

* --------------------------------------------------------------------
* Desc:
* Stack:
* --------------------------------------------------------------------
LABEL convl2_2!
::					( d1 d1 arg1 arg2 )
  convl2_1!				( d1 c_arg-arg d1 c_arg+arg )
  SWAP RNEGext SWAP			( d1 c_arg-arg -d1 c_arg+arg )
;

LABEL x-SIMPCOS
:: 2SYMBINCOMP MetaSub metasimp addtCOS CKSYMBN ;

LABEL x+SIMPCOS
:: 2SYMBINCOMP MetaAdd metasimp addtCOS CKSYMBN ;

LABEL x-SIMPSIN
:: 2SYMBINCOMP MetaSub metasimp addtSIN CKSYMBN ;

LABEL x+SIMPSIN
:: 2SYMBINCOMP MetaAdd metasimp addtSIN CKSYMBN ;

* --------------------------------------------------------------------
* Desc:
* Stack:
* --------------------------------------------------------------------
LABEL convl3!
::
  num#0=case 2DROP2SWAP
  2SWAP
  num#0=case 2DROP2SWAP
  6ROLL 6ROLL #2+
;

LABEL 2DROP2SWAP
:: 2DROP 2SWAP ;

* --------------------------------------------------------------------
* Desc:		Insure that trigo fnct is 1 or SIN or COS
* Stack:	( coef, fnct trigo -> fnct trigo, coef )
* --------------------------------------------------------------------
LABEL trigarrange!
::
  DUPTYPESYMB? NOTcase trigarrangereal
						( coef fnct )
  DUP LASTCOMP { xNEG xCOS xSIN } ListPos	( coef fnct # )
  DUP#0=csedrp x*%1SWAP				( -> %1 coef*fnct )
  #1= NOTcase SWAP				( -> fnct coef )
						( coef neg_fnct )
  RNEGext trigarrange! RNEGext
;

LABEL x*%1SWAP
::
  x*ext %1 SWAP
;

LABEL trigarrangereal
::						( coef obj )
  DUPTYPEREAL? IT R>Z x*%1SWAP			( -> %1 coef*obj )
;

**********************************************************************
* Desc:		Back conversion of trig-meta to symbolic
* Stack:	( Meta --> symb )
**********************************************************************
NULLNAME MTRIG2SYMB
::				( fnc1 coef1...fncn coefn #2n )
  Z0ONE ROT #2/			( fnc1 coef1...fncn coefn meta-res #2n )
  ZERO_DO (DO)			( fnc1 coef1...fncn coefn meta-res )
    get1 rotomo			( fnc1 coef1...coefn-1 meta-res coef fnc )
    DUPTYPEREAL? IT R>Z		( fnc1 coef1...coefn-1 meta-res coef fnc )
    SWAP SYMBINCOMP metasimp	( fnc1 coef1...coefn-1 meta-res fnc meta-coef )
    get1 SYMBINCOMP		( fnc1 coef1... meta-res meta-coef meta-fnc )
    MetaMul MetaAdd		( fnc1 coef1...coefn-1 meta-res )
  LOOP
				( meta-res )
  CKSYMBN
;

**********************************************************************
* Desc:		Collect logarithms
* Stack:	( symb --> symb' )
**********************************************************************
NULLNAME LNCOLCext
::
  lncolc!
  INCOMPDROP				( reste, ln )
  COLCext CKLN x+ext
;

* --------------------------------------------------------------------
* Desc:
* Stack:	( symb -> { reste arg_ln } )
* --------------------------------------------------------------------
LABEL lncolc!
::
  DUP Z1 TWO{}N SWAP			( {}, symb )
  DUPTYPESYMB? NOTcaseDROP
  DUP' xLN ' EQ POSCOMP #0= caseDROP	( {}, symb )
  EXPR>					( {}, arg1, .., argn, %n, oper )
  SWAP COERCE
  { xLN xNEG x* x+ x- }
  3PICK ' EQ POSCOMP			( {}, arg1, .., argn, op, #n, #pos )
  DUP#0=csedrp lncolc0!
  ROTDROP				( {}, arg1, .., argn, #n, #pos )
  ONE   EQUALcasedro lncolc1!
  TWO   EQUALcasedro lncolc2!
  THREE EQUALcasedro lncolc3!
					( {}, arg1, arg2, #2, # )
  SWAPDROP 4ROLLDROP			( arg1, arg2, # )
  FOUR #= UNROT lncolc!
  SWAP lncolc!				( ?=x+, arg2, arg1 )
  INCOMPDROP ROT INCOMPDROP		( ?=x+, reste1, ln1, reste2, ln2 )
  4ROLLROT 5PICK ITE x+ext x-ext UNROT	( ?=x+, reste, ln1, ln2 )
  4ROLL ITE x*ext x/ext
  TWO{}N
;

* --------------------------------------------------------------------
* Desc:		Calls recursively LNCOLC for all arg of this function
* Stack:
* --------------------------------------------------------------------
LABEL lncolc0!
::					( {}, arg1, .., argn, op, #n )
  DUP#0_DO (DO)
     get1 LNCOLCext symcomp
     UNROT
  LOOP					( {}, arg1, .., argn, op, #n )
  SWAP IDNTLAM? 			( {}, arg1, .., argn, #n, op, ? )
  ITE SIMPUSERFCN :: SWAPDROP EVAL ;	( {} res )
  SWAPDROP Z1 TWO{}N
;

* --------------------------------------------------------------------
* Desc:		LN()
* Stack:
* --------------------------------------------------------------------
LABEL lncolc1!
::				( {}, arg1, #1 )
  DROPSWAPDROP			( arg1 )
  Z0 SWAP TWO{}N
;

* --------------------------------------------------------------------
* Desc:		-LN() -> LN(INV())
* Stack:
* --------------------------------------------------------------------
LABEL lncolc2!
::			( {}, arg1, #1 )
  DROPSWAPDROP		( arg1 )
  lncolc!		( { reste arg_ln } )
  INCOMPDROP
  SWAP RNEGext
  Z1 ROT NDXFext
  TWO{}N		( {} )
;

* --------------------------------------------------------------------
* Desc:		Collect z*LN()-> LN()^z
* Stack:	
* --------------------------------------------------------------------
LABEL lncolc3!
::					( {}, arg1, arg2, #2 )
  DROP					( {}, arg1, arg2 )
  OBJINT? ?SKIP SWAP			( {}, arg1, arg2 )
  OBJINT? NOTcase2DROP			( -> {} )
					( {}, arg1, z )
  ROTDROPSWAP lncolc!			( z, {} )
  INCOMPDROP				( z reste arg_ln )
  SWAP3PICK RMULText			( z arg_ln reste*z )
  UNROTSWAP x^ext TWO{}N		( {} )
;

**********************************************************************
* Desc:		Expand tan[meta]
* Stack:	( meta -> tan[meta] )
**********************************************************************
NULLNAME METATANEXPA
::
  M1st+case TANEXPA+
  M1st-case TANEXPA-
  M1st*case TANEXPA*
  M1st/case TANEXPA/
  addtTAN
;

LABEL TANEXPA*
::
  dup METASINEXPA psh METACOSEXPA MetaDiv
;

LABEL TANEXPA/
::
  DISTRIB/ NOTcase addtTAN
  METATANEXPA
;

LABEL TANEXPA+
::					( a+b )
  plDRPpZparg METATANEXPA psh METATANEXPA	( tan[a] tan[b] )
  dup rotdup				( tan[b] tan[b] tan[a] tan[a] )
  rot MetaAdd unrot psh MetaMul Z1 ONE 
  psh MetaSub MetaDiv			( tan[a]+tan[b]/1-tan[a]tan[b] ) 
;

LABEL TANEXPA-
::					( a-b )
  plDRPpZparg METATANEXPA psh METATANEXPA	( tan[a] tan[b] )
  dup rotdup				( tan[b] tan[b] tan[a] tan[a] )
  rot MetaSub unrot psh MetaMul Z1 ONE 
  psh MetaAdd MetaDiv			( tan[a]-tan[b]/1+tan[a]tan[b] ) 
;

**********************************************************************
* Desc:		Main transcendental expansion program
* Stack:	( symb -> symb )
**********************************************************************
NULLNAME TEXPAext
::	
  { xTAN xSIN xCOS xEXP xLN }
  { METATANEXPA METASINEXPA METACOSEXPA METAEXPEXPA METALNEXPA }
  REPLACE2BY1
;
**********************************************************************
