INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE global.s
INCLUDE keydefs.h

	STITLE	ALG48 Mod Utilities
**********************************************************************
*	ALG48 Library:	Modular arithmetic utilities
*	Added Mod from original pmod.s
**********************************************************************

ASSEMBLE
*---------------------------------------------------------------------
* Modular math preparation
* Stack:	( Q Zn )
* Exit:		sSHORTMOD=1	R4[W]=Zn	D0=->Q
*		sSHORTMOD=0	R4[A]=->Zn	D0=->Q
*---------------------------------------------------------------------

=ModPrepare	GOSBVL	=SAVPTR
		A=DAT1	A		->Zn
		D1=D1+	5
		R4=A	A		R4[A] = ->Zn
		ST=0	sSHORTMOD	Assume Zn is long
		D0=A
		GOSBVL	=SHORTPOP?	A[W]C[S] = Zn
		GOC	+		It was long - continue
		ST=1	sSHORTMOD	It wasn't - flag so and save Zn
		R4=A	W
+		A=DAT1	A
		D0=A			D0 = ->Q
		RTN

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Modular inverse
* Entry:	D0	= ->Z
*		R4[W]	= Zn
* Exit:		A[W]	= Z inv Zn
*		D0	= ->Ztail
* Alters:	A[W] B[W] C[W] D[W] R0[W] R1[W] R2[W] R3[W]
* Stack:	4
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

=ZMODINV	C=R4	W
		R3=C	W		R3[W]=P=N
		GOSUB	ZMODPSHORT	A[W] = Zmod	C[W] = Zn
		R2=A	W		R2[W] = Z
		?A=0	W
		RTNYES			ModInv(1,1) = 0
		A=A-1	A
		?A#0	W
		GOYES	+
		A=A+1	A		ModInv(1,any) = 1
		RTN
+		A=A+1	A
		GOSBVL	=SHORTGCD	A[W] = GCD(Z,Zn)
		A=A-1	W
		?A=0	W
		GOYES	+
		A=0	W		ModInv(n*x,m*x) = 0
		RTN
+		R1=A	W		R1[W]=S=0
		A=A+1	A
		R0=A	W		R0[W]=R=1

ZMODINVLOOP	A=R2	W		A[W]=Z
		C=R3	W		C[W]=P
		?C=0	W
		GOYES	ZMODINVOK	Until P=0
		GOSBVL	=BDIV		A[W]=Q=Z/P
		C=A	W		C[W]=Q
		AR4EX	W		R4[W]=Q
		B=A	W		B[W]=N
		A=R1	W		A[W]=S
		GOSBVL	=SHORTMULMOD	A[W]=C[W]=SQ mod N
		C=R0	W		C[W]= R
		C=C-A	W		C[W]=R-SQ mod N
		GONC	+
		C=C+B	W		C[W]=T=R-SQ+N mod N
+		CR1EX			S=T
		R0=C	W		R=S
		A=R3	W		A[W]=P
		C=B	W
		CR4EX	W		C[W]=Q	R4[W]=N
		GOSBVL	=SHORTMULMOD	A[W]=C[W]=PQ mod N
		C=R2	W		C[W]=Z
		C=C-A	W		C[W]=Z-PQ mod N
		GONC	+
		C=C+B	W		C[W]=T=Z-PQ+N mod N
+		CR3EX	W		P=T
		R2=C	W		Z=P
		GOTO	ZMODINVLOOP
ZMODINVOK	A=R0	W		A[W]=R
		RTN

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Compute symmetric modulo Z1 mod Z2 where Z2 is short
* In:		D0	->Z1
*		C[W]	Z2
* Out:		D0	->Z1tail
*		A[W]	Zmod
*		D[S]	Zmodsign
*		C[W]	Z2
* Alters:	D0 A[W] B[W] C[W] D[S,14,A]
* Stack:	3
* Note:		This always returns the symmetric modulo, not the
*		antisymmetric modulo as the old system RPL code would
*		when Z2 is negative.
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=ZMODSHORT	GOSUB	ZREMSHORT	A[W]=remainder	C[W]=Z2
		GOSUB	ZMODSYM
		?A=0	W		Ignore next if rem is negative
		RTNYES
		B=C	S		Negate modulo of Z1 is negative
		D0=D0-	1
		C=DAT0	S		C[S] = Z1sign
		D0=D0+	1
		CBEX	S
		?B=0	S
		RTNYES
		D=-D-1	S
		RTN

* Little utility to convert MODP to MOD

=ZMODSYM	D=0	S		Assume modulo is positive
		?A=0	W
		RTNYES			It is if remainder is zero
		B=A	W		B[W] = remainder
		B=B+B	W
		GOC	+
		?B<=C	W		If 2*rem <= Zn then is symmetric
		RTNYES
+		A=A-C	W
		A=-A	W
		D=-D-1	S
		RTN
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Compute positive modulo Z1 mod Z2 where Z2 is short
* In:		D0	->Z1
*		C[W]	Z2
* Out:		D0	->Z1tail
*		A[W]	Zmod
*		D[S]	Zmodsign
*		C[W]	Z2
* Alters:	D0 A[W] B[W] C[W] D[S,14,A]
* Stack:	3
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
=ZMODPSHORT	GOSUB	ZREMSHORT	A[W]=remainder	C[W]=Z2
		B=C	S		Negate modulo of Z1 is negative
		D0=D0-	1
		C=DAT0	S		C[S] = Z1sign
		D0=D0+	1
		CBEX	S
		?B=0	S
		RTNYES			rem is result
		?A=0	W		Don't change 0 to Z2 !!!
		RTNYES
		A=A-C	W
		A=-A	W		Z2-rem is result for Z1<0
		RTN

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Compute remainder in Z1/Z2 calculation.
* In:		D0	->Z1
*		C[W]	Z2		Z2 is short and unsigned
* Out:		D0	->Z1tail
*		A[W]	Zrem
*		C[W]	Z2
* Alters:	D0 A[W] B[W] C[W] D[S,A]
* Stack:	2
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

=ZREMSHORT	B=C	W		B[W] = divisor
		D0=D0+	5
		A=DAT0	A		Z1size

		C=C-1	W
		GOC	z1rem1		Handle Z1 REM 0
		C=C-1	W
		GOC	z1rem1		Handle Z1 REM 0
		C=C-1	W
		GOC	z1rem2		Handle Z1 REM 2

		LC(5)	5+16+1
		?A>C	A
		GOYES	zremlong

* The integer in D0 is short - we can do the entire calculation in registers

		GOSBVL	=shortpop	A[W]C[S] = Z1
		CBEX	W		A[W]B[S] = Z1
		GOSBVL	=BDIV		C[W]=divisor
		A=B	W		A[W]=remainder
		RTN

z1rem1		CD0EX
		SETHEX
		C=C+A	A		Skip Z1
		SETDEC
		CD0EX
		C=B	W		C[W] = Z2
		A=0	W		A[W] = remainder = 0
		RTN
z1rem2		D0=D0+	5
		C=DAT0	B		C[0] = low nibble in Z1 (0 for Z0)
		CD0EX
		SETHEX
		C=C+A	A		Skip Z1
		SETDEC
		CD0EX
		D0=D0-	5		->Z1tail
		A=0	W		A[W] = remainder = 0
		?CBIT=0	0
		GOYES	+
		A=A+1	A		A[W] = remainder = 1
+		C=B	W		C[W] = divisor
		RTN

* The integer in D0 is too long to fit into a register, but the
* remainder can be calculated in registers. (See ZQUO for comparison)
* The old code was actually unnecessarily complicated, this new
* version simply starts with a zero remainder, and starts adding
* the nibbles from the dividend. At each phase the new remainder
* is made modulo the divisor. This way there is no need to call
* BDIV, which implies shifting back and forth. For comparison
* see the new ZQUO inner loop for division by an integer shorter
* than 8 nibbles.
		
zremlong	A=A-CON	A,7		A[A]=nibbles-1
		CD0EX
		SETHEX
		C=C+A	A
		SETDEC
		CD0EX
		D0=D0+	6		D- = ->Zsign

		C=0	W		C[W] = remainder = 0

zremlonglp	A=C	S		A[S]C[W] = remainder*10
		CSL	W
		D0=D0-	1		And add the next digit from dividend
		C=DAT0	1		

zremlongmodlp	C=C-B	W		Now perform division without
		GONC	zremlongmodlp	any quotient output
		A=A-1	S
		GONC	zremlongmodlp
		C=C+B	W		Fix back the extra substraction
*		A=0	S		Not needed
		SETHEX
		A=A-1	A		1 less nibble to go
		SETDEC
		GONC	zremlonglp

* Now C[W] contains the final remainder

		D0=D0-	5
		A=DAT0	A		A[A] = Z1len
		CD0EX
		SETHEX
		C=C+A	A
		SETDEC
		CD0EX			D0 = ->Z1tail
		A=C	W		A[W] = remainder
		C=B	W		C[W] = Z2
		RTN

RPL
**********************************************************************

**********************************************************************
* Desc:		Make Z modulo N
* Stack:	( Z Zn --> Z' )
**********************************************************************
NULLNAME Mod
CODE
		GOSUBL	ModPrepare	D0 = ->Z	R4[W] = Z / ->Zn
		?ST=0	sSHORTMOD
		GOYES	ModCheckRPL

		C=R4	W		C[W] = Z
		SETDEC
		GOSUBL	ZMODPSHORT	A[W] = Zmod	C[W] = Z
		SETHEX
		GOVLNG	=Drp2PushNLoop

ModCheckRPL	GOSBVL	=ModRPL:

ENDCODE
::
  2DUP ZNLT? ?SKIP				( Divide if too big )
  :: DUPUNROT QDivRem SWAPDROPSWAP ;
  OVER ZIsNeg? case QAdd			( Add N if negative )
  DROP						( Else just drop N )
;
