	STITLE  Arithmetic with polynomials

INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s

**********************************************************************
* EXPORT:
*
* PARTFRAC	( o -> symb )
*		partial fraction expansion of o with respect to the cur. var
*
* PFext		( o, {}, # -> 
*		  {}, Q, n1, d1, f1, m1, ..., nk, dk, fk, mk, #4k )
*		{} is the list of variables of object
*		# is the type of factorization
*		Q the integer part of the partial fraction decomposition
*		n1 numerator of the 1st factor f1=d1^m1
*		d1 1st factor (without multiplicity)
*		m1 multiplicity
*
* REGCDext	( a, b -> d, u, v )
*		a, b may be fractions, d="gcd[a,b]", au+bv=d
*
* EGCDext	( Qa, Qb -> Qd, Qu, Qv )
*		Qd=gcd[Qa,Qb], Qa*Qu+ Qb*Qv=Qd )
*
* DERext	( P -> P' )
*		polynomial derivation
*
* INTEGRext	( P -> { an/n ... a1/2 a0 0 } )
*		polynomial integration
*		Warning: the result is not a Q object
*
* LRDMext	( {} # -> [] )
*		Adds 0 to the left of the list to get a vector with length
*		#+1 (looks like a polynomial of degree #)
*
* RRDMext	( {} # -> {} )
*		Like LRDMext but adds 0 to the right and creates a polynomial
*
* DEGREext	( {} -> # )
*		( obj -> #0 )
*		returns the degree of a list polynomial
*
* HORNext	( P, r -> P1, r, P[r] )
*		Horner evaluation of a polynomial P at a scalar r
*		P1=P[X] div [X-r]
*
* MHORNext	( P,r -> P1, r, P[r] )
*		Horner evaluation if P is a polynomial with 
*		matrix coefficients, used by diagonalization
*
* PTAYLext	( P, r -> P1 )
*		P1(X)=P(X+r)
*******************************************************

**********************************************************************
* Desc:		partial fraction expansion of o with respect to the cur. var
* Stack:	( o -> symb )
**********************************************************************
NULLNAME PARTFRAC
::
*  VX>LVARext
  LVARext VX!
* end change
  DUPZERO 2LAMBIND	( symb lvar )
  INPARTFRAC		( symb )
  ABND 
;

**********************************************************************
* Desc:		partial fraction expansion of o
* Stack:	( o, list -> symb )
* Note:		lvar must be bound to LAM2, 
*		list is =lvar if o is in external format
*		list is NULL{} if o is still in internal format
**********************************************************************
NULLNAME INPARTFRAC
::			( o lvar )
  SIX PFext		( lvar quo num1 deno1 f1 m1..numk denok fk mk #4k )
  PARTFRACRAT		( lvar quot resultat )
  ROTDROPSWAP		( res quot )
  ( * check if quot has non cst denominator * )
  DUP FXNDext 		( res q q-n q-d )
  NOTLISTcase INPARTOUT	( -> symb )
  ROTDROP SWAPDUP	( res q-d q-n q-n )
  ( * extract constant part from quot * )
  :: dLISTcase LASTCOMP
     DUPTYPECOL? case DROPZ0
  ;
  DUPUNROT		( res q-d q-n.cst q-n q-n.cst )
  ONE>POLY RSUBext	( res q-d q-n.cst q-n.not_cst )
  3PICK RDIVext		( res q-d q-n.cst q.not_cst )
  CKRISCH13 VAL1	( res q-d q-n.cst add_to_res )
  4ROLL EvalNoCKx+	( q-d q-n.cst res )
  SWAPROT CARCOMP RDIVext ( res q-n.cst/q.d )
  2GETLAM CDRCOMP 2PUTLAM ( res q-n.cst/q.d )
  NULL{} INPARTFRAC	( res symb )
  EvalNoCKx+		( symb )
;

LABEL INPARTOUT	
::				( res q q-n q-d )
  2DROP CKRISCH13 		( res q )
  VAL1M get1 SYMBINCOMP metaadd CKSYMBN
;

LABEL CKRISCH13
::
  DUPTYPESYMB? NOT?SEMI
  DUP FXNDext
  CK&DISPATCH1
  seco 2DROP
  #80  2DROP
  any :: 2DROP RISCH13 ;
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME PARTFRACRAT
::
  Z0 SWAP 
  DUP#0=csDROP
  FOUR #/ SWAPDROP
  ZERO_DO (DO)		( lvar quo num1 deno1 f1 m1...numk denk fk mk res )
     5UNROLL
     ZEROSWAP
     %ABSCOERCE		( lvar quo num1 f1 m1...res numk denk fk resinter #mk )
     ZERO_DO (DO)		( numk denk fk resinter )
       4ROLL 3PICK PSEUDODIV	( denk fk resinter a quot reste )
       ROT 6ROLL RMULText	( fk resinter quot reste newdenok )
       DUPUNROT RDIVext		( fk resinter quot newdenok reste )
       4ROLLSWAP		( fk quot newdenok resinter reste )
       5PICK SWAP VAL1M		( fk quot newdenok resinter fk reste-meta )
       get1 VAL1M 		( fk quot newdenok resinter reste fk )
       ISTOP-INDEX UNCOERCE
       meta^ metadiv 		( fk quot newdenok resinter reste/fk^ )
       get1 DUP#0= ITE_DROP
         ::
           SYMBINCOMP psh metaadd
         ;
       SYMBN			( fk quot newdenok resinter )
       4ROLLSWAP
     LOOP
			( lvar quo num1 f1 m1...res numk denk fk resinter )
     4UNROLL3DROP
     x+ext
  LOOP
			( lvar quot res )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME PFext
::				( symb, lvar, # )
  RFACT2ext			( lvar, n/d, meta-fact )

* d = produit des facteurs de meta-fact

  DUP#0= ?SEMI
  get1 FXNDext
  DUPUNROT NEWDIVext		( lvar, meta-fact, deno, quot, num=rest )
  SWAP4PICK #4+ UNROLL		( lvar, quot, meta-fact, deno, num=rest )
  SWAP				( lvar, quot, meta-fact, num, deno )

* elimine tous les facteurs constants et les multiplie a num et deno

  TWO psh 
  Z1 				( lvar quot num pden a #2 f1...fk mk #2k mult )
  SWAPDUP #2/
  ZERO_DO (DO)			( ..., f1, ..., fk, mk, mult, #2k )
     get1 OVER #2+ROLL		( ..., f2, ..., fk, mk, mult, #2k, f1, m1 )
     OVER DUPTYPELIST?
     ITE LENCOMP DROPONE	( ..., f2, ..., fk, mk, mult, #2k, f1, m1, # )
     #1= ITE
       ::
	 EXACT? NOTcase NONEXACTERR
         DUPUNROT
         %ABSCOERCE RP# 	( ..., mult, #2k, m1, f1^|m1| )
         4ROLLSWAP ROT 		( ..., #2k, m1, mult, f1^|m1| )
         %0> ITE RMULText RDIVext
         SWAP #2- 
       ;
       2SWAP 
  LOOP
			( lvar quot num pden #2 f1 m1... fk mk mult #2k )

* deno = 1/mult* produit des fi^mi

  DUP#0=csedrp
   ::				( lvar, quot, num, deno, #2, mult, #0 ) 
      4DROP ZERO 		( -> lvar, quot, #0 )
   ;
  DUP #2= casedrop
   ::				( lvar, quotient, num, den, #2, f1, m1, mult )
     4ROLLDROP 4ROLLDROP	( lvar, quotient, num, f1, m1, mult )
     4ROLL RMULText FXNDext 	( lvar, quotient, f1, m1, num1, deno1 )
     2SWAP FOUR			( lvar, quotient, num1, den1, f1, m1, #4 )
   ;
			( lvar quot num pden #2 f1... fk mk mult #2k )
  #1+ psh		( lvar quot f1...fk mk mult #2k+1 pnum pden #2 )
  DROP 4PICK RMULText
  4ROLLROT RMULText
  FXNDext

* {{ numn numd }}
DEFINE numn 2GETLAM
DEFINE numd 1GETLAM
DEFINE numn! 2PUTLAM
DEFINE numd! 1PUTLAM

  2LAMBIND		( lvar quot f1...fk mk #2k+1 pden )
  SWAP #1-DUP #2/	( lvar quot f1...fk mk pden num #2k #k )
  ONE_DO (DO)		( lvar quotient f1 m1... fk mk deno #termes ) 
     get1 OVER #2+ROLL	( lvar quotient f2 m2... deno #termes f1 m1 ) 

* deno = deno/(fk^mk)* (fk^mk ) = newdeno * (fk^mk )
* Bezout gcd = newdeno * a_den + fk^mk * a_num
* num-n/num-d= num-n/[num-d*gcd] * gcd
* num /deno = newnum / newdeno + A / (fk^mk)
* avec newnum = a_num *num-n / num-d*gcd ( modulo newdeno )
* et A = a_den * num-n / num-d * gcd ( modulo fk^mk )

    2DUP %ABSCOERCE RP#	( lvar quo f2 m2...deno #terms f1 m1 f1^m1 ) 
    5ROLL OVER QUOText	( lvar quo f2 m2..#terms f1 m1 f1^m1 den/f1^m1 )
    2DUP EGCDext	( lvar quo f2 m2..#terms f1 m1 f1^m1 den/f1^m1  gcd a_f a_den )
    numd		( lvar quo f2 m2..#terms f1 m1 f1^m1 den/f1^m1 gcd a_f a_den num-d )
    4ROLL RMULText 	( lvar quo f2 m2..#terms f1 m1 f1^m1 den/f1^m1 a_f a_den num-d*gcd )
    SWAP numn RMULText	( lvar quo f2 m2..#terms f1 m1 f1^m1 den/f1^m1 a_f num-d*gcd a-den*num-n )
    5ROLL PSEUDODIV
    SWAPDROP		( lvar quo f2 m2..#terms f1 m1 den/f1^m1 a_f num-d*gcd a a-den*num-n )
    SWAP3PICK RMULText	( lvar quo f2 m2..#terms f1 m1 den/f1^m1 a_f num-d*gcd A=a-den*num-n mod f1^m1 num-d*gcd )
    7UNROLL 7UNROLL
    SWAP		( lvar quo f2 m2...#terms A num-d*gcd f1 m1 newdeno num-d*gcd a_f )
    numn RMULText
    3PICK PSEUDODIV
    SWAPDROP		( lvar quo f2 m2...#terms A num-d*gcd f1 m1 newdeno num-d*gcd a num-n*a_f )
    numn! RMULText
    numd!		( lvar quo f2 m2...#terms A f1 m1 newdeno )
    6ROLL #2+
  LOOP
			( lvar quo f1 m1..numk denok fk mk deno #termes ) 
  SWAPDROP
  numn OVER#2+UNROL
  1GETABND OVER#2+UNROL
  #2+			( lvar quotient num1 deno1 f1 m1..numk denok fk mk #4k )
;


**********************************************************************
* Desc:		Bezout for integers
* Stack:	( a, b  -> d, u, v )
* Note: 	d=au+bv=gcd(a,b) 
**********************************************************************
NULLNAME IEGCDext
::					( a, b  )
					( a, b )
  stepflag TestSysFlag NOTcase RIEGCD
  CK&DISPATCH1
  zintzint IEGCDSTEP
  any RIEGCD
;

LABEL RIEGCD
::
  num#0=case :: DROP Z1Z0 ;
  DUPUNROT IDIV2			(  B, A div B, A mod B )
  ROTSWAP				( A div B, B, A mod B )
  RIEGCD 				( A div B, d, u1, v1 )

  ( avec : d= B*u1+[A mod B]*v1 )
  ( [A mod B]=A-[A div B]*B )
  ( donc: d= A*v1+ B*[u1-[A div B]*v1] )
  ( v=v1, u=u1-[A div B]*v1 )
					( A div B, d, u1, v1 )
  4ROLL OVER QMul			( d, u1, v1, [A div B]*v1 )
  ROTSWAP RSUBext			( d, v1, u1-v1*A div B )
;


**********************************************************************
* Desc:
* Stack:	( a, b -> d, u, v t.q. d=pgcd[a,b], au+bv=d )
**********************************************************************
NULLNAME REGCDext
::
  FXNDext UNROTSWAP		( b.d, b.n, a )
  FXNDext UNROTSWAP		( b.d, a.d, a.n, b.n )
  EGCDext			( b.d, a.d, d, u, v )

* d=a.n*u+b.n*v -> a*a.d u + b*b.d = d

  SWAP4ROLL RMULText
  SWAP4ROLL RMULText
;

**********************************************************************
* Desc:		identit de Bezout pour des polynomes
* Stack:	( a, b -> d, u, v t.q. d=pgcd[a,b], au+bv=d )
**********************************************************************
NULLNAME EGCDext
::
  CODE
egcd00		ST=0	sSWAP
egcd0		C=DAT1	A
		CD1EX
		A=DAT1	11
		CD1EX
		C=A	W
		LCZ0
		?A#C	W
		GOYES	egcd1
drop01		C=B	A	execute SEMI
		CD0EX
		D0=D0-	5
		A=DAT0	A
		AD0EX
		B=A	A
		D=D+1	A
		?ST=0	sSWAP
		GOYES	drop01-
		LA(5)	=Z0
		GOTO	drop01+
drop01-		LA(5)	=Z1
drop01+		DAT1=A	A
		?ST=0	sSWAP
		GOYES	drop01--
		LA(5)	=Z1
		PC=(A)
drop01--	LA(5)	=Z0
		PC=(A)
egcd1		?ST=0	sSWAP
		GOYES	egcd1+
		ST=0	sSWAP
		GOTO	egcd1-
egcd1+		ST=1	sSWAP
egcd1-		LC(5)	=DOLIST
		?A=C	A
		GOYES	egcd3
egcd2		A=DAT1	A
		D1=D1+	5
		DAT1=A	A
		D1=D1-	5
		GOTO	drop01
egcd3		LA(5)	=DUPLENCOMP
		PC=(A)
  ENDCODE

						( a, b, #b )
  #1= IT
  CODE
		GOTO	egcd2
  ENDCODE
  SWAP
  CODE
		ST=1	sSWAP
		GOTO	egcd0
  ENDCODE
  #1= IT
  CODE
		GOTO	egcd2
  ENDCODE

						( b, a )
  SWAP						( a, b )

  Z1
  DUP 2LAMBIND

  ( a, b avec a polynome de degre >=1 )

  INEGCD					( d, u, v )
  ABND						( d, u, v )
  SWAPROT LPGCDext				( v, u, d, pgcd )
  ONE>POLY 4PICK GCDext 3PICK GCDext		( v, u, d, pgcd )
  DUPUNROT QUOText 4UNROLL			( d, v, u, pgcd )
  DUPUNROT QUOText UNROT QUOText		( d, u, v )
  3PICK DUPTYPECOL? NOTcaseDROP			( d, u, v, d )
  DUPLENCOMP #2= ITE RCONJext QCONJext		( d, u, v, d-barre )
  THREE ZERO_DO (DO)
     DUPROT RMULText 4UNROLL
  LOOP

  Z0
  2DROP

;

**********************************************************************
* Desc:
* Stack:	( a, b  -> d, u, v )
**********************************************************************
NULLNAME INEGCD
LABEL egcdlam1!
::
  ( avec d=au+bv )
  ( lam2 = g, lam1 = h )
  ( a, b )
  CODE
	GOTO	egcd00
  ENDCODE
  DUP#1= IT
  CODE
	D1=D1+	5
	D=D+1	A
	GOTO	egcd2
  ENDCODE
				( a, b, #b )
  OVER CARCOMP ONE>POLY		( a, b, #b, b1 )
  4PICK LENCOMP ROT		( a, b, b1, #a, #b )
  2DUP#< case2drop EGCDSWAP
  #-				( a, b, b1, #a-#b )
  DUPUNROT #1+ RP#		( a, b, #a-#b, alpha )
  DUP 5ROLL RMULText		( b, #a-#b, alpha, A=alpha*a )
  4ROLL DUPUNROT QDivRem	( #a-#b, alpha, b, A div b, A mod b )
  ROTSWAP			( #a-#b, alpha, A div b, b, A mod b )
  1GETLAM 6PICK RP#		( # alpha A div b, b, A mod b, h^# )
  2GETLAM RMULText		( # alpha A div b, b, A mod b, g*h^# )
  DUP4UNROLL RDIVext		( # alpha A div b beta=g*h^# A'=b B'=A mod B/beta )
  OVER CARCOMP ONE>POLY DUP 2PUTLAM
				( # alpha A\b beta=g*h^# A'=b B'=A mod B/beta g' )
  7PICK RP#			( #, alpha, A div b, beta, A', B', g'^# )
  1GETLAM 8ROLL			( alpha, A div b, beta, A', B', g'^#, h, # )
  EGCDNEWG 1PUTLAM		( alpha, A div b, beta, A', B' )
  INEGCD			( alpha, A div b, beta, d, u1, v1 )

  ( avec : d= b*u1+[A mod b]/beta*v1 )
  ( [A mod b]=a*alpha-[A div b]*b )
  ( donc: beta*d= a*[alpha*v1]+ b*[u1*beta-[A div b]*v1] )

				( alpha, A div b, beta, d, u1, v1 )
  2SWAP OVER RMULText		( alpha, A div b, u1, v1, beta, beta*d )
  3PICK 7ROLL RMULText		( A div b, u1, v1, beta, beta*d, v1*alpha )
  ROT 5ROLL RMULText
  4ROLL 5ROLL RMULText RSUBext	( d*beta, v1*alpha, u1*beta- [A div B]* v1 )
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME EGCDSWAP
::
  DROPSWAP INEGCD SWAP
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME EGCDNEWG
::
  DUP#0=csedrp RMULText
  DUP#1= case2DROP
  #1- RP# QUOText
;

**********************************************************************
* Desc:
* Stack:	( polynome-liste -> derivee )
**********************************************************************
NULLNAME PDer
LABEL DERext
::
  DUPTYPELIST? NOTcasedrop Z0
  INNERDUP
  #1= case2drop Z0
  DUP #2= case2drop ONE>POLY
  UNROTDROP OVER UNROLL DUPTWO
  DO
    SWAPINDEX@ #>Z RMULText
    OVER UNROLL
  LOOP
  #1- {}POLY
;

**********************************************************************
* Desc:		calcule la primitive s'annulant en 0
* Stack:	( polynome-liste -> primitive )
**********************************************************************
NULLNAME INTEGRext
::
  CKINNERCOMP #1+DUP
  ONE_DO (DO)
     SWAPINDEX@ #>Z RDIVext
     OVER UNROLL
  LOOP
  Z0 SWAP >POLYTRIM
;

**********************************************************************
* Desc:		Left ReDiMension
* Stack:	( P # -> [] )
* Note:		adds 0 to the left of polynomial to get a symbolic vector
*		of lenght #+1
**********************************************************************
NULLNAME LRDMext
::				( P # )
  OVER TYPELIST?		( P # flag )
  NOTcase LRDMNOTLIST		( -> [] )
  #1+ OVER LENCOMP		( P #+1 #P )
  2DUP#<

  ( liste, entier+1, longueur de liste , ?entier+1<longueur de liste)

  case LRDMSHRINK

  #-				( P # )
  Z0 SWAP NDUPN			( P z0 .. z0 # )
  get1 INNERCOMP top& []N	( [] )
;

LABEL LRDMNOTLIST
::			( P # )
  Z0 SWAP NDUPN get1 SWAP#1+ []N
;

LABEL LRDMSHRINK
( liste trop longue: on supprime a gauche )
::			( P #+1 #P )
  DUPROT #-+1 SWAP
  SUBCOMP {}TO[]	( [] )
;	

**********************************************************************
* Desc:		Right ReDiMension: like LRDM but 0 at the right and {} 
* Stack:	( {} # -> {} )
**********************************************************************
NULLNAME RRDMext
::
  OVER TYPELIST? NOTcase RRDMNOTLIST
  #1+ OVER LENCOMP 2DUP#<

  ( liste, entier+1, longueur de liste , ?entier+1<longueur de liste)
  casedrop
    :: ONESWAP SUBCOMP ;	( liste trop longue: on supprime a droite )

  #-SWAP INNERCOMP		( # meta-P )
  get1 Z0 SWAP NDUPN top&	( meta )
  {}POLY
;

LABEL RRDMNOTLIST
:: Z0 SWAP NDUPN #1+ {}POLY ; 

**********************************************************************
* Desc:		degree of a list-polynomial
* Stack:	( {} -> degre )
**********************************************************************
NULLNAME DEGREext
::
  DUPTYPELIST? NcaseSIZEERR
  LENCOMP #1-
;

**********************************************************************
* Desc:		Schema de Horner
* Stack:	( polynome-liste P/d, r -> P[X] div [X-r]/d, r, P[r]/d )
**********************************************************************
NULLNAME FHORNER
::
  SWAP FXNDext			( r P d )
  CONSTANTPOLY? NcaseSIZEERR	
				( r P d )
  UNROTSWAP			( d P r )
  HORNext			( d Q r P[r] )
  ROT 4PICK QDiv		( d r P[r] Q/d )
  4UNROLLROT CKCARCOMP QDiv	( Q/d r P[r]/d )
;

**********************************************************************
* Desc:		Schema de Horner
* Stack:	( polynome-liste P, r -> P[X] div [X-r], r, P[r] )
**********************************************************************
NULLNAME HORNext
::
  CK&DISPATCH1
  EIGHTY  HORN1
  any     :: Z0 UNROTSWAP ; 
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME HORN1
::					( poly, r )
  SWAPINCOMP				( r, a0, .., an, #n+1)
  DUP#1= casedrop :: Z0 UNROT ;
  DUP #2+PICK				( r, a0, ..., an, #n+1, r )
  ( * test for r=infinity * )
  TESTINFINI				( r a0..an #n+1 r ? )
  case HORNINF
					( r a0, ..., an, #n+1, r )
  SWAP get1				( r a1, ..., an, r, #n+1, a0)
  UNROTDUP				( r a1, ...., an, b0, r, #n+1, #n+1 )
  ONE_DO (DO)				( r ak...an, b0, b1...bk-1, r, #n+1 )
     get1 2OVER RMULText RADDext UNROT
  LOOP
					( r b0, b1, ..., bn, r, #n+1)
  UNROTDROP				( r b0, ..., bn-1, #n+1, bn )
					( r, b0, ..., bn-1, #n+1, bn )
  OVER #1+UNROLL #1- {}POLY		( r, bn , {b0 ... bn-1 } )
  UNROT					( B, r, bn )
;

LABEL HORNINF
* evaluation at infinity for non-cst polynomial
::					( r a0..an #n+1 r )
  OVER TWO #/ DROP			( r a0..an #n+1 r #mod )
  #1= casedrop HORNUNSIGNEDINF
					( r a0..an #n+1 r )
  INNERCOMP metainftype			( r a0..an #n+1 # )
  #0= NOTcase HORNPINF
					( r a0..an #n+1 )
  NDROP ?ext SWAPDUP			( ? inf inf )
;

LABEL HORNUNSIGNEDINF
:: 					( r a0..an #n+1 )
  #1- NDROP 				( r a0 )
  ?ext UNROT PLUSINFext HORN*		( ? r r*a0 )
;

LABEL HORNPINF
:: 					( r a0..an #n+1 )
  #1- NDROP 				( r a0 )
  ?ext UNROTOVER HORN*			( ? r r*a0 )
;

LABEL HORN*
::
  2SYMBINCOMP top&addt* SYMBN
;

**********************************************************************
* Desc:		Horner pour des matrices, utilises par Jordan
* Stack:	( polynome-liste P, r -> P[X] div [X-r], r, P[r] )
**********************************************************************
NULLNAME MHORNext
::				( poly de matrices, r )
  SWAPINCOMP			( r, a0, .., an, #n+1)
  DUP#1= casedrop
   :: DUP Z0 MAT*SCL UNROT ;
  DUP #2+PICK			( r, a0, ..., an, #n+1, r )
  SWAP get1			( r, a1, ..., an, r, #n+1, a0)
  UNROTDUP			( r, a1, ...., an, b0, r, #n+1 , #n+1 )
  ONE_DO (DO)			( r, ak,...,an, b0, b1, ..., bk-1, r, #n+1 )
     get1 2OVER MAT*SCL 
     MADD
     UNROT
  LOOP
				( r, b0, b1, ..., bn, r, #n+1)
  UNROTDROP			( r, b0, ..., bn-1, #n+1, bn )
  OVER #1+UNROLL #1- {}POLY	( r, bn , {b0 ... bn-1 } )
  UNROT				( B, r, bn )
;

**********************************************************************
* Desc:		Taylor for polynomials
* Stack:	( polynome-liste P, r -> dvpt Taylor de P en r )
**********************************************************************
NULLNAME PTAYLext
::				( P r )
  num#0=case DROP
  CK&DISPATCH1
  EIGHTY :: OVER LENCOMP	( P r #n )
            DUP4UNROLL
            ZERO_DO (DO)
               HORNext 4UNROLL
            LOOP
            2DROP reversym {}POLY
	 ;
  symbany :: SWAP FXNDext			( r P d )
             CONSTANTPOLY? NcaseSIZEERR
						( r P d )
             UNROT DUPTYPESYMB? caseSIZEERR	
						( d r P )
             SWAP PTAYLext			( d P' )
             SWAP QDiv				( P'/d )
	     ;
  any DROP
;
**********************************************************************

**********************************************************************
* Desc:		Lagrange interpolation
* Stack:	( matrix -> symb )
*		for [ [ x1 .. xn ] [ f(x1) .. f(xn) ] ]
*		returns a polynomial P such that P(xi)=f(xi)
**********************************************************************
NULLNAME LAGRANGEext
::
  LVARext VXXLext			( lvar, liste )
  DUP CARCOMP				( lvar, liste, lrac )
  INNERCOMP Z1 SWAP
  ZERO_DO (DO)
     SWAP RACTOFACext NDXFext RMULText
  LOOP
					( lvar, liste, polynome Pi[x-x_i] )
  Z0 SWAPDUP FXNDext DROP LENCOMP
  ONE_DO				( lvar, liste, resultat, produit )
    DUP 4PICK CARCOMP
    INDEX@ NTHCOMPDROP		( lvar list result produit produit xk )
    FHORNER DROP		( lvar list result produit prod' xk )
    OVERSWAP HORNER1ext		( lvar list result produit prod' prod'[xk] )
    5PICK TWO NTHCOMPDROP
    INDEX@ NTHCOMPDROP	( lvar list result produit prod' prod'[xk] yk )
    SWAP RDIVext 
    ONE>POLY RMULText	( lvar list result produit yk/prod'[xk]*prod' )
    ROT RADDext
  SWAPLOOP
				( lvar, liste, resultat, produit )
  DROPSWAPDROP			( lvar, resultat )
  ZERO SWAPROT RCLVX >HCOMP
  VAL2ext
;
