INCLUDE erromid.h
INCLUDE cas2extdec.h
INCLUDE keydefs.h
INCLUDE global.s

	STITLE Conversion from internal to user type and positivity testing

**********************************************************************
* EXPORT:
*
* HSECO2RCext	( obj -> obj )
*		Conversion of all constants from internal to user form
*
* SECO2CMPext	( :: -> symb )
*		Back conversion of complex
*
* VALOBJext	( #, {..{ Q } ..}, { var1 ... varn } -> {..{ obj } ..}
*		Back conversion of objects embedded at depth # in lists
*		Simplifies var1 ... varn
*
* VAL2ext	( #, {..{ Q } ..}, { var1 ... varn } -> {..{ obj } ..}
*		Back conversion of objects embedded at depth # in lists
*		Does not simplify var1 ... varn
*		Conversion is done in asc. power if increaseflag is set
*		(useful for SERIES/LIMIT)
*
* INVAL2	( P, # -> poly-symb )
*		lam2 must contain the list of var, # is the depth 
*
* METAVAL2	( #, list-meta -> list-meta of symbolics )
*		lvar must be in LAM2, LAM1 is modified
*
* VAL1		( obj -> obj )
*		LAM2 must be bound to list of vars, LAM1 is modified
*
* VAL1M		( obj -> symbolic-meta )
*		LAM2 must be bound to list of vars, LAM1 is modified
*
* SYMBINCOMP	( obj -> obj, ONE )
*		( symb -> o1, ..., on, #n )
**********************************************************************


* Desc:		Set sign of P1 according to leading coeff of P2
* Stack:	( P1 P2 --> sign[P2]*P1 )
**********************************************************************
NULLNAME PSetSign
::

*---------------------------------------------------------------------
* Desc:		Get the leading integer coefficient of P
* Stack:	( P --> Z )
*---------------------------------------------------------------------
   NULLNAME PLCZ
   CODE
		C=DAT1	A
		CD0EX
		RSTK=C
		LC(5)	=DOLIST
PLCZloop	A=DAT0	A
		D0=D0+	5
		?A=C	A
		GOYES	PLCZloop
		D0=D0-	5
		C=RSTK
		CD0EX
		DAT1=C	A
		GOVLNG	=Loop
   ENDCODE

   ZIsNeg?		( P1 flag )
   NOT?SEMI QNeg	( P1' )
;

**********************************************************************
* Desc:		Conversion of constants from internal to user form.
* Stack:	( ob --> ob' )
**********************************************************************
NULLNAME HSECO2RCext
::
*  RMSYMext
  DUPTYPESYMB? ITE INNERDUP ONEONE
  #1-SWAP ZERO_DO
     get1
     DUPTYPECOL? IT VAL1SECO
  SWAPLOOP
  #1+ CKSYMBN
;

**********************************************************************
* Desc:		Back conversion of complex
* Stack:	( seco --> symb )
* Note:		polarflag should be disabled if not at the top level
*		of rational expressions
**********************************************************************
NULLNAME SECO2CMPext
::
  polarflag SysITE SECO2CMPPOL SECO2CMPCART
;

* --------------------------------------------------------------------
* Desc:		Conversion of a complex in polar coordinates
* Stack:
* Note:		should be used only at the top level of rational expr.
* --------------------------------------------------------------------
NULLNAME SECO2CMPPOL
::					( C )
  ( * must be in radians mode * )
  RADON
  polarflag ClrSysFlag
  DUP CNORMext XSQRext VAL1M		( C meta-|C| )
  get1 CXIRext VAL1 SWAP VAL1 SWAP	( meta-|C| im re )
  ARG2					( meta-|C| meta-arg )
  metai psh MetaMul			( meta-|C| arg*i )
  OVER Z0 EQUAL 			( meta-|C| arg*i flag )
  ITE NDROPZ1ONE addtEXPonly 
  MetaMul CKSYMBN
  polarflag SetSysFlag
;

* --------------------------------------------------------------------
* Desc:		Conversion of a complex in cartesian coordinates
* Stack:
* --------------------------------------------------------------------
NULLNAME SECO2CMPCART
::
  CXIRext			( im re )
  valmz get1			( meta-re im )
  valmz metai metamult		( meta-re meta-im*i )
  metaadd CKSYMBN
;

**********************************************************************
* Desc:		Back conversion of objects embedded at depth #
*		in lists. Simplifies var1..varn.
* Stack:	( # {..{Q}..} {var1..varn} --> {..{ob}..} )
**********************************************************************
NULLNAME VALOBJext
::
  SLVARext VAL2ext
;

**********************************************************************
* Desc:		Back conversion of a scalar object
* Stack:	( lvar obj --> obj )
**********************************************************************
NULLNAME R2SYM
::
  ZERO SWAPROT VAL2ext
;

**********************************************************************
* Desc:		Back conversion of objects embedded at depth #
*		in lists. Does not simplify var1..varn.
*		Conversion is done in asc. power if positivfflag is set,
*		which is useful for SERIES and LIMIT commands.
* Stack:	( # {..{Q}..} {var1..varn} --> {..{ob}..} )
**********************************************************************
NULLNAME VAL2ext
::						( # P Lvar )

* valprof is used for the current variable in valvar

  DEFINE valprof  1GETLAM
  DEFINE valprof! 1PUTLAM
  DEFINE valvar   2GETLAM

  ROTSWAP				( P # lvar )
  ZERO 2LAMBIND				( P # )
  INVAL2				( symb )
  ABND
;

**********************************************************************
* Desc:
* Stack:	( P # --> symbpoly )
* Note:		LAM2 must contain Lvar, # is the depth.
**********************************************************************
NULLNAME INVAL2
::
  DUP#0=csedrp VAL1			( composite # -> composite )
  #1-SWAP DUPTYPETAG?			( # composite flag )
  DUP4UNROLL
  IT :: 'EvalNoCK: xOBJ> 4UNROLL ;
					( [tag] tagged? # composite )
  DUP OBJ>R INVAL2INCOMP		( [tag] tagged? # meta )
  METAVAL2				( [tag] tagged? meta )
  R>OBJ ERABLETYPE COMPN 			( [tag] tagged? composite )
  SWAP NOT?SEMI
  SWAP >TAG
;

LABEL INVAL2INCOMP
::
  dLISTcase INNERCOMP
  dMATRIXcase INNERCOMP
  ONE
;

**********************************************************************
* Desc:
* Stack:	( # Meta_list --> Meta_symb )
* Note:		LMA2 must contain Lvar, LAM1 is modified
**********************************************************************
NULLNAME METAVAL2
::
  DUP#0=case SWAPDROP
  get1 OVER
  ZERO_DO (DO)			( ak, ..., an, a1,...,ak-1, #n, # )
     OVER #2+ROLL OVER		( ak+1, ..., an, a1,...,ak-1,#n, #, ak,#)
     INVAL2 UNROT
  LOOP
  DROP 
;

**********************************************************************
* Desc:
* Stack:	( ob --> ob )
* Note:		LAM2 must contain Lvar, LAM1 is modified
**********************************************************************
NULLNAME VAL1
::
  CK&DISPATCH1
  zint    VAL1Z
  seco    VAL1SECO
  list    VAL1LIST
  symb    VAL1SYMB
  ereal   %%>%
  ecmp    C%%>C%
  TAGGED  VAL1TAG
  any     NOP
;

**********************************************************************
* Desc:
* Stack:	( ob --> Meta_symb )
* Note:		LAM2 must contain Lvar, LAM1 is modified
*		returns always a valid meta
**********************************************************************
NULLNAME VAL1M
::			( ob )
  val1m!		( meta )
  DUP#0= NOT?SEMI
  Z0 SWAP#1+
;


LABEL val1m!
* obj -> meta, might return a null meta
CODE
val1m		C=DAT1	A
		CD1EX
		A=DAT1	A
		CD1EX
		LC(5)	=DOINT
		?A#C	A
		GOYES	valmtlst
valmhxs		LA(5)	valmhxs!
		PC=(A)
valmtlst	LC(5)	=DOLIST
		?A#C	A
		GOYES	valmtsymb
valmlist	LA(5)	valmlist!
		PC=(A)
valmtsymb	LC(5)	=DOSYMB
		?A#C	A
		GOYES	valmnhxs
valmsymb	LA(5)	valmsymb!
		PC=(A)
valmnhxs	A=DAT1	A
		LC(5)	#70000
		?A>C	A
		GOYES	valnotinrom
		LC(5)	=%%0
		?A=C	A
		GOYES	valnotinrom
		LA(5)	=ONE
		PC=(A)
valnotinrom	LA(5)	valmnhxs!
		PC=(A)
ENDCODE

* zint -> meta, returns zabs neg for a neg zint
* WARNING returns null-meta for 0
LABEL valmhxs!
::
  Z0 EQUALcasedro ZERO
  valmz
;

LABEL valmnhxs!
* obj -> meta
::
  VAL1 DUPTYPESYMB? ITE INNERCOMP ONE
;

* zint -> meta, returns zabs neg for a negative zint
LABEL valmz
::
  DupZIsNeg?				( z, ?neg )
  NOTcase valmznock
  ZABS VAL1Z ' xNEG TWO
;

* zint -> meta as zint one
LABEL valmznock
:: VAL1Z ONE ;

* zint -> zint, push again zint for rom pointers
LABEL VAL1Z
CODE
	C=DAT1	A
	CD1EX
	D1=D1+	5
	A=DAT1	A	Z length
	CD1EX
	LC(5)	6
	?A#C	A
	GOYES	+
	LA(5)	=Z0
	DAT1=A	A
	GOTO	val1zlp
+	LC(1)	8
	?A<C	A
	GOYES	++	true if Z too long
val1zlp	GOVLNG	=Loop
++	GOSBVL	=SAVPTR
	A=0	W
	C=DAT1	A
	CD1EX
	D1=D1+	10
	A=DAT1	1
	D1=D1+	1
	C=DAT1	S
	GOVLNG	=DrpPushZLoop
ENDCODE

* seco -> symb
LABEL VAL1SECO
::
  POSITIFext DUPUNROT ?SKIP RNEGext	( signe, obj )
  DUP CARCOMP DUPTYPEZ?			( signe, obj, 1er element, ?hxs )
  casedrop 
  ::					( signe, obj )
    SECO2CMPext				( signe, obj )
    SWAP ?SKIP RNEGext
  ;
					( signe, obj, 1er element )
  EQ: x<< NOTcase SWAPDROP
  INNERCOMP #5= NOTcase INTERNALERR
  DROP 4ROLLDROP			( flag a, b, c )
  ( * if c is a zint, and a or b are complex, we check for polarmode * )
  complexflag TestSysFlag NOTcase VAL1IRRQ
  DUPTYPEZ? NOTcase VAL1IRRQ
  ( ?>0 a b c )
  polarflag TestSysFlag NOTcase VAL1IRRQ
  NDXQext SWAP ?SKIP QNeg
  ( irrq )
  SECO2CMPPOL
;

LABEL VAL1IRRQ
::
  SWAP VAL1M				( flag a c meta-b )
  ( * if meta-b is a square root or something*sqrt, combine with meta-c * )
  M1st*case VAL1IRRQ*
  AEQ1stcase x\v/ VAL1IRRQCOMBINESQRT
  get1 VAL1M addtSQRTonly		( flag a meta-b meta-sqrtc )
  MetaMul				( flag a meta-b*meta-sqrtc )
  VAL1IRRQEND				( -> symb )
;

LABEL VAL1IRRQ*
::					( flag a c meta-b )
  get1 VAL1M psh			( flag a meta-c meta-b )
  pulldrop larg				( flag a meta-c meta-b1 meta-b2 )
  AEQ1stcase x\v/ VAL1IRRQSQRT*
					( flag a meta-c meta-b1 meta-b2 )
  MetaMul psh addtSQRTonly 		( flag a meta-b1*meta-b2 meta-cSQRT )
  MetaMul VAL1IRRQEND
;

LABEL VAL1IRRQSQRT*
::					( flag a meta-c meta-b1 meta-b2 )
  pulldrop rot MetaMul addtSQRTonly	( flag a meta-b1 meta-b2*cSQRT )
  MetaMul VAL1IRRQEND
;

LABEL VAL1IRRQCOMBINESQRT
::					( flag a c meta-bSQRT )
  pulldrop get1 VAL1M MetaMul		( flag a meta-b*c )
  addtSQRTonly VAL1IRRQEND		( -> symb )
;

LABEL VAL1IRRQEND
::					( flag a meta-b*meta-sqrtc )
  get1 VAL1M psh MetaAdd		( flag meta )
  get1 case CKSYMBN			( -> symb )
  metaneg CKSYMBN
;

* tagged -> tagged
LABEL VAL1TAG
::
  'EvalNoCK: xOBJ>
  SWAP VAL1
  SWAP >TAG
;

* symb -> symb
LABEL VAL1SYMB
::
  valmsymb!
  CKSYMBN
;

* ( symb -> meta )
LABEL valmsymb!
::
  DUP >R LENCOMP 			( # )
  ZEROSWAP				( meta # )
  ZERO_DO (DO)				( meta )
    RSWAP 'R RSWAP			( meta obj )
    valmsymblp				( meta )
  LOOP
  EXACT? NOT?SEMI
  M1st/case valmsymb/!
;

LABEL valmsymblp
::
  DUPEQ: xABS casedrop addtABSonly	( meta )
					( meta obj )
  VAL1MH				( meta meta )
  top&					( meta )
;

LABEL VAL1MH
* obj -> meta
* if obj is a zint tries to rewrite it as a rom pointer
::
  DUPTYPEZ? NOTcase val1m!
  VAL1Z ONE
;

LABEL valmsymb/!
::
  pulldrop larg 			( meta-n meta-d )
  ( * check for ABS factored in deno * )
  valmsymb/*				( meta-n meta-d meta- )
  dup addtSQ				( meta-n meta-d meta- meta- )
  rot metamult unrot			( meta-*meta-d meta-n meta- )
  addtABS metamult psh metadiv		( -> )
;

LABEL valmsymb/*
::					( meta-d )
  AEQ1stcase xABS :: pulldrop Z1ONE psh ;	( -> )
  OVER EQ: x*				( meta-d ? )
  NOTcase Z1ONE				( -> meta-d meta-abs )
  pulldrop larg				( meta-df1 meta-df2 )
  valmsymb/* rot valmsymb/*		( meta-d2 meta-abs2 meta-d1 meta-abs1 )
  rot metamult unrot			( meta-abs meta-d2 meta-d1 )
  psh metamult psh			( meta-d meta-abs )
;

* list -> symb
LABEL VAL1LIST
::
  valmlist! ( was: val1m! )
  DUP#0=csedrp Z0
  OVER EQ: xNEG NOTcase CKSYMBN
  3PICK TYPEZ? NOTcase CKSYMBN
  2DROP QNeg
;

LABEL valmlist!
::	
  POSITIFext DUP OBJ>R
  ?SKIP QNeg				( list )

  ( * !: QNeg might return a non-list * )
  DUPTYPELIST? ?SKIP ONE{}N		( list ) 

  valvar valprof #1+DUP valprof! NTHELCOMP	
					( liste, [ var, TRUE | FALSE ] )
  ITE vallistsym! ONE			( meta )
  valprof #1- valprof!			( meta )
  R>OBJ ?SEMI
  metackneg
;


LABEL vallistsym!
( list, idnt/sym -> meta )
::
  increaseflag SysITE valasc! valdec!	( idnt, meta-res )
  get1 DROP				( meta-res )
;

LABEL valdec!
::
					( list, idnt )
  SWAPDUP				( idnt, liste, liste )
  LENCOMP DUP#1= casedrop valdeccst	( idnt, liste, #len )
  SWAP >R				( idnt, #len )
  'R 3PICK				( idnt, #len, an, idnt )
  OBJ>R val1m! get1			( idnt, meta-an, #len )
  R>OBJ					( idnt, meta-an, #len, idnt )
  SYMBINCOMP get1 #1-DUP OBJ>R addtpow!	( idnt, meta-an*X^n, idnt )
  metamult R>OBJ			( idnt, meta-an*X^n, #n-1 )
  DUP#1= casedrop valdecend
  ONE_DO (DO)				( idnt, meta-res )
	DUP #2+PICK RSWAP 'R RSWAP	( idnt, meta-res, idnt, ak )
	SWAP OBJ>R val1m! R>OBJ		( idnt, meta-res, meta-ak, idnt )
	CODE
		D1=D1+	5
		A=DAT1	A
		D1=D1-	5
		LC(5)	=ZERO
		?A#C	A		first test if ak=0
		GOYES	valdecn0
		D1=D1+	10
		D=D+1	A		yes, then drop meta-ak and idnt
		D=D+1	A
		GOVLNG	=Loop
valdecn0	D1=D1+	10		ak!=0, test sign of meta-ak
		A=DAT1	A
		D1=D1-	10
		LC(5)	=xNEG
		?A#C	A
		GOYES	valdecpos
		D1=D1+	15
		A=DAT1	A		meta-ak is < 0, is it -1?
		D1=D1-	15
		LC(5)	=Z1
		?A#C	A
		GOYES	valdecneg
		A=DAT1	A		meta-ak is -1, do 4unroll3drop
		D1=D1+	15
		D=D+CON	A,3
		DAT1=A	A
		LA(5)	=valdec-1!
		PC=(A)
valdecneg	LA(5)	=valdecneg!
		PC=(A)
valdecpos	LC(5)	=Z1		meta-ak > 0, is it 1?
		?A#C	A
		GOYES	valdecnorm
		A=DAT1	A		meta-ak is 1, do unrot2drop
		D1=D1+	10
		D=D+1	A
		D=D+1	A
		DAT1=A	A
		LA(5)	=valdec1!
		PC=(A)
valdecnorm	LA(5)	=valdecnorm!	meta-ak > 0 and not 1
		PC=(A)
	ENDCODE
	LOOP
					( idnt, meta-res )
  ( Warning, calling valdecend needs to COLA it )
  ( since return stack is used for data )
  'R val1m! DUP#0=csDROP
  valmetaadd
;

LABEL valdecend
::
  'R val1m! DUP#0=csDROP
  valmetaadd
;

LABEL valdec1!
::
  SYMBINCOMP ISTOP-INDEX addtpow! valmetaadd
;

LABEL valdec-1!
::
  SYMBINCOMP ISTOP-INDEX addtpow! valmetasub
;

LABEL valdecneg!
::
  ( suppress xNEG from meta-ak )
  ROTDROP SWAP#1-SWAP
  SYMBINCOMP ISTOP-INDEX addtpow!	( idnt, meta-res, meta-ak, meta-idnt^ )
  top& addt* valmetasub			( idnt, meta-res )
;

LABEL valdecnorm!
::
					( idnt, meta-res, meta-ak, idnt )
  SYMBINCOMP ISTOP-INDEX addtpow!	( idnt, meta-res, meta-ak, meta-idnt^ )
  top& addt* valmetaadd			( idnt, meta-res )
;

LABEL valmetaadd
::
  top& M1stNEGcase addtNEG+ addt+
;


LABEL valmetasub
::
  top& M1stNEGcase addtNEG- addt-
;

LABEL addtNEG+
::
  pulldrop addt-
;

LABEL addtNEG-
::
  pulldrop addt+
;

LABEL valdeccst
::
  ( idnt, liste )
  CARCOMP val1m!
;

LABEL addtpow!
( meta, # -> meta' )
::
  DUP#1= caseDROP
  ( * we simplify meta-ABS^# in real mode * )
				( meta # )
  COMPLEX? case addtpownow
  3PICK EQ: xABS NOTcase addtpownow
				( meta # )
  1LAMBIND
  larg				( meta1 meta2-ABS )
  pulldrop			( meta1 meta2 )
  1GETLAM TWO #/		( meta1 meta2 #rem #quot )
  DROP #1=case addtpowodd
				( meta1 meta2 )
  1GETABND addtpownow top&
;

LABEL addtpowodd
::				( meta1 meta2 )
  dup addtABSonly psh 		( m1 m2-abs m2 )
  1GETABND #1- addtpownow	( m1 m2-abs m2^ )
  metamult top&			( meta )
;

LABEL addtpownow
::
  EXACT? ITE #>Z UNCOERCE
  SWAP#1+ ' x^ SWAP#1+ 
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
LABEL valasc!
::

  ( returns monomials in ascending powers )
  SWAPINCOMP DUP#0<> NOTcase INTERNALERR

				( X an...a0 #n+1 )
  reversym get1 SWAPDUP		( a0...an X #n+1 #n+1 )
  ZERO_DO (DO)			( aj...an X met-res' #met-a+met-res )
    get1			( aj+1...an X, met-res' #met-a+met-res aj )
    val1m!			( aj+1...an X met-res' #met-a+met-res met-aj )
				( aj-1...a0 X met-res' #met-a+met-res met-aj )
    INDEX@
    ( * changed from ML to srpl for room * ) 
    #0=ITE valasc0! valascn0!
*    CODE
*		A=DAT1	A
*		D1=D1+	5
*		D=D+1	A
*		LC(5)	=ZERO
*		?A=C	A
*		GOYES	valasc0
*		LA(5)	=valascn0!
*		PC=(A)
*valasc0		LA(5)	=valasc0!
*		PC=(A)
*    ENDCODE
				( aj-1...a0 X met-res' #met-a+met-res+1 )
    #1-
  LOOP
;

* 1er element du meta, on ne met donc pas de +/- et si le meta est nul on met 0
* il n'y a pas de puissance a rajouter ici

LABEL valasc0!
:: 		( aj-1, ..., a0, X, meta-res', #meta-a+meta-res, meta-aj )
  DUP#0= NOTcase top&
  DROPZ0 SWAP#1+ 
;

LABEL valascn0!
::			( aj-1...a0 X met-res' #met-a+met-res met-aj )
  DUP#0=csDROP
  DUP #3+PICK Z0 EQ DUP OBJ>R
  IT addt0meta top&	( aj-1...a0 X new-met-res' #met-a+met-res )
  M1stNEGcase valascneg
			( aj-1...a0 X new-met-res' #met-a+met-res #n+1-index )
  valasc1!
  R>OBJ ?SKIP addt+ 
  top&
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
NULLNAME addt0meta
:: 
  psh pulldrop psh 
;

**********************************************************************
* Desc:
* Stack:
**********************************************************************
LABEL valascneg
::
  pulldrop valasc1!
  R>OBJ ITE addtNEG addt- 
  top&
;

* rajoute la puissance [cas ascendant]
LABEL valasc1!
:: 
  DUP ISTOP-INDEX #-	( aj-1...a0 X new-meta-res' #meta-a+meta-res # )
  #2+PICK 		( aj-1...a0 X new-meta-res' #meta-a+meta-res X )
  DUPTYPESYMB?
  ITE INNERCOMP ONE
  INDEX@ addtpow!	( aj-1...a0 X new-meta-res' #meta-a+meta-res meta-X^ )

* regarder si c'est 1*meta-X^

  DUP #3+PICK
  Z1 EQ NOTcase addt*
  psh pulldrop psh 
;
**********************************************************************
