INCLUDE cas2romid.h
INCLUDE erextdec.h
INCLUDE cas2ext.h
INCLUDE ../CAS/keydefs.h
INCLUDE ../CAS/global.s

	STITLE Trigonometric fraction integration utilities

**********************************************************************
* Desc:		Test if a fraction (internal rep) is even/odd/none
* Stack:	( fr --> Z )
*		Z=1 if even, -1 if odd, 0 if neither even nor odd
**********************************************************************
NULLNAME FRACPARITY
::
  FXNDext POLYPARITY		( n z )
  SWAPOVER 			( z n z )
  Z0 EQUAL caseDROP
				( z n )
  SWAP POLYPARITY QMul		( z )
;

**********************************************************************
* Desc:		Test if a polynomial (internal rep) is even/odd/none
* Stack:	( poly --> Z )
*		Z=1 if even, -1 if odd, 0 if neither even nor odd
**********************************************************************
NULLNAME POLYPARITY
::					( P )
  DUPTYPELIST? NOTcasedrop Z1
  DUPNULL{}? caseSIZEERR
  DUPLENCOMP DUP#1= case2drop Z1

  TWO #/				( P #rem #quo )
  SWAP #1= ITE Z1 Z-1			( P #quo z )
  ROT >R SWAP				( z #quo )
  ZERO_DO (DO)
    RSWAP 'R'R RSWAP			( z an an-1 )
    SWAPDROP Z0 EQUAL			( z flag )
    ?SKIP DROPZ0_7
  LOOP					( z )
*  ticR caseDROP				( remove remaining coeff if present )
  RDROP
;

LABEL DROPZ0_7
:: DROP Z0 ;

NULLNAME PARITYTEST
**********************************************************************
* Desc:		Test if a fraction of tan[x/2] can be integrated
*		using cos[x]/sin[x] or tan[x] change of var
* Stack:	( tan fraction-l -> sin/cos/tan fraction-l )
**********************************************************************
::
  EXACT? NOT?SEMI
  COSTEST ?SEMI
					( var fr )
  DUP FR2ND% %1+			( var fr N D % )
* check degree: n-d+2%=0
  DUP %+				( var fr N D 2% )
  3PICK DUPCKLEN{} SWAPDROP UNCOERCE %+	( var fr N D 2%+n+1 )
  OVER DUPCKLEN{} SWAPDROP UNCOERCE %-	( var fr N D 2%+n-d )
  %0= NOTcase2DROP			( -> var fr )
					( var fr N D )
  SINTEST ?SEMI

* TAN test					( var fr N D )
* the fraction, including dt, must be invariant by t->-1/t change of var
* or [equivalent] tF(t) must change sign
* if so, then u=-(2t)/(t^2-1)
* or 2/u=-t+1/t
* and F(t)dt= t F(t) (t^2+1)/(t^2-1) [-du/u]
* look for N and D symmetry
  OVER POLYASYM				( var fr N D ZN )
  DUP Z0 EQUALcase 3DROP			( -> var fr )
					( var fr N D ZN )
  OVER POLYASYM				( var fr N D ZN ZD )
  DUP Z0 EQUALcase 4DROP			( -> var fr )
  					( var fr N D ZN ZD )
  QMul Z-1 EQUAL			( var fr N D ? )
  NOTcase2DROP				( -> var fr )

					( var fr N D )
  SWAP					( var fr D N )
  MULTIPLY2				( var fr D N )
  SWAP SHRINK2ASYM			( var fr N' D' )
  Z-2 INTERNALX NDEvalN/D		( var fr n d )
  INTERNALX QNeg QMul QDiv SWAPDROPSWAP	( n/d*u var )
  INNERCOMP pulldrop meta2* 
  metasimp addtTAN SYMBN 		( n/d*u TAN )
  SWAP 					( var fr )
;

NULLNAME COSTEST
**********************************************************************
* Desc:		Test if a fraction of tan[x/2] can be integrated
*		using cos[x] change of var
* Stack:	( tan fraction-l -> cos/tan fraction-l )
**********************************************************************
::
* is the fraction odd wrt to t -> -t
* if so, we will use u=cos[x] instead of t=tan[x/2]
* we have u=[1-t^2]/[1+t^2], hence t^2=[1-u]/[1+u]
* and t*dt = - du/[1+u^2]
* if no parity detected, continue with tan[x/2]
  DUP FXNDext			( tan fr n d )
  POLYPARITY 			( tan fr n z )
  Z0 EQUALcasedro DROPFALSE	( -> tan fr false )
				( tan fr n zd )
  SWAP POLYPARITY		( tan fr zd zn )
  Z0 EQUALcasedro DROPFALSE	( -> tan fr false )
				( tan fr zd zn )
  Z1 EQUALcase COSTESTn1
  				( tan fr zd )
  Z1 EQUAL NOTcaseFALSE		( -> tan fr zd )
				( tan fr )
  ( * num is odd, deno is even * )
  FXNDext 			( tan n d )
  SWAP INTERNALX QDiv SWAP	( tan n/x d )
  COSMAKE
;

LABEL COSTESTn1
* num is even, test that deno is odd
::				( tan fr zd )
  Z1 EQUAL caseFALSE
  FXNDext INTERNALX QMul	( tan n d*x )
  COSMAKE
;

LABEL COSMAKE
::				( tan n d )
  SHRINK2EVEN			( tan n' d' )
  APPLY1-u/1+u			( tan num deno )
  INTERNALX+1^2 QNeg QMul	( tan num deno' )
  QDiv				( tan fr )
  SWAPINCOMP pulldrop		( fr meta )
  meta2* metasimp addtCOS	( fr meta-COS )
  SYMBN SWAP TRUE		( cos fr )
;

LABEL INTERNALX+1^2
:: { ZINT 1 ZINT 2 ZINT 1 } TOTEMPOB ;

LABEL SHRINK2EVEN
::					( n d )
  SWAP SHRINKEVEN SWAP SHRINKEVEN	( n' d' )
;

**********************************************************************
* Desc:		change var Y=X^2 in an even polynomial
* Stack:	( P -> P' )
**********************************************************************
NULLNAME SHRINKEVEN
::
  DUPTYPELIST? NOT?SEMI
  DUPLENCOMP 
  DUP#1= caseDROP
  #2/ SWAP >R ZEROSWAP			( meta # )
  ZERO_DO (DO)
    RSWAP 'R'R RSWAP			( meta an an-1 )
    DROP SWAP#1+			( meta )
  LOOP
  ticR NcaseSIZEERR
  SWAP#1+ {}N
;

LABEL APPLY1-u/1+u
::
  Z-1 Z1 TWO{}POLY
  Z1Z1 TWO{}POLY
  NDEvalN/D
;

NULLNAME SINTEST
**********************************************************************
* Desc:		Test if a fraction of tan[x/2] can be integrated
*		using sin[x] change of var
* Stack:	( tan fraction-l N D -> sin/tan fraction-l true )
*		(                    -> tan fr N D false )
**********************************************************************
::
* the fraction, including dt, must be invariant by t->1/t change of var
* or [equivalent] tF(t) must change sign
* if so, then u=(2t)/(1+t^2)
* or 2/u=t+1/t
* and F(t)dt= t F(t) (t^2+1)/(t^2-1) [-du/u]
					( var fr N D )
* look for N and D symmetry
  OVER POLYSYM				( var fr N D ZN )
  DUP Z0 EQUALcase DROPFALSE			( -> var fr N D FALSE )
					( var fr N D ZN )
  OVER POLYSYM				( var fr N D ZN ZD )
  DUP Z0 EQUALcase 2DROPFALSE			( -> var fr N D FALSE )
  					( var fr N D ZN ZD )
  QMul Z-1 EQUAL			( var fr N D ? )
  NOTcaseFALSE				( -> var fr N D FALSE )

					( var fr N D )
  MULTIPLY2 				( var fr N D )
  SHRINK2SYM				( var fr N' D' )
  Z2 INTERNALX NDEvalN/D		( var fr n d )
  INTERNALX QNeg QMul QDiv SWAPDROPSWAP	( n/d*u var )
  INNERCOMP pulldrop meta2* 
  metasimp addtSIN SYMBN 		( n/d*u SIN )
  SWAP TRUE				( var fr true )
;

LABEL MULTIPLY2
::					( N D -> N D )
  SWAP INTERNALt^2+1 QMul SWAP INTERNALt^2-1 QMul SIMPext
;

LABEL INTERNALt^2+1
:: { ZINT 1 ZINT 0 ZINT 1 } TOTEMPOB ;

LABEL INTERNALt^2-1
:: { ZINT 1 ZINT 0 ZINT -1 } TOTEMPOB ;

**********************************************************************
* Desc:		Shrink 2 polynomials using symmetry properties
* Stack:	( N D -> N' D' )
**********************************************************************
NULLNAME SHRINK2SYM
:: SWAP SHRINKSYM SWAP SHRINKSYM ;

**********************************************************************
* Desc:		Shrink 1 polynomial using symmetry properties
* Stack:	( N -> N' )
*		degree of N must be even
*		note that if it is odd then N should be divided by X+1
**********************************************************************
NULLNAME SHRINKSYM
::
  DUPTYPELIST? NOT?SEMI
  DUPLENCOMP DUP#1= caseDROP
					( N # )
  #2/ SWAP 				( #quo N )
  INTERNALt^2+1 3PICK RP#		( #quo N x^2+1^# )
  ROT DUP1LAMBIND			( N x^2+1^# #quo )
  ZERO_DO (DO)				( N x^2+1^# )
    DUP INTERNALt^2+1 QUOText UNROT	( ...res x^2+1^#-1 N x^2+1^# )
    QDivRem SWAP CARCOMPext SWAP	( ...res x^2+1^#-1 quo' rem )
    INTERNALX QUOText 			( ...res x^2+1^#-1 quo rem/x )
    ROT					( ...res quo newN x^2+1^#-1 )
  LOOP					( ...res newN x^ )
  DROP CARCOMPext			( ...res newN' )
  1GETABND #1+ >POLYTRIM		( N' )
;


**********************************************************************
* Desc:		Shrink 2 polynomials using antisymmetry properties
* Stack:	( N D -> N' D' )
**********************************************************************
NULLNAME SHRINK2ASYM
:: SWAP SHRINKASYM SWAP SHRINKASYM ;

**********************************************************************
* Desc:		Shrink 1 polynomial using antisymmetry properties
* Stack:	( N -> N' )
*		degree of N must be even
*		note that if it is odd then N should be divided by X+1
**********************************************************************
NULLNAME SHRINKASYM
::
  DUPTYPELIST? NOT?SEMI
  DUPLENCOMP DUP#1= caseDROP
					( N # )
  #2/ SWAP 				( #quo N )
  INTERNALt^2-1 3PICK RP#		( #quo N x^2-1^# )
  ROT DUP1LAMBIND			( N x^2-1^# #quo )
  ZERO_DO (DO)				( N x^2-1^# )
    DUP INTERNALt^2-1 QUOText UNROT	( ...res x^2-1^#-1 N x^2-1^# )
    QDivRem SWAP CARCOMPext SWAP	( ...res x^2-1^#-1 quo' rem ) 
    INTERNALX QUOText 			( ...res x^2-1^#-1 quo' rem/x )
    ROT					( ...res quo newN x^2-1^#-1 )
  LOOP					( ...res newN x^ )
  DROP CARCOMPext			( ...res newN' )
  1GETABND #1+ >POLYTRIM		( N' )
;

**********************************************************************
* Desc:		Extract trivial power of fraction
* Stack:	( fraction-l -> N D % )
*		fraction=N/D*t^%
**********************************************************************
NULLNAME FR2ND%
::
  FXNDext P2P# UNCOERCE		( N D %d )
  ROT P2P# UNCOERCE		( D %d N %n )
  ROT %-			( D N % )
  ROTSWAP
;

**********************************************************************
* Desc:		Test symmetry of coef of polynomial
* Stack:	( P -> Z )
*		Z=1 for symmetric, -1 for anti, 0 otherwise
**********************************************************************
NULLNAME POLYSYM
::
  DUPTYPELIST? NOTcasedrop Z1
  DUPINCOMP reversym >POLYTRIM		( P Pbarre )
  2DUP QSub 				( P Pbarre P-Pbarre )
  Z0 EQUAL case2drop Z1
  QAdd Z0 EQUAL case Z-1 Z0
;

**********************************************************************
* Desc:		Test "antisymmetry" of coef of polynomial
* Stack:	( P -> Z )
*		Z=1 for symmetric, -1 for anti, 0 otherwise
**********************************************************************
NULLNAME POLYASYM
::
  DUPTYPELIST? NOTcasedrop Z1
  DUPDUP LENCOMP			( P P # )
  DUP #2/ TWO #/ DROP			( P P # #/2_mod_2 )
  #0=ITE Z-1 Z1				( P P # z )
  ROT >R				( P # z )
  ZERO SWAPROT				( P meta z # )
  ZERO_DO (DO)				( P meta z )
    RSWAP 'R RSWAP			( P meta z ak )
    SWAP QNeg DUPUNROT			( P meta -z ak -z )
    QMul				( P meta -z ak )
    UNROT SWAP#1+SWAP			( P meta&ak -z )
  LOOP
  DROP reversym {}N			( P Pbarre )

  2DUP QSub 				( P Pbarre P-Pbarre )
  Z0 EQUAL case2drop Z1
  QAdd Z0 EQUAL case Z-1 Z0
;

**********************************************************************
* Desc:		Extract trivial power of poly
* Stack:	( P -> P' # )
*		fraction=P'*t^#
* Note:		P must be a valid poly (if list, begin with a non zero coeff)
**********************************************************************
NULLNAME P2P#
::
  NOTLISTcase ZERO
  DUPINCOMP 				( P p1..pk #k )
  OVER Z0 EQUAL 			( P p1..pk #k ? )
  NOTcase NDROPZERO			( -> P #0 )
					( P p1..pk #k )
  ZERO
  BEGIN					( p1..pi #i # )
    3PICK Z0 EQUAL			( p1..pi #i # ?pi=0 )
  WHILE
    ROTDROP SWAP#1-SWAP #1+		( p1..pi-1 #i-1 #+1 )
  REPEAT
  OVER#2+UNROLL {}N 			( P # P' )
  ROTDROPSWAP				( P' # )
;

**********************************************************************
* Desc:		Eval list poly over a list fraction
* Stack:	( num deno n d -> num" deno" )
*		where num[n/d]/deno[n/d]=num"/deno"
**********************************************************************
NULLNAME NDEvalN/D
::					( num deno n d )
  4ROLL 3PICK ROT			( deno n num' n d )
  PEvalN/D				( deno n num' d #d )
  5UNROLL SWAP 4UNROLL			( #d num deno n d )
  PEvalN/D				( #d num' deno' d #n )
  5ROLL 2DUP#>				( num' deno' d #n #d ?n>d )
  case NDEvalN/D>
					( num' deno' d #n #d )
  SWAP#- RP# QMul			( num" deno" )
;

LABEL NDEvalN/D>
::					( num' deno' d #n #d )
  #- RP# 				( num' deno' d^#n-#d )
  ROT QMul SWAP				( num" deno" )
;

**********************************************************************
* Desc:		Eval list poly over a list fraction
* Stack:	( P n d -> num d # )
*		where P[n/d]=num/d^#
**********************************************************************
NULLNAME PEvalN/D
::					( P n d )
  ROT DUPTYPELIST? 			( n d P ? )
  NOTcase :: ROTDROPSWAP ZERO ;		( -> P d #1 )
					( n d P )
  DUPLENCOMP DUP#0= caseSIZEERR
					( n d P #k+1 )
  ( * Warning: we must make coeff of P one level deeper * )
  #1- Z1 SWAPROT			( n d d^ # P )
  5UNROLL 4NULLLAM{} BIND		( P )
  FIVE NEXTCOMPOB DROP ONE>POLY		( P # pk )
  [BEGIN]				( P # num )
    UNROT NEXTCOMPOB
  [WHILE]
    PEvalN/DW
  [REPEAT]
					( num P )
  DROP					( num )
  3GETLAM 1GETABND			( num d # )
;

LABEL PEvalN/DW
::					( num P # pj )
  ONE>POLY 2GETLAM 3GETLAM QMul		( num P # pj d^ )
  DUP 2PUTLAM QMul			( num P # pj*d^ )
  4ROLL 4GETLAM QMul QAdd		( P # pj*d^+num*n )
;